/* @(#) ShortTermTimer_FT.c 96/07/17 1.1 */
/*	
 *	File:				ShortTermTimer_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains contains a functional test which verifies that the independent clocks
 *	created by the user keeps accurate time over a short period of time. There are four tests.
 *
 *	Test 1:	Repeatedly read the times from the four independent clocks (set at the same clock rate)
 *			Verify that the time is increasing for each of the clocks at the same rate as the global
 *			clock.
 *
 *	Test 2:	Set the clock rates to rates which are different and multiples of one another.  Check
 *			to see that the rates in which the clocks are increasing are the same multiples.
 *
 *	Test 3:	Run the fastest clock for extended period of time to see if it keeps accurate time.
 *
 *	Test 4:	Test GetAudioClockRate by finding the clock rates of each of the independent clocks.
 *
 *	Test 5:	Test AudioTimeLaterThan, AudioTimeLaterThanOrEqual, CompareAudioTimes
 *
 *
 *	Audio Functions called:
 *
 *	OpenAudioFolio		CloseAudioFolio				CreateAudioClock			GetAudioClockRate		
 *	ReadAudioClock		DeleteAudioClock			SleepUntilAudioTime				
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <stdio.h>

#include "TestMacros.h"
#include <tests/3do_scaf.h>


#define		MAX_ITERATIONS		4
/**************************************Test Environment Globals****************/
/******************************************************************************/

int main (int argc, char ** argv)
{
	AudioTime	time1, time2, time3, time4, globalTime,
				prevTime1 = 0, prevTime2 = 0, prevTime3 = 0, prevTime4 = 0, prevGlobalTime = 0,
				diff1, diff2, diff3, diff4, diffGlobal,
				tolerance = 1;	
	Err			theError;
	float32		clockRate1, clockRate2, clockRate3, clockRate4;
	int			i, j, counter = 1;
	Item		clock1, clock2, clock3, clock4, theCue;


	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT, theCue,"CreateCue", cleanup);

	/*	Create the clocks	*/
	clock1	= CreateAudioClock(NULL);
	CHECKRESULT_SETUP(FT,clock1,"CreateAudioClock",cleanup);
	clock2	= CreateAudioClock(NULL);
	CHECKRESULT_SETUP(FT,clock2,"CreateAudioClock",cleanup);
	clock3	= CreateAudioClock(NULL);
	CHECKRESULT_SETUP(FT,clock3,"CreateAudioClock",cleanup);
	clock4	= CreateAudioClock(NULL);
	CHECKRESULT_SETUP(FT,clock4,"CreateAudioClock",cleanup);


	/****************************************************************************************************
	 *	Test 1:	Print out the times for the four independent clocks.  Calculate the difference between
	 *			clock's current time and its previous time.  This tests checks to see if the differences
	 *			between current and previous times between the four independent and the global clock are
	 *			the same.  The differences should be the same because all of the clocks are running at
	 *			the same rate.
	 ****************************************************************************************************/
	LogComment( "1. Test to see if the four independent clocks are in sync with the global clock");
	for ( i = 0; i < MAX_ITERATIONS; i++ )
	{
		ReadAudioClock( clock1, &time1 );
		ReadAudioClock( clock2, &time2 );
		ReadAudioClock( clock3, &time3 );
		ReadAudioClock( clock4, &time4 );

		globalTime	= GetAudioTime();

		LogComment("Clock #1: %5d.\t\tClock #2: %5d.\t\tClock #3: %5d.\t\tClock #4: %5d.\t\tClock (global): %5d.", time1, time2, time3, time4, globalTime);
		
		if ( i != 0 )
		{
			diff1		= time1 - prevTime1;
			diff2		= time2 - prevTime2;
			diff3		= time3 - prevTime3;
			diff4		= time4 - prevTime4;
			diffGlobal	= globalTime - prevGlobalTime;

			LogComment("Diff  #1: %5d.\t\tDiff  #2: %5d.\t\tDiff  #3: %5d.\t\tDiff  #4: %5d.\t\tDiff  (global): %5d.", diff1, diff2, diff3, diff4, diffGlobal);		

			BeginTest(FT, "ShortTermTimer: Global/ Local Clock Synchronization");
			
			if (	( (diff1 == (diff2 + tolerance)) || (diff1 == (diff2 - tolerance)) || (diff1 == diff2) )	&&
					( (diff2 == (diff3 + tolerance)) || (diff2 == (diff3 - tolerance)) || (diff2 == diff3) )	&&
					( (diff3 == (diff4 + tolerance)) || (diff3 == (diff4 - tolerance)) || (diff3 == diff4) )	&&
					( (diff4 == (diffGlobal + tolerance)) || (diff4 == (diffGlobal - tolerance)) || (diff4 == diffGlobal) )  )
				LogResult(PASSED, "" );
			else
				LogResult(FAILED, "Clocks are not in sync.  Same amount of time has not elapsed for each clock" );

		}
	
		prevTime1		= time1;
		prevTime2		= time2;
		prevTime3		= time3;
		prevTime4		= time4;
		prevGlobalTime	= globalTime;
	}

	/****************************************************************************************************
	 *	Test 2:	Set different rates for the four independent clocks.  The rates are whole number ratios
	 *			of each other. Print out the times for the four independent clocks.  Calculate the 
	 *			difference between clock's current time and its previous time.  This tests checks to see
	 *			if the differences between current and previous times between the four independent and the 
	 *			global clock are the same ratios as the clock rates.  
	 ****************************************************************************************************/
	LogComment("Set different rates of independent clocks");
	LogComment("44100 Hz\t\t\t\t\t480 Hz\t\t\t240 Hz\t\t\t\t60 Hz\t\t\t\t\t240Hz");

	SetAudioClockRate(clock1, 44100);
	SetAudioClockRate(clock2, 480);
	SetAudioClockRate(clock3, 240);
	SetAudioClockRate(clock4, 60);
	
	for ( i = 0; i < MAX_ITERATIONS; i++ )
	{
		ReadAudioClock( clock1, &time1 );
		ReadAudioClock( clock2, &time2 );
		ReadAudioClock( clock3, &time3 );
		ReadAudioClock( clock4, &time4 );

		globalTime	= GetAudioTime();

		LogComment("Clock #1: %5d.\t\tClock #2: %5d.\t\tClock #3: %5d.\t\tClock #4: %5d.\t\tClock (global): %5d.", time1, time2, time3, time4, globalTime);
		
		if ( i != 0 )
		{
			diff1		= time1 - prevTime1;
			diff2		= time2 - prevTime2;
			diff3		= time3 - prevTime3;
			diff4		= time4 - prevTime4;
			diffGlobal	= globalTime - prevGlobalTime;

			LogComment("Diff  #1: %5d.\t\tDiff  #2: %5d.\t\tDiff  #3: %5d.\t\tDiff  #4: %5d.\t\tDiff  (global): %5d.", diff1, diff2, diff3, diff4, diffGlobal);		

			BeginTest(FT, "ShortTermTimer: Local/ Local Clock Synchronization");

			if ( (diff3 == (diff2/2 + tolerance)) || (diff3 == (diff2/2 - tolerance)) || (diff3 == diff2/2) )
				LogResult(PASSED, "" );
			else
				LogResult(FAILED, "Clock 3 at 240 Hz is not running half as fast as Clock 2 at 480 Hz" );
			
			BeginTest(FT, "ShortTermTimer: Local/ Local Clock Synchronization");

			if ( (diff4 == (diff3/4 + tolerance)) || (diff4 == (diff3/4 - tolerance)) || (diff4 == diff3/4) )
				LogResult(PASSED, "" );
			else
				LogResult(FAILED, "Clock 4 at 60 Hz is not running a fourth as fast as Clock 3 at 240 Hz" );
			
		}
	
		prevTime1		= time1;
		prevTime2		= time2;
		prevTime3		= time3;
		prevTime4		= time4;
		prevGlobalTime	= globalTime;
	}

	/****************************************************************************************************
	 *	Test 3:	Run fastest independent clock for a while. 
	 ****************************************************************************************************/
	printf("\n3. Run fastest clock for a while. Read times from left to right.\n");

	for ( i = 0; i < 20; i++ )
	{
		for ( j = 0; j < MAX_ITERATIONS; j++ )
		{
			ReadAudioClock( clock1, &time1 );
			printf("%3d. %5d\t", counter, time1);
			counter++;
		}
		printf("\n");
	}

	GetAudioClockRate( clock1, &clockRate1 );
	GetAudioClockRate( clock2, &clockRate2 );
	GetAudioClockRate( clock3, &clockRate3 );
	GetAudioClockRate( clock4, &clockRate4 );


	/****************************************************************************************************
	 *	Test 4:	Test GetAudioClockRate by reading the independent clocks' rate 
	 ****************************************************************************************************/
	LogComment("4. Test GetAudioClockRate() for the independent clocks.");

			
	BeginTest(FT, "GetAudioClockRate");

	if ( (clockRate1 - 44100.0 < 1) && (clockRate1 - 44100.0 > -1) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected 44100, but got %f from GetAudioClockRate(clock1)",clockRate1 );

	BeginTest(FT, "GetAudioClockRate");

	if ( (clockRate2 - 480.0 < 1) && (clockRate2 - 480.0 > -1) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected 480, but got %f from GetAudioClockRate(clock2)",clockRate2 );

	BeginTest(FT, "GetAudioClockRate");

	if ( (clockRate3 - 240.0 < 1) && (clockRate3 - 240.0 > -1) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected 240, but got %f from GetAudioClockRate(clock3)",clockRate3 );

	BeginTest(FT, "GetAudioClockRate");

	if ( (clockRate4 - 60.0 < 1) && (clockRate4 - 60.0 > -1) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected 60, but got %f from GetAudioClockRate(clock4)",clockRate4 );


	/****************************************************************************************************
	 *	Test 5:	Test AudioTimeLaterThan, AudioTimeLaterThanOrEqual, CompareAudioTimes
	 ****************************************************************************************************/

	ReadAudioClock( clock1, &time1 );
	SleepUntilTime( theCue, GetAudioTime() + 60 );
	ReadAudioClock( clock1, &time2 );
	
	BeginTest(FT, "AudioTimeLaterThan");

	if ( AudioTimeLaterThan(time2, time1) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "AudioTimeLaterThan(%d, %d) returned FALSE", time2, time1 );


	BeginTest(FT, "AudioTimeLaterThan");

	if ( AudioTimeLaterThan(time1, time2) )
		LogResult(FAILED, "AudioTimeLaterThan(%d, %d) returned TRUE", time1, time2 );
	else
		LogResult(PASSED, "");

	
	BeginTest(FT, "AudioTimeLaterThanOrEqual");

	if ( AudioTimeLaterThanOrEqual(time2, time1) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "AudioTimeLaterThanOrEqual(%d, %d) returned FALSE", time2, time1 );

	BeginTest(FT, "AudioTimeLaterThanOrEqual");

	if ( AudioTimeLaterThanOrEqual(time1, time2) )
		LogResult(FAILED, "AudioTimeLaterThanOrEqual(%d, %d) returned TRUE", time1, time2 );
	else
		LogResult(PASSED, "");

	
	BeginTest(FT, "AudioTimeLaterThanOrEqual");

	if ( AudioTimeLaterThanOrEqual(time1, time1) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "AudioTimeLaterThanOrEqual(%d, %d) returned FALSE\n", time1, time1 );

	
	BeginTest(FT, "CompareAudioTimes");

	if ( CompareAudioTimes(time2, time1) > 0 )
		LogResult(PASSED, "" );
	else
		LogResult(FAILED, "CompareAudioTimes(%d, %d) <= 0", time2, time1 );

	BeginTest(FT, "CompareAudioTimes");

	if ( CompareAudioTimes(time1, time2) < 0)
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "CompareAudioTimes(%d, %d) >= 0", time1, time2 );

	BeginTest(FT, "CompareAudioTimes");

	if ( CompareAudioTimes(time1, time1) == 0 )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "CompareAudioTimes(%d, %d) != 0", time1, time1 );

	
	DeleteCue(theCue);
	DeleteAudioClock(clock1);
	DeleteAudioClock(clock2);
	DeleteAudioClock(clock3);
	DeleteAudioClock(clock4);
	
	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT, theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}

