/* @(#) Tuning_FT.c 96/07/26 1.2 */
/*	
 *	File:				Tuning_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *		
 *	This file contains the functional tests for the tuning.  we create various tuning lists and play a
 *	sawtooth instrument using those tuning lists.
 *
 *	Test 1:	Play the sawtooth for MIDI pitches 60 - 71 using the default tuning.
 *
 *	Test 2:	Play the sawtooth for MIDI pitches 60 - 71 using the first four notes of a  random tuning
 *
 *	Test 3:	Play the squareWave (from a template) for MIDI pitches 60 - 71 using a  random tuning
 *
 *	Test 4:	Play the squareWave (from a template) after deleting its tuning
 *
 *	Uses the following calls:
 *
 *	OpenAudioFolio	 		CreateCue			LoadInstrument		
 *	CreateTuning			TuneInstrument		CloseAudioFolio
 *	ConnectInstrumentParts	StartInstrument		StopInstrument
 *	DeleteCue				DeleteTuning		UnloadInstrument		SleepUntilTime	
 *
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <stdio.h>
#include <string.h>

#include "TestMacros.h"
#include <tests/3do_scaf.h>

/**************************************Test Environment Globals****************/
/******************************************************************************/

int main (int argc, char ** argv)
{
	
	Err					theError;
	const	float32		pitches0[12]	= {123.0, 261.7, 423.5, 123.0, 261.7, 423.5, 123.0, 261.7, 423.5, 123.0, 261.7, 423.5 };
	const	float32		pitches1[12]	= {60.0, 120.5, 2000.5, 200.0, 10000.7, 600.5, 140.0, 320.7, 60.5, 500.0, 800.7, 15000.5 };
	Item 				theItem, theLineOut, theCue, tuning[6], theTempItem, theTemplate;
	uint8				i, j;



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);

	#define expected 0

	/****************************************************************************************************/
	/*	Load the sound instrument and the line_out instrument (which provides the sound).				*/
	/*	Create the tunings.									*/
		
	tuning[0]		= CreateTuning( pitches0, 12, 12, 60 );
	CHECKRESULT(FT,tuning[0],"CreateTuning(pitches1, 12, 12, 60)");
		
	tuning[1]		= CreateTuning( pitches1, 2, 2, 60 );
	CHECKRESULT(FT,tuning[1],"CreateTuning(pitches1, 2, 2, 60)");
		
	tuning[2]		= CreateTuning( pitches1, 12, 12, 60 );
	CHECKRESULT(FT,tuning[1],"CreateTuning(pitches1, 12, 12, 60)");
		
	theItem			= LoadInstrument("sawtooth.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,theItem,"LoadInstrument(sawtooth.dsp)",cleanup);

	theTemplate		= LoadInsTemplate("square.dsp", NULL);
	CHECKRESULT_SETUP(FT,theItem,"LoadInstrument(sawtooth.dsp)",cleanup);

	theError		= TuneInsTemplate( theTemplate, tuning[2] );		
	CHECKRESULT(FT,theError,"TuneInsTemplate()");

	theTempItem		= CreateInstrument(theTemplate, NULL);
	CHECKRESULT_SETUP(FT,theItem,"LoadInstrument(sawtooth.dsp)",cleanup);

	theLineOut		= LoadInstrument("line_out.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,theLineOut,"LoadInstrument(line_out.dsp)",cleanup);
	
	ConnectInstrumentParts(	theItem, "Output", 0, theLineOut, "Input", 0 );		/*	Connect the left channel	*/

	ConnectInstrumentParts(	theItem, "Output", 0, theLineOut, "Input", 1 );		/*	Connect the right channel - Mono Sample	*/

	/****************************************************************************************************/
	/****************************************************************************************************/
	/*	Play Sawtooth with Default Tuning																	*/
	/****************************************************************************************************/
	BeginTest(FT, "Tuning");

	for( i = 60; i < 72; i++ )
	{
		LogComment("(Default Tuning): Playing sawtooth Wave at MIDI Pitch %d", i );

		StartInstrumentVA( theItem, AF_TAG_PITCH, i, TAG_END );					/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		SleepUntilTime( theCue, GetAudioTime() + 1*240 );						/*	Enter a Wait state until sample is finished	*/

		/****************************************************************************************************/
		StopInstrument( theItem, NULL );										/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );
	}

	LogResult(InteractiveVerify("You should have heard 12 notes at the given pitch. Did Test Pass?"), "");

	/****************************************************************************************************/
	/*	Play Sawtooth with Assigned 12-note Tuning														*/
	/****************************************************************************************************/
	printf("The tuning is " );
	for ( j = 0; j < 12; j++ )
		printf("Note %2d: %f\t", j, pitches0[j] );
	printf("\n\n" );

	theError	= TuneInstrument( theItem, tuning[0] );		
	CHECKRESULT(FT,theError,"TuneInstrument()");

	BeginTest(FT, "Tuning");

	for( i = 60; i < 72; i++ )
	{
		LogComment("(Assigned Tuning): Playing sawtooth Wave at MIDI Pitch %d:\tFreq. %f", i, pitches0[i-60] );

		StartInstrumentVA( theItem, AF_TAG_PITCH, i, TAG_END );			/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		SleepUntilTime( theCue, GetAudioTime() + 1*240 );				/*	Enter a Wait state until sample is finished	*/

		/****************************************************************************************************/
		StopInstrument( theItem, NULL );								/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );
	}

	LogResult(InteractiveVerify("You should have heard 12 notes at the assigned pitch. Did Test Pass?"), "");

	/****************************************************************************************************/
	/*	Play sawtooth wave with the first two notes of the Assigned 12-note Tuning						*/
	/****************************************************************************************************/
	
	printf("The tuning is " );
	for ( j = 0; j < 12; j++ )
		printf("Note %2d: %f\t", j, pitches1[j] );
	printf("\n\n" );

	theError	= TuneInstrument( theItem, tuning[1] );		
	CHECKRESULT(FT,theError,"TuneInstrument()");

	BeginTest(FT, "Tuning");

	for( i = 60; i < 72; i++ )
	{
		LogComment("(1st 2 notes of Assigned Tuning): Playing sawtooth Wave at MIDI Pitch %d", i );

		StartInstrumentVA( theItem, AF_TAG_PITCH, i, TAG_END );				/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		SleepUntilTime( theCue, GetAudioTime() + 1*240 );					/*	Enter a Wait state until sample is finished	*/

		/****************************************************************************************************/
		StopInstrument( theItem, NULL );									/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );
	}

	LogResult(InteractiveVerify("You should have heard the first two notes of the assigned tuning being played at over and over at consecutively higher octaves. Did Test Pass?"), "");

	/*	Disconnect the sawtooth instrument from the output and connect the square instrument to the output	*/
	DisconnectInstrumentParts(	theLineOut, "Input", 0 );		
	DisconnectInstrumentParts(	theLineOut, "Input", 1 );	
	
	ConnectInstrumentParts(	theTempItem, "Output", 0, theLineOut, "Input", 0 );		/*	Connect the left channel	*/
	ConnectInstrumentParts(	theTempItem, "Output", 0, theLineOut, "Input", 1 );		/*	Connect the right channel - Mono Sample	*/

	
	/****************************************************************************************************/
	/*	Play square Wave with the  Assigned Random Tuning												*/
	/****************************************************************************************************/
	
	printf("The tuning is " );
	for ( j = 0; j < 12; j++ )
		printf("Note %2d: %f\t", j, pitches1[j] );
	printf("\n\n" );

	BeginTest(FT, "Tuning");

	for( i = 60; i < 72; i++ )
	{
		LogComment("(Assigned Random Tuning): Playing square Wave at MIDI Pitch %d:\tFreq. %f", i, pitches1[i-60] );

		StartInstrumentVA( theTempItem, AF_TAG_PITCH, i, TAG_END );					/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		SleepUntilTime( theCue, GetAudioTime() + 1*240 );							/*	Enter a Wait state until sample is finished	*/

		/****************************************************************************************************/
		StopInstrument( theTempItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );
	}

	LogResult(InteractiveVerify("You should have heard 12 notes at the assigned pitch. Did Test Pass?"), "");

	for ( i = 0; i < 3; i++ )
	{
		theError	= DeleteTuning( tuning[i] );
		CHECKRESULT(FT,theError,"DeleteTuning()");
	}
		
	/****************************************************************************************************/
	/*	Play square Wave after deleting tuning															*/
	/****************************************************************************************************/
	
	BeginTest(FT, "Tuning");

	for( i = 60; i < 85; i++ )
	{
		LogComment("(Default Tuning): Playing square Wave after deleting its Tuning" );

		StartInstrumentVA( theTempItem, AF_TAG_PITCH, i, TAG_END );			/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		SleepUntilTime( theCue, GetAudioTime() + 1*240 );					/*	Enter a Wait state until sample is finished	*/

		/****************************************************************************************************/
		StopInstrument( theTempItem, NULL );								/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );
	}

	LogResult(InteractiveVerify("You should have heard 15 notes at the default pitch. Did Test Pass?"), "");

	DeleteCue( theCue );
	UnloadInstrument(theItem);
	UnloadInstrument(theLineOut);
	UnloadInsTemplate(theTemplate);
	/****************************************************************************************************/

	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT,theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}

