/* @(#) Patch_FT.c 96/07/26 1.3 */
/*	
 *	File:				Patch_FT.c
 *	Author:				
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *		
 *	This file contains the functional tests for Audio Patch Folio.  The code for this test was taken
 *	from the patch example program, windpatch.c.  This test program has added error checks which allow
 *	the tester to monitor the program as it builds and plays its patch.
 *
 *
 *	Uses the following calls:
 *
 *	OpenAudioFolio	 		CreateCue			CreateInstrument			SleepUntilTime
 *	StartInstrument			StopInstrument		ConnectInstrumentParts		LoadInsTemplate
 *	DeleteCue				DefinePatchPort		CreatePatchCmdBuilder		AddTemplateToPatch
 *	SetPatchConstant		ConnectPatchPorts	DeletePatchCmdBuilder		UnloadInsTemplate
 *	CreatePatchTemplateVA	GetPatchCmdList
 *
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */

#include <audio/audio.h>
#include <audio/patch.h>
#include <kernel/operror.h>
#include <kernel/task.h>
#include <stdio.h>

#include <TestMacros.h>
#include <tests/3do_scaf.h>

/* -------------------- Local functions */

static Err DoWindSound (void);
static Item CreateWindPatch (void);


/**************************************Test Environment Globals****************/
/******************************************************************************/

/***********************************************************************/
int main (int argc, char *argv[])
{
    Err errcode;
	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/
    errcode = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,errcode,"OpenAudioFolio",clean);

    DoWindSound();

	BeginTest(FT, "AudioPatchFolio");
	LogResult(InteractiveVerify("You should have heard a wind sound. Did Test Pass?"), "");


clean:

	errcode	= CloseAudioFolio();
	CHECKRESULT(FT,errcode,"CloseAudioFolio");

	/******************************************************************************/
	TermScaffold();

    return 0;
}


/***********************************************************************/
static Err DoWindSound (void)
{
    Item	windPatch, theCue;
    Item	leftWindIns, rightWindIns;
    Item	outputIns;
    Err		errcode;


	theCue	= CreateCue( NULL );
	CHECKRESULT(FT,theCue,"CreateCue");


    windPatch		= CreateWindPatch();
	CHECKRESULT(FT,windPatch,"CreateWindPatch");
	
    leftWindIns 	= CreateInstrument (windPatch, NULL);
	CHECKRESULT_SETUP(FT,leftWindIns,"CreateInstrument(windPatch)", clean);

    rightWindIns	= CreateInstrument (windPatch, NULL);
	CHECKRESULT_SETUP(FT,rightWindIns,"CreateInstrument(windPatch)", clean);
	
    outputIns 		= LoadInstrument ("line_out.dsp", 0, 100);
	CHECKRESULT_SETUP(FT,outputIns,"LoadInstrument(line_out.dsp)", clean);

    /*
         Connect 2 different wind instruments from the same template on
         different channels, because we can and it sounds spookier than just
         a single instrument connected up monophonically to both outputs.
     */
    errcode = ConnectInstrumentParts (	leftWindIns, "Output",  0, outputIns, "Input", AF_PART_LEFT);
	CHECKRESULT(FT,errcode,"ConnectInstrumentParts()");

    errcode = ConnectInstrumentParts (rightWindIns, "Output", 0, outputIns, "Input", AF_PART_RIGHT);
	CHECKRESULT(FT,errcode,"ConnectInstrumentParts()");

    errcode = StartInstrument (outputIns, NULL);
	CHECKRESULT(FT,errcode,"StartInstrument(line_out.dsp)");

    errcode = StartInstrument (leftWindIns, NULL);
	CHECKRESULT(FT,errcode,"StartInstrument(windpatch)");

    errcode = StartInstrument (rightWindIns, NULL);
	CHECKRESULT(FT,errcode,"StartInstrument(windpatch)");

        /* Play for about 10 seconds */
    SleepUntilTime( theCue, GetAudioTime() + 10*240 );

    errcode = StopInstrument (outputIns, NULL);
	CHECKRESULT(FT,errcode,"StopInstrument(line_out.dsp)");
	
    UnloadInstrument (outputIns);
	

clean:
    errcode = DeletePatchTemplate (windPatch);
	CHECKRESULT(FT,errcode,"DeletePatchTemplate(windpatch)");

    DeleteCue (theCue);
    return errcode;
}


/***********************************************************************/
/*
    This function creates a Patch Template out of constituent templates and
    returns the result.

    The patch basically consists of a white noise generator played through
    a low pass filter. The cutoff frequency and resonance of the filter
    are controlled by two independent random signals from red noise generators
    running at very low frequency.

    Note: this function has most of its error checking removed for brevity.
    In this case, the patch compiler will reject the PatchCmd list if an
    error was encountered while constructing it.
*/
static Item CreateWindPatch (void)
{
    const Item 		windNoiseTmp     = LoadInsTemplate ("noise.dsp", NULL);
    const Item 		windFilterTmp    = LoadInsTemplate ("svfilter.dsp", NULL);
    const Item 		windRedNoiseTmp  = LoadInsTemplate ("rednoise.dsp", NULL);
    const Item 		windTimesPlusTmp = LoadInsTemplate ("timesplus.dsp", NULL);
	Err				errcode;
    PatchCmdBuilder *pb = NULL;
    Item			result	= -1;


 	CHECKRESULT_SETUP(FT,windNoiseTmp,"LoadInsTemplate(noise.dsp)",clean);
	CHECKRESULT_SETUP(FT,windFilterTmp,"LoadInsTemplate(svfilter.dsp)",clean);
	CHECKRESULT_SETUP(FT,windRedNoiseTmp,"LoadInsTemplate(rednoise.dsp)",clean);
	CHECKRESULT_SETUP(FT,windTimesPlusTmp,"LoadInsTemplate(timesplus.dsp)",clean);

 
        /* create PatchCmdBuilder */
    result = CreatePatchCmdBuilder (&pb);
	CHECKRESULT_SETUP(FT,result,"CreatePatchCmdBuilder()",clean);

        /* Define knobs and ports */
    errcode	= DefinePatchPort (pb, "Output", 1, AF_PORT_TYPE_OUTPUT, AF_SIGNAL_TYPE_GENERIC_SIGNED);
	CHECKRESULT_SETUP(FT,errcode,"DefinePatchPort()",clean);

        /* Add constituent templates to patch */
    errcode	= AddTemplateToPatch (pb, "noise", windNoiseTmp);
	CHECKRESULT(FT,errcode,"AddTemplateToPatch(noise.dsp)");

    errcode	= AddTemplateToPatch (pb, "modsignal1", windRedNoiseTmp);
	CHECKRESULT(FT,errcode,"AddTemplateToPatch(rednoise.dsp)");

    errcode	= AddTemplateToPatch (pb, "modsignal2", windRedNoiseTmp);
	CHECKRESULT(FT,errcode,"AddTemplateToPatch(rednoise.dsp)");

    errcode	= AddTemplateToPatch (pb, "modroute1", windTimesPlusTmp);
	CHECKRESULT(FT,errcode,"AddTemplateToPatch(timesplus.dsp)");

    errcode	= AddTemplateToPatch (pb, "modroute2", windTimesPlusTmp);
	CHECKRESULT(FT,errcode,"AddTemplateToPatch(timesplus.dsp)");

    errcode	= AddTemplateToPatch (pb, "filter", windFilterTmp);
	CHECKRESULT(FT,errcode,"AddTemplateToPatch(svfilter.dsp)");

        /* Connect signal path: noise -> filter -> output */
    errcode	= SetPatchConstant (pb, "noise", "Amplitude", 0, 0.3);
	CHECKRESULT(FT,errcode,"SetPatchConstant(noise.dsp, Amplitude)");

    errcode	= ConnectPatchPorts (pb, "noise",  "Output", 0, "filter", "Input", 0);
	CHECKRESULT(FT,errcode,"ConnectPatchPorts()");

    errcode	= ConnectPatchPorts (pb, "filter", "Output", 0, NULL,     "Output", 0);
	CHECKRESULT(FT,errcode,"ConnectPatchPorts()");

        /* Connect one rednoise modulation signal to filter cutoff frequency */
    errcode	= ConnectPatchPorts (pb, "modsignal1", "Output", 0, "modroute1", "InputA", 0);
	CHECKRESULT(FT,errcode,"ConnectPatchPorts()");

    errcode	= ConnectPatchPorts (pb, "modroute1",  "Output", 0, "filter", "Frequency", 0);
	CHECKRESULT(FT,errcode,"ConnectPatchPorts()");

	
    errcode	= SetPatchConstant (pb, "modsignal1", "Frequency", 0, 0.7);
	CHECKRESULT(FT,errcode,"SetPatchConstant(rednoise.dsp, Frequency)");

    errcode	= SetPatchConstant (pb, "modroute1", "InputB", 0, 0.0427);
	CHECKRESULT(FT,errcode,"SetPatchConstant(timesplus.dsp, InputB)");

    errcode	= SetPatchConstant (pb, "modroute1", "InputC", 0, 0.0854);
	CHECKRESULT(FT,errcode,"SetPatchConstant(timesplus.dsp, InputC)");


        /* Connect the other rednoise modulation signal to filter resonance */
    errcode	= ConnectPatchPorts (pb, "modsignal2", "Output", 0, "modroute2", "InputA", 0);
	CHECKRESULT(FT,errcode,"ConnectPatchPorts()");

    errcode	= ConnectPatchPorts (pb, "modroute2", "Output", 0, "filter", "Resonance", 0);
	CHECKRESULT(FT,errcode,"ConnectPatchPorts()");

   
    errcode	= SetPatchConstant (pb, "modsignal2", "Frequency", 0, 0.7);
	CHECKRESULT(FT,errcode,"SetPatchConstant(rednoise.dsp, Frequency)");

    errcode	= SetPatchConstant (pb, "modroute2", "InputB", 0, 0.0854);
	CHECKRESULT(FT,errcode,"SetPatchConstant(timesplus.dsp, InputB)");

    errcode	= SetPatchConstant (pb, "modroute2", "InputC", 0, 0.0916);
	CHECKRESULT(FT,errcode,"SetPatchConstant(timesplus.dsp, InputC)");


        /* create Patch Template from PatchCmd list returned from
           PatchCmdBuilder */
    result = CreatePatchTemplateVA (GetPatchCmdList (pb), TAG_ITEM_NAME, "wind.patch", TAG_END);
	CHECKRESULT(FT,result,"CreatePatchTemplateVA()");


clean:
        /* we don't need this stuff after we've built the patch */
    DeletePatchCmdBuilder (pb);

    errcode	= UnloadInsTemplate (windNoiseTmp);
	CHECKRESULT(FT,errcode,"UnloadInsTemplate()");

    errcode	= UnloadInsTemplate (windFilterTmp);
	CHECKRESULT(FT,errcode,"UnloadInsTemplate()");

    errcode	= UnloadInsTemplate (windRedNoiseTmp);
	CHECKRESULT(FT,errcode,"UnloadInsTemplate()");

    errcode	= UnloadInsTemplate (windTimesPlusTmp);
	CHECKRESULT(FT,errcode,"UnloadInsTemplate()");

    return result;
}
