/* @(#) Filter1o1z_FT.c 96/07/29 1.3 */
/*	
 *	File:				Filter1o1z_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests for filter1o1z.dsp.  This program draws the frequency response of
 *	svfilter with a given resonance to a sinewave centered around middle C.  The frequency range tested
 *	goes from 130.5 to 523 Hz.  The lowpass (output) frequency response is drawn on the television screen.
 *	The user can exit by pressing the X button of the control pad.
 *
 *	The user can set test parameters on the command line.  There are nine parameters the user can set.
 *
 *	-0		filter1o1z's A0 coefficient. default is .5
 *	-1		filter1o1z's A1 coefficient. default is .5
 *	-l		the low end of the frequency range. default is 130.0
 *	-h		the high end of the frequency range. default is 523.0
 *	-d		the time (in ticks) to wait before the test program reads the response to a certain frequency 
 *			(the longer, the better - 60 ticks should be the maximum value needed. Any longer will make the
 *			test program way too slow). default is 60.0
 *	-p		print the graphed values into the debugger window. Default is 0 - "do not print"
 *	-v		vertical graphing scaling factor. default is .9
 *
 *	The equation for the frequency response of the filter is:	y(n) = A0*x(n) + A1*x(n-1) 
 *
 *	This test program works by creating a loop ranging from the lowest frequency to the highest frequency.
 *	Each time in the loop, the sinewave sample is played at that frequency into the filter's input.  The
 *	output of the filter is connected the an envfollower which follows the level of the filter's output.
 *	The test program waits a specified amout of time before reading the envfollower's output into an array.
 *	(We must allow the envfollower time to rise to the final output value.  The test program takes this array
 *	of values and plots them on the TV screen.
 *
 *	In order to build this program, you must -laudiodemo to the LIBS line of Integrator_FT.make, and you must
 *	place libaudiodemo.a in the 3DO:Interfaces:Libs:Default folder.
 *
 *	Audio Functions called:
 *
 *		CloseAudioFolio		CreateCue			CreateAttachment		LoadInstrument
 *		OpenAudioFolio		CreateInstrument	CreateKnob				CreateProbe
 *		SetKnobPart			StartInstrument		StopInstrument			LoadSample
 *		SleepUntilTime		DeleteAttachment	DeleteCue				UnloadInstrument
 *		DeleteProbe			UnloadSample		ConnectInstrumentParts
 *
 *	Dsp Instruments used:
 *
 *	sampler_16_f1.dsp		filter1o1z.dsp		envfollower.dsp
 *
 *	Audio files Used (These files must be present in the current directory)
 *
 *	sinewave.aiff				
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <audio/parse_aiff.h>
#include <graphics/graphics.h>
#include <kernel/operror.h>
#include <misc/event.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include <audiodemo/portable_graphics.h>
#include <TestMacros.h>
#include "TestUtilities.h"
#include <tests/3do_scaf.h>

/**************************************Test Environment Globals****************/
/******************************************************************************/

#define	MAX_NUM_OF_SAMPLES			2000

void WaitButton( uint32 button );


float32		 				outSampleValues[MAX_NUM_OF_SAMPLES],
							vertScale	= .9;
int32						ivResult;
int32						diffFreq, lowFreq	= 130, highFreq	= 523;
PortableGraphicsContext		*pgcon;

int main (int argc, char ** argv)
{
	
	char		c;
	Err			theError;
	float32		theOutputValue, scaledValue, flipValue,
				a0Arg	= .5, a1Arg	= .5, ampArg	= 1.0, freqCounter;
	int			doPrint	= 0, newArgc;
	int32		i, duration	= 60;
	Item 		filter, a0, a1, sineWave,
				out_envfollower,
				outTestProbe,
				theCue, sleepCue, theAttachment, theSample;
				


	newArgc	= InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	BeginTest(FT, "Filter1o1z_FT");
	/******************************************************************************/

	while( --newArgc > 0  && (*++argv)[0] == '-' )
	{
		c	= *++argv[0];
		switch( c )
		{
			case	'0'	:									/*	Get A0 coefficient from the command line	*/
							if ( --argc > 0 )
							{
								a0Arg	= atof32( *++argv );
								if ( (a0Arg < 0) || (a0Arg > 1.0) )  
								{
									printf( "\nFilter1o1z_FT: %f is out of range.\tA0 coefficient must be between 0.0 and 1.0\n", a0Arg );
									exit(1);
								}
							}
							break;
			case	'1'	:									/*	Get A1 coefficient from the command line	*/
							if ( --argc > 0 )
							{
								a1Arg	= atof32( *++argv );
								if ( (a1Arg < 0) || (a1Arg > 1.0) )  
								{
									printf( "\nFilter1o1z_FT: %f is out of range.\tA1 coefficient must be between 0.0 and 1.0\n", a1Arg );
									exit(1);
								}
							}
							break;
			case	'l'	:									/*	Get low frequency range  from the command line	*/
							if ( --argc > 0 )
							{
								lowFreq	= atol( *++argv );
								if ( (lowFreq < 130) || (lowFreq > 523) ) 
								{
									printf( "\nFilter1o1z_FT: %d is out of range.\tLow Frequency Range must be between 130 and 523\n", lowFreq );
									exit(1);
								}
							}
							break;
			case	'h'	:									/*	Get high frequency range  from the command line	*/
							if ( --argc > 0 )
							{
								highFreq	= atol( *++argv );
								if ( (highFreq < 130) || (highFreq > 523) ) 
								{
									printf( "\nFilter1o1z_FT: %d is out of range.\tHigh Frequency Range must be between 130 and 523\n", highFreq );
									exit(1);
								}
							}
							break;
			case	'd'	:									/*	Get duration (the time to wait before reading probes)  from the command line	*/
							if ( --argc > 0 )
							{
								duration	= atol( *++argv );
								if ( (duration < 0) || (duration > 60) )
								{
									printf( "\nFilter1o1z_FT: %d is out of range.\tWait time must be between 0 and 60\n", duration );
									exit(1);
								}
							}
							break;
			case	'p'	:									/*	Get the printing option from the command line	*/
							if ( --argc > 0 )
							{
								doPrint	= atoi( *++argv );
								if ( doPrint < 0 ) 
								{
									printf( "\nFilter1o1z_FT: %d is out of range.\tPrint option must be either 1 to print or 0 disable printing\n", ampArg );
									exit(1);
								}
							}
							break;
			case	'v'	:									/*	Get vertical scaling factor from the command line	*/
							if ( --argc > 0 )
							{
								vertScale	= atof32( *++argv );
								if ( (vertScale < 0) || (vertScale > 1) ) 
								{
									printf( "\nFilter1o1z_FT: %f is out of range.\tVertical scale must be between 0.0 and 1.0\n", ampArg );
									exit(1);
								}
							}
							break;
			default		:	
							printf( "\nFilter1o1z_FT: Illegal option -%c\n", c );
							printf( "\nUsage:\tFilter1o1z_FT\t-a (amplitude) -f (critical frequency) -r (resonance)\n\t\t\t\t\t-l (low frequency) -h(high frequency) -d(wait duration)\n" );
							exit(1);
							break;
		}
	}
	
	if ( highFreq < lowFreq )
	{
		printf("Error: High part of the frequency range is lower than low part of frequency range.\nExiting program...\n");
		exit(1);
	}
		



	/* Initialize the EventBroker. */
	theError = InitEventUtility(1, 0, LC_FocusListener);
	CHECKRESULT_SETUP(FT,theError,"InitEventUtility",cleanup);

	/* Set up double buffered display. */
	theError = pgCreateDisplay( &pgcon, 2 );
	CHECKRESULT_SETUP(FT,theError,"pgCreateDisplay",cleanup);

	/* Draw Background so we can see pixel colors. */
	pgClearDisplay( pgcon, 0.0, 0.0, 0.0 );


	theError =	 OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);

	sleepCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,sleepCue,"CreateCue",cleanup);


	/****************************************************************************************************
	 *	Load the svfilter instrument and create the amplitude and offset knobs
	 ****************************************************************************************************/
		
	filter		= LoadInstrument("filter_1o1z.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,filter,"LoadInstrument(filter_1o1z.dsp)",cleanup);

	a0			= CreateKnob( filter, "A0", NULL );
	CHECKRESULT_SETUP(FT,a0,"CreateKnob",cleanup);
	
	a1			= CreateKnob( filter, "A1", NULL );
	CHECKRESULT_SETUP(FT,a1,"CreateKnob",cleanup);
	

	/****************************************************************************************************
	 *	Load the sine wave instrument
	 ****************************************************************************************************/
		
	sineWave		= LoadInstrument("sampler_16_v1.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,sineWave,"LoadInstrument(sampler_16_v1.dsp)",cleanup);
		
	theSample		= LoadSample("sinewave.aiff");
	CHECKRESULT_SETUP(FT,theSample,"LoadSample(sinewave.aiff)",cleanup);
	
	theAttachment	= CreateAttachment( sineWave, theSample, NULL );
	MonitorAttachment( theAttachment, theCue, CUE_AT_END );


	/****************************************************************************************************
	 *	Create the envelope follower instruments
	 ****************************************************************************************************/

	out_envfollower		= LoadInstrument("envfollower.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,out_envfollower,"LoadInstrument(envfollower.dsp)",cleanup);
		
		
	/****************************************************************************************************
	 *	Create the testprobes for the envelope follower instruments
	 ****************************************************************************************************/

	outTestProbe	= CreateProbe( out_envfollower, "Output", NULL );
	CHECKRESULT_SETUP(FT,outTestProbe,"CreateProbe",cleanup);
	
	
	/****************************************************************************************************
	 *	Connect the output of the sampler_16_f1.dsp to the input of the svfilter.dsp.
	 *	Connect the Output of the filter1o1z.dsp to input of out_envfollwer..dsp
	 ****************************************************************************************************/
		
	ConnectInstrumentParts(	sineWave, "Output", 0, filter, "Input", 0 		);
	ConnectInstrumentParts(	filter, "Output", 0, out_envfollower, "Input", 0 		);

							
	SetKnob( a0, a0Arg);
	SetKnob( a1, a1Arg);

	LogComment("\tfilter_1o1z.dsp Setup: (Amp. = %f; A0. = %f; A1. = %f)", ampArg, a0Arg, a1Arg);		
	LogComment("\tFrequency range is from %d Hz to %d Hz.  Wait time is %d ticks.  Vertical scaling factor is %f", lowFreq, highFreq, duration, vertScale);		
		
	StartInstrument( filter, NULL );
	StartInstrument( out_envfollower, NULL );

	printf("Gathering data.  Please wait ...\n\n");		
	
	diffFreq	= highFreq - lowFreq;
	freqCounter	= lowFreq;
	
	for ( i = 0; i < diffFreq; i++ )
	{
		StartInstrumentVA( sineWave, AF_TAG_FREQUENCY_FP, ConvertFP_TagData(freqCounter), TAG_END);
		SleepUntilTime( sleepCue, GetAudioTime() + duration );

		ReadProbe( outTestProbe, &theOutputValue );
		outSampleValues[i]	= theOutputValue;
		
		freqCounter	+= 1.0;
	}	

	StopInstrument( sineWave, NULL );
	StopInstrument( filter, NULL );
	StopInstrument( out_envfollower, NULL );


	/* Draw white rectangle. */
	pgSetColor( pgcon, 1.0, 1.0, 1.0 );

	/*	Draw the lowpass scaled output to the screen. Scale the x values so that they will map to 0.0 - 1.0
		screen.  Invert the y values so that graph will look right side up.
	*/
	pgMoveTo( pgcon, 0, 1.0 );
	for ( i = 0; i < diffFreq; i++ )
	{
		scaledValue	= (float32)i/(float32)diffFreq;
		flipValue	= 1.0 - outSampleValues[i]*vertScale;
		pgDrawTo( pgcon, scaledValue,flipValue );
	}
	
	/* Draw in pink. */
	pgSetColor( pgcon, .7, .5, .8 );
	pgMoveTo( pgcon, .4, .9 );
	
	pgDrawText( pgcon, "Output" );

	if ( doPrint != 0 )
	{
		for ( i = 0; i < diffFreq; i++ )
			printf("%3d Hz. out  = %f\n",	i+lowFreq, outSampleValues[i] );
	}

	printf("\nPress the X button to exit program\n\n");
	
	WaitButton(ControlX);
	
	/* Clear the screen before exiting */
	pgClearDisplay( pgcon, 0.0, 0.0, 0.0 );

	DeleteAttachment( theAttachment );
	UnloadSample( theSample );
	UnloadInstrument( filter );
	UnloadInstrument( sineWave );
	UnloadInstrument( out_envfollower );

	DeleteCue( theCue );
	DeleteCue( sleepCue );

	DeleteProbe( outTestProbe );
	
	KillEventUtility();

	pgDeleteDisplay( &pgcon );
	
	BeginTest(FT, "filter1o1z.dsp");
	LogResult(InteractiveVerify("Did the frequency response plot look correct. Did Test Pass?"), "");

	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT, theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}


void WaitButton( uint32 button )
{
	ControlPadEventData cped;

	do
	{
		GetControlPad (1, TRUE, &cped);   /* Wait for button press. */
	}
	while( (cped.cped_ButtonBits & button) == 0 );
}
