/* @(#) Integrator_FT.c 96/07/29 1.3 */
/*	
 *	File:				Integrator_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests for integrator.dsp.  This program sends a squareWave produced by a 
 *	square.dsp instrument through an integrator.dsp instrument.  This program plots the output of the integrator
 *	on the televison screen.  Also the user hears what the output sounds like.  The user can set five parameters
 *	on the command line.  The user can set the amplitude of the squareWave (-a), the squareWave frequency (-f),
 *	the integrator amplitude (-h), the integrator's offset (-o), the vertical scaling factor of the graph(-v),
 *	and the number of samples taken of the sound (-n).  Each parameter has a default value.  At the end of the test
 *	the user is prompted as to whether the test has passed.  If the test passed, the user presses Shift-A on the
 *	control pad, if the test has failed, press Shift-X on the control pad.
 *
 *	In order to build this program, you must -laudiodemo to the LIBS line of Integrator_FT.make, and you must
 *	place libaudiodemo.a in the 3DO:Interfaces:Libs:Default folder.
 *
 *	Audio Functions called:
 *
 *		CloseAudioFolio		CreateCue			LoadInstrument			
 *		OpenAudioFolio		CreateKnob			CreateProbe					ReadProbe
 *		SetKnob				StartInstrument		StopInstrument
 *		SleepUntilTime		DeleteProbe			DeleteCue					UnloadInstrument
 *
 *	Dsp Instruments used:
 *
 *	integrator.dsp		line_out.dsp		square.dsp
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <stdio.h>

#include <audiodemo/portable_graphics.h>
#include <TestMacros.h>
#include <TestUtilities.h>
#include <tests/3do_scaf.h>

/**************************************Test Environment Globals****************/
/******************************************************************************/

#define	MAX_NUM_OF_SAMPLES		100




PortableGraphicsContext		*pgcon;

int main (int argc, char ** argv)
{
	char		c;
	Err			theError;
	float32		intOffsetArg	=.5,
				intAmplitudeArg	= .5,
				sqAmplitudeArg	= .1,
				sqFrequencyArg	= 2000.0,
				vertScale		= .9,
				scaledValue, flipValue, theOutputValue, outSampleValues[4000];
	int			newArgc;
	int32		i,
				numofSamples	= MAX_NUM_OF_SAMPLES;
	Item 		integrator, intOffset, intAmplitude, squareWave, sqAmplitude, sqFrequency,
				outTestProbe, lineOut, theCue;


	newArgc	= InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	while( --newArgc > 0  && (*++argv)[0] == '-' )
	{
		c	= *++argv[0];
		switch( c )
		{
			case	'a'	:									/*	Get squareWave amplitude from the command line	*/
							if ( --argc > 0 )
							{
								sqAmplitudeArg	= atof32( *++argv );
								if ( (sqAmplitudeArg < 0) || (sqAmplitudeArg > 1) ) 
								{
									printf( "\nIntegrator_FT: %f is out of range.\tSquareWave Amplitude must be between 0 and 1.0\n", sqAmplitudeArg );
									exit(1);
								}
							}
							break;
			case	'f'	:									/*	Get SquareWave frequency from the command line	*/
							if ( --argc > 0 )
							{
								sqFrequencyArg	= atof32( *++argv );
								if ( (sqFrequencyArg < 0) || (sqFrequencyArg > 22050) )  
								{
									printf( "\nIntegrator_FT: %f is out of range.\tSquareWave frequency must be between 0.0 and 22050.0\n", sqFrequencyArg );
									exit(1);
								}
							}
							break;
			case	'h'	:									/*	Get Integrator amplitude from the command line	*/
							if ( --argc > 0 )
							{
								intAmplitudeArg	= atof32( *++argv );
								if ( (intAmplitudeArg < 0) || (intAmplitudeArg > 1) ) 
								{
									printf( "\nIntegrator_FT: %f is out of range.\tIntegrator Amplitude must be between 0 and 1.0\n", intAmplitudeArg );
									exit(1);
								}
							}
							break;
			case	'n'	:									/*	Get the number of samples for output  from the command line	*/
							if ( --argc > 0 )
							{
								numofSamples	= atol( *++argv );
								if ( (numofSamples < 0) ||  (numofSamples > 1000) )
								{
									printf( "\nIntegrator_FT: %d is out of range.\tThe number of samples must be between 0 and 1000\n", numofSamples );
									exit(1);
								}
							}
							break;
			case	'o'	:									/*	Get the integrator offset from the command line	*/
							if ( --argc > 0 )
							{
								intOffsetArg	= atof32( *++argv );
								if ( (intOffsetArg < 0) || (intOffsetArg > 1.0) )  
								{
									printf( "\nIntegrator_FT: %f is out of range.\tIntegrator offset must be between 0.0 and 1.0\n", intOffsetArg );
									exit(1);
								}
							}
							break;
			case	'v'	:									/*	Get vertical scaling factor from the command line	*/
							if ( --argc > 0 )
							{
								vertScale	= atof32( *++argv );
								if ( (vertScale < 0) || (vertScale > 1) ) 
								{
									printf( "\nIntegrator_FT: %f is out of range.\tVertical scale must be between 0.0 and 1.0\n", vertScale );
									exit(1);
								}
							}
							break;
			default		:	
							printf( "\nIntegrator_FT: Illegal option -%c\n", c );
							printf( "\nUsage:\tIntegrator_FT\t-a (squareWave amplitude) -f (squareWave frequency) -h (integrator amplitude) -o(integrator offset)\n" );
							exit(1);
							break;
		}
	}

	/* Set up double buffered display. */
	theError = pgCreateDisplay( &pgcon, 2 );
	CHECKRESULT_SETUP(FT,theError,"pgCreateDisplay",cleanup);

	/* Draw Background so we can see pixel colors. */
	pgClearDisplay( pgcon, 0.0, 0.0, 0.0 );


	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT, theCue,"CreateCue", cleanup);

	/****************************************************************************************************
	 *	Load the integrator instrument and create the amplitude and offset knobs
	 ****************************************************************************************************/
		
	integrator		= LoadInstrument("integrator.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,integrator,"LoadInstrument(integrator.dsp)",cleanup);
		
	intAmplitude	= CreateKnob( integrator, "Amplitude", NULL );
	CHECKRESULT_SETUP(FT,intAmplitude,"CreateKnob",cleanup);
	
	intOffset		= CreateKnob( integrator, "Offset", NULL );
	CHECKRESULT_SETUP(FT,intOffset,"CreateKnob",cleanup);
	

	/****************************************************************************************************
	 *	Create a test probe for the integrator
	 ****************************************************************************************************/
	outTestProbe	= CreateProbe( integrator, "Output", NULL );
	CHECKRESULT_SETUP(FT,outTestProbe,"CreateProbe",cleanup);

	
	/****************************************************************************************************
	 *	Load the square wave instrument and create the amplitude and frequency knobs
	 ****************************************************************************************************/
		
	squareWave		= LoadInstrument("square.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,squareWave,"LoadInstrument(square.dsp)",cleanup);
		
	sqAmplitude		= CreateKnob( squareWave, "Amplitude", NULL );
	CHECKRESULT_SETUP(FT,sqAmplitude,"CreateKnob",cleanup);
	
	sqFrequency		= CreateKnob( squareWave, "Frequency", NULL );
	CHECKRESULT_SETUP(FT,sqFrequency,"CreateKnob",cleanup);

	/****************************************************************************************************
	 *	Load the line_out instrument
	 ****************************************************************************************************/
		
	lineOut		= LoadInstrument("line_out.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,lineOut,"LoadInstrument(line_out.dsp)",cleanup);
		

	/****************************************************************************************************
	 *	Connect the output of the square.dsp to the input of the integrator.
	 *	Connect the output of the integrator.dsp to the right and left input of the line_out.dsp
	 ****************************************************************************************************/
		
	ConnectInstrumentParts(	squareWave, "Output", 0, integrator, "Input", 0 	);
	ConnectInstrumentParts(	integrator, "Output", 0, lineOut, "Input", 0 		);
	ConnectInstrumentParts(	integrator, "Output", 0, lineOut, "Input", 1 		);
							

	/****************************************************************************************************
	 *	Set the knobs
	 ****************************************************************************************************/
		
		
	SetKnob( sqAmplitude, sqAmplitudeArg);
	SetKnob( sqFrequency, sqFrequencyArg);
	SetKnob( intAmplitude, intAmplitudeArg);
	SetKnob( intOffset, intOffsetArg);

	
	StartInstrument( lineOut, NULL );
	StartInstrument( integrator, NULL );
	StartInstrument( squareWave, NULL );

	SleepUntilTime(theCue, GetAudioTime() + 480 );

	for ( i = 0; i < numofSamples; i++ )
	{
		ReadProbe( outTestProbe, &theOutputValue );
		outSampleValues[i]	= theOutputValue;
	}
		
	/* Draw white graph. */
	pgSetColor( pgcon, 1.0, 1.0, 1.0 );
	pgMoveTo( pgcon, 0, 1.0 );
	for ( i = 0; i < numofSamples; i++ )
	{
		scaledValue	= (float32)i/(float32)numofSamples;
		flipValue	= 1.0 - outSampleValues[i]*vertScale;
		pgDrawTo( pgcon, scaledValue,flipValue );
	}
	

	/* Draw text in pink. */
	pgSetColor( pgcon, .7, .5, .8 );
	pgMoveTo( pgcon, .4, .9 );
	pgDrawText( pgcon, "Output" );


	BeginTest(FT, "integrator.dsp");
	LogResult(InteractiveVerify("You should see a triangle wave on the screen and hear the sound. Did Test Pass?"), "");

	StopInstrument( squareWave, NULL );
	StopInstrument( integrator, NULL );
	StopInstrument( lineOut, NULL );


	/* Clear the screen before exiting */
	pgClearDisplay( pgcon, 0.0, 0.0, 0.0 );
	pgDeleteDisplay( &pgcon );

	UnloadInstrument( integrator );
	UnloadInstrument( squareWave );
	UnloadInstrument( lineOut );

	DeleteCue( theCue );
	DeleteProbe( outTestProbe );
	
	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT, theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}

