/* @(#) Latch_FT.c 96/07/22 1.2 */
/*	
 *	File:				Latch_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional test for latch.dsp. latch.dsp is supposed to pass the input to the
 *	output if the gate is greater than 0. A structure is created which contains the values 
 *	for various input and gate combinatios.  The output of latch.dsp when given the input
 *	combination is compared against the expected output.
 *
 *	The folowing patches need to be in the current working directory in order for this test to run.
 *
 *		ampout.patch
 *
 *	Audio Functions called:
 *
 *		CloseAudioFolio		CreateCue			CreateKnob			LoadInstrument			CreateProbe			
 *		OpenAudioFolio		StartInstrument		StopInstrument		SleepUntilTime			SetKnob
 *		DeleteCue			DeleteProbe			ReadProbe			UnloadInstrument		UnloadPatchTemplate
 *		ConnectInstrumentParts
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <audio/patch.h>
#include <audio/patchfile.h>
#include <stdio.h>

#include "TestMacros.h"
#include "TestUtilities.h"
#include <tests/3do_scaf.h>

/**************************************Test Environment Globals****************/
/******************************************************************************/

typedef	struct
		{
			float32	input;
			float32	gate;
			float32	expectedOutput;
		}  TestValues;


#define	NUM_OF_TESTS		6


int main (int argc, char ** argv)
{
	
	Err		theError;
	float32	theOutputValue;
	int		i;
	Item 	theItem, theTestProbe, theCue, gate, theInputVoltage, constantVoltagePatch, theVoltageKnob;
	TestValues	testList[NUM_OF_TESTS]	= 	{
												1.0,		0,		0,
												1.0,		.5,		1.0,
												1.0,		-.5,	1.0,
												.25,		-.2,	1.0,
												-.6,		.2,		-.6,
												.9,			4,		.9,
											};



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);


	/****************************************************************************************************
	 *	Load the latch instrument ant create the two knobs for it.
	 ****************************************************************************************************/
		
	theItem			= LoadInstrument("latch.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,theItem,"LoadInstrument",cleanup);
		
	gate	= CreateKnob( theItem, "Gate", NULL );
	CHECKRESULT_SETUP(FT,gate,"CreateKnob",cleanup);
	
	/*	Create the test probe	*/
	theTestProbe	= CreateProbe( theItem, "Output", NULL );
	CHECKRESULT_SETUP(FT, theTestProbe,"CreateProbe",cleanup);
	
	/*	Create an instrument patch which supplies a constant voltage. Set the voltage to .1	*/
	constantVoltagePatch	= LoadPatchTemplate( "ampout.patch" );
	theInputVoltage			= CreateInstrument (constantVoltagePatch, NULL);
		
	if ( theInputVoltage < 0 )
	{
		LogComment("Could not create an instrument for  ampout.patch\n" );
		PrintfSysErr(theInputVoltage);
	}
		
	theVoltageKnob			= CreateKnob( theInputVoltage, "Amp", NULL );

	/*	Connect the constant voltage instrument to the latch		*/		
	ConnectInstrumentParts(	theInputVoltage, "Output", 0, theItem, "Input", 0 	);



	for ( i = 0; i < NUM_OF_TESTS; i++ )
	{
		StopInstrument( theItem, NULL );
		StopInstrument( theInputVoltage, NULL );

		/*	Test latch if gate is negative	initially	*/
		SetKnob( theVoltageKnob, testList[i].input );
		SetKnob( gate, testList[i].gate );

		StartInstrument( theItem, NULL );
		StartInstrument( theInputVoltage, NULL );
	
		SleepUntilTime( theCue, GetAudioTime() + 4 );
		ReadProbe( theTestProbe, &theOutputValue );
		
		BeginTest(FT, "latch.dsp");
		if ( (theOutputValue - testList[i].expectedOutput < .1) && (theOutputValue - testList[i].expectedOutput > -.1))
			LogResult(PASSED, "");
		else
			LogResult(FAILED, "Expected %f, but got %f when Input = %f;  gate = %f.",
							 testList[i].expectedOutput, theOutputValue, testList[i].input, testList[i].gate );
	}
	
	StopInstrument( theItem, NULL );
	StopInstrument( theInputVoltage, NULL );

	UnloadInstrument( theItem );
	UnloadPatchTemplate( constantVoltagePatch );
	DeleteCue( theCue );
	DeleteProbe( theTestProbe );
	
	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT,theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}

