/* @(#) PlayMixerDSPs_FT.c 96/07/29 1.2 */
/*	
 *	File:				PlayMixerDSPs_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests for mixers.  There are three tests.
 *
 *	Test 1:	Test 1 x N mixers.  Create a 1 x N mixer. Connect a voltage knob to the input.
 *			Set the gain between the input and each output to maximum.  Check each output
 *			to see if the input has gotten through.  Test MixerSpecToNumOut on each mixer
 *
 *	Test 2:	Test N x 1 mixers.  Create a N x 1 mixer. Connect a voltage knob to each input.
 *			Set the gain between the inputs and the output output to maximum.  Check the output
 *			to see if it is the cumulative sum of the inputs.  Test MixerSpecToNumIn on each mixer
 *
 *	Test 3:	Test 32 x 8 mixers.  Create a 32 x 8 mixer. Connect a random set of inputs that are
 *			at differing voltage levels to a random output. Set the gain between those inputs
 *			and the chosen output to maximum.  Do this until all of the inputs are connect to
 *			one of the outputs.  Check to see if the outputs are the expected values.  This
 *			test ensures that no crosstalk is occurring
 *
 *	Test 4:	Test MixerSpecToFlags
 *
 *	The folowing patches need to be in the current working directory in order for this test to run.
 *
 *		ampout.patch		ampMultipleOut.patch
 *
 *	Audio Functions called:
 *
 *		CloseAudioFolio				CreateProbe			MakeMixerSpec			LoadInstrument			
 *		OpenAudioFolio				CreateKnob			SetKnobPart				StartInstrument
 *		StopInstrument				UnloadInstrument	CreateMixerTemplate		CreateInstrument
 *		CreateProbe					ReadProbePart		DeleteProbe				UnloadPatchTemplate
 *		DeleteMixerTemplate			CalcMixerGainPart	MixerSpecToNumIn		MixerSpecToNumOut
 *		MixerSpecToFlags
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <audio/patch.h>
#include <audio/patchfile.h>
#include <kernel/operror.h>
#include <stdio.h>

#include <TestMacros.h>
#include <TestUtilities.h>
#include <tests/3do_scaf.h>


#define		SOUND_DURATION		240
#define		MAX_NUM_OF_OUTPUTS	8	
#define		MAX_NUM_OF_INPUTS	32	
/**************************************Test Environment Globals****************/
/******************************************************************************/




int main (int argc, char ** argv)
{
	
	Err			theError;
	float32		theOutput, expectedOutput;
	int32		theMixerKnobPartNum;
	Item 		theMixerTemplate, theMixerInstrument, theInputVoltage,
				theMixerKnob, theTestProbe, theMixerMasterKnob, constantVoltagePatch, theVoltageKnob;
	MixerSpec	theMixerSpec, mixerSpec1, mixerSpec2, mixerSpec3, mixerSpec4;
	uint8		i, j, numOfInputs, numOfOutputs;
	uint16		theFlags;



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	/****************************************************************************************************/
	/*	Test #1:	Test the various outputs of a 1-input mixer to make sure they are receiving sound	*/
	/*				from the input.  Use CalcMixerGain to determine the Gain for that Input/Output		*/
	/*				combination, and adjust volume for the output.										*/
	/****************************************************************************************************/

	for ( i = 0; i < MAX_NUM_OF_OUTPUTS; i++ )
	{

		/*	Create Mixer					*/
		theMixerSpec			= MakeMixerSpec(1, i+1, AF_F_MIXER_WITH_AMPLITUDE);
		theMixerTemplate		= CreateMixerTemplate( theMixerSpec, NULL );

		if ( theMixerTemplate < 0 )
		{
			LogComment("FAILED: CreateMixerTemplate could not create the 1x%d Mixer Template", i+1 );
			PrintfSysErr(theMixerTemplate);
		}

		theMixerInstrument		= CreateInstrument( theMixerTemplate, NULL );
		
		if ( theMixerInstrument < 0 )
		{
			LogComment("FAILED: CreateInstrument could not create the 1x%d Mixer Instrument", i+1 );
			PrintfSysErr(theMixerInstrument);
		}
		
		/*	Create a probe for the mixer's output	*/
		theTestProbe			= CreateProbe( theMixerInstrument, "Output", NULL );
		
		if ( theTestProbe < 0 )
		{
			LogComment("FAILED: CreateProbe could not create a probe for the 1x%d Mixer Instrument", i+1 );
			PrintfSysErr(theTestProbe);
		}
		
		/*	Create and Set Mixer Master Knob		*/
		theMixerKnob			= CreateKnob( theMixerInstrument, "Gain", NULL );
		theMixerMasterKnob		= CreateKnob( theMixerInstrument, "Amplitude", NULL );

		if ( theMixerKnob < 0 )
		{
			LogComment("FAILED: CreateKnob could not create the knobs for the 1x%d Mixer Instrument", i+1 );
			PrintfSysErr(theMixerKnob);
		}
		
		if ( theMixerMasterKnob < 0 )
		{
			LogComment("FAILED: CreateKnob could not create the master knob for the 1x%d Mixer Instrument", i+1 );
			PrintfSysErr(theMixerMasterKnob);
		}
		
		SetKnobPart( theMixerMasterKnob, 0, 1.0 );

		/*	Create an instrument patch which supplies a constant voltage. Set the voltage to .5	*/
		constantVoltagePatch	= LoadPatchTemplate( "ampout.patch" );
		theInputVoltage			= CreateInstrument (constantVoltagePatch, NULL);
		
		if ( theInputVoltage < 0 )
		{
			LogComment("FAILED: Could not create an instrument for  ampout.patch" );
			PrintfSysErr(theInputVoltage);
		}
		
		theVoltageKnob			= CreateKnob( theInputVoltage, "Amp", NULL );
		SetKnobPart( theVoltageKnob, 0, .5 );

		/*	Connect the constant voltage instrument to the mixer		*/		
		ConnectInstrumentParts(	theInputVoltage, "Output", 0, theMixerInstrument, "Input", 0 	);

		/*	Test all of the outputs of a 1xj mixer from output 0 to output j(i)	*/
		for ( j = 0; j <= i; j++ )
		{
		
			/*	Set the gain between the input and the output			*/
			theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 0, j );
			SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

			StartInstrument( theInputVoltage, NULL );
			StartInstrument( theMixerInstrument, NULL );					
		
			ReadProbePart( theTestProbe, j, &theOutput );	
		
			StopInstrument( theInputVoltage, NULL );
			StopInstrument( theMixerInstrument, NULL );					

			BeginTest(FT, "1x%d Mixer", i+1);
			
			if ( (theOutput - .5 < .01) && (theOutput - .5 > -.01) )
				LogResult(PASSED, "");
			else
				LogResult(FAILED, "Expected .5, but got %f from output %d of a 1 x %d Mixer", 
								theOutput, j, i+1 );
		}
		
		BeginTest(FT, "MixerSpecToNumOut");
			
		if ( (numOfOutputs = MixerSpecToNumOut(theMixerSpec))  == i+1 )
				LogResult(PASSED, "");
		else
				LogResult(FAILED, "Expected %d, but MixerSpecToNumOut( %dx1 Mixer ) yielded %d", 
								i+1, i+1, numOfOutputs );

		DeleteProbe( theTestProbe );
		UnloadPatchTemplate(constantVoltagePatch);
		DeleteMixerTemplate(theMixerTemplate);							/*	Removes the template and the instrument	*/
	}

	
	
	/****************************************************************************************************/
	/*	Test #2:	Test to see if the inputs of all the different channels are added together properly	*/
	/****************************************************************************************************/

	for ( i = 0; i < MAX_NUM_OF_INPUTS; i++ )
	{

		/*	Create Mixer	*/
		theMixerSpec			= MakeMixerSpec( i+1, 1, 0);
		theMixerTemplate		= CreateMixerTemplate( theMixerSpec, NULL );

		if ( theMixerTemplate < 0 )
		{
			LogComment("FAILED: CreateMixerTemplate could not create the %dx1 Mixer Template", i+1 );
			PrintfSysErr(theMixerTemplate);
		}

		theMixerInstrument		= CreateInstrument( theMixerTemplate, NULL );
		
		if ( theMixerInstrument < 0 )
		{
			LogComment("FAILED: CreateInstrument could not create the %dx1 Mixer Instrument", i+1 );
			PrintfSysErr(theMixerInstrument);
		}
		
		theMixerKnob			= CreateKnob( theMixerInstrument, "Gain", NULL );

		if ( theMixerKnob < 0 )
		{
			LogComment("FAILED: CreateKnob could not create the knobs for the %dx1 Mixer Instrument", i+1 );
			PrintfSysErr(theMixerKnob);
		}
		
		/*	Create a probe for the mixer's output		*/
		theTestProbe			= CreateProbe( theMixerInstrument, "Output", NULL );

		if ( theTestProbe < 0 )
		{
			LogComment("FAILED: CreateProbe could not create a probe for the %dx1 Mixer Instrument", i+1 );
			PrintfSysErr(theTestProbe);
		}
		
		/*	Create the constant voltage instrument.  Set voltage .1		*/
		constantVoltagePatch	= LoadPatchTemplate( "ampout.patch" );
		theInputVoltage			= CreateInstrument (constantVoltagePatch, NULL);
	
		if ( theInputVoltage < 0 )
		{
			LogComment("FAILED: Could not create an instrrument for ampout.patch" );
			goto cleanup;
		}
		
		theVoltageKnob			= CreateKnob( theInputVoltage, "Amp", NULL );
		SetKnobPart( theVoltageKnob, 0, .03 );

		

		for ( j = 0; j <= i; j++ )
		{
			/*	Set the gain between the input and the output			*/
			theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, j, 0 );
			SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );
		
			ConnectInstrumentParts(	theInputVoltage, "Output", 0, theMixerInstrument, "Input", j 	);
			
			StartInstrument( theInputVoltage, NULL );
			StartInstrument( theMixerInstrument, NULL );
	
			ReadProbePart( theTestProbe, 0, &theOutput );	
		
			StopInstrument( theInputVoltage, NULL );
			StopInstrument( theMixerInstrument, NULL );
			
			BeginTest(FT, "%dx1 Mixer", i+1);
			
			if ( (theOutput - ((j+1)*.03) < .01) && (theOutput - ((j+1)*.03) > -.01) )
				LogResult(PASSED, "");
			else
				LogResult(FAILED, "Expected %f, but got %f %dx1 Mixer with .03 going into %d inputs", 
								(j+1)*.03, theOutput,i+1, j+1 );
		}

		BeginTest(FT, "MixerSpecToNumIn");
			
		if ( (numOfInputs = MixerSpecToNumIn(theMixerSpec))  == i+1 )
				LogResult(PASSED, "");
			else
				LogResult(FAILED, "Expected %d, but MixerSpecToNumIn( %dx1 Mixer ) yielded %d", 
								i+1, i+1, numOfInputs );


		DeleteProbe( theTestProbe );
		UnloadPatchTemplate(constantVoltagePatch);
		DeleteMixerTemplate(theMixerTemplate);							/*	Removes the template and the instrument	*/
	}

		

	/****************************************************************************************************/
	/*	Test #3:	Create the largest mixer possible.  Select a random set of inputs and connect them	*/
	/*				at their maximum values to each output.  Compare the value of the output to the		*/
	/*				calculated value.  If the values do not match, then crosstalk has occurred.			*/
	/****************************************************************************************************/

	/*	Create Mixer					*/
	theMixerSpec			= MakeMixerSpec(32, 8, 0);
	theMixerTemplate		= CreateMixerTemplate( theMixerSpec, NULL );

	if ( theMixerTemplate < 0 )
	{
		LogComment("FAILED: CreateMixerTemplate could not create the 32x8 Mixer Template" );
		PrintfSysErr(theMixerTemplate);
	}

	theMixerInstrument		= CreateInstrument( theMixerTemplate, NULL );
		
	if ( theMixerInstrument < 0 )
	{
		LogComment("FAILED: CreateInstrument could not create the 32x8 Mixer Instrument" );
		PrintfSysErr(theMixerInstrument);
	}
		
	/*	Create a probe for the mixer's output	*/
	theTestProbe			= CreateProbe( theMixerInstrument, "Output", NULL );
		
	if ( theTestProbe < 0 )
	{
		LogComment("FAILED: CreateProbe could not create a probe for the 32x8 Mixer Instrument", i+1 );
		PrintfSysErr(theTestProbe);
	}
		
	/*	Create and Set Mixer Knob				*/
	theMixerKnob			= CreateKnob( theMixerInstrument, "Gain", NULL );

	if ( theMixerKnob < 0 )
	{
		LogComment("FAILED: CreateKnob could not create the knobs for the 32x8 Mixer Instrument", i+1 );
		PrintfSysErr(theMixerKnob);
	}
		
	
	/*	Create the constant voltage instrument.  Set voltage .1		*/
	constantVoltagePatch	= LoadPatchTemplate( "ampMultipleOut.patch" );
	theInputVoltage			= CreateInstrument (constantVoltagePatch, NULL);
	
	if ( theInputVoltage < 0 )
	{
		LogComment("FAILED: Could not create an instrrument for ampMultipleOut.patch" );
		goto cleanup;
	}
		
	theVoltageKnob			= CreateKnob( theInputVoltage, "Amp", NULL );

	/*	Set the voltage for each part of the constant voltage instrument.  The voltage rise at
		intervals of .01 for each part.  Connect each output part of the constant voltage
		instrument to each input of the mixer
	*/	
	for ( i = 0; i < 32; i++ )
	{
		SetKnobPart( theVoltageKnob, i, .01*(i+1) );
		ConnectInstrumentParts(	theInputVoltage, "Output", i, theMixerInstrument, "Input", i 	);
	}

	printf("\nEach input which is used for each crosstalk test remains connected to its output for the next test.\n");
	printf("The last four tests have all inputs connected to some output.  In test 8, we force the mixer\n");
	printf("to clip its positive output.  In test 9, we force the mixer to clip its negative output.\n");
	printf("In test 10, we see if the mixer works for negative and positive inputs\n\n");
	/**********************************************************************************************************************/
	/*	Connect mixer inputs 0, 31, 13, 7, 27 to output 0	*/
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 0, 0 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 7, 0 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 13, 0 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 27, 0 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );
	
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 31, 0 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	expectedOutput	= .01*(0+1) + .01*(31+1) + .01*(13+1) + .01*(7+1) + .01*(27+1);
	
	StartInstrument( theInputVoltage, NULL );
	StartInstrument( theMixerInstrument, NULL );
	
	ReadProbePart( theTestProbe, 0, &theOutput );	
		
	StopInstrument( theInputVoltage, NULL );
	StopInstrument( theMixerInstrument, NULL );
			
	BeginTest(FT, "32x8 Mixer: Crosstalk test");
			
	if ( (theOutput - expectedOutput < .001) && (theOutput - expectedOutput > -.001) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected %f, but got %f with Inputs 0 (%f), 7 (%f), 13 (%f), 27 (%f), 31 (%f)", 
						expectedOutput, theOutput, .01*(0+1), .01*(31+1), .01*(13+1), .01*(7+1), .01*(27+1) );
	/**********************************************************************************************************************/
	/**********************************************************************************************************************/
	/*	Connect mixer inputs 16, 17, 22, 5, 1 to output 1	*/
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 1, 1 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 5, 1 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 16, 1 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 17, 1 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );
	
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 22, 1 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	expectedOutput	= .01*(1+1) + .01*(5+1) + .01*(16+1) + .01*(17+1) + .01*(22+1);
	
	StartInstrument( theInputVoltage, NULL );
	StartInstrument( theMixerInstrument, NULL );
	
	ReadProbePart( theTestProbe, 1, &theOutput );	
		
	StopInstrument( theInputVoltage, NULL );
	StopInstrument( theMixerInstrument, NULL );
			
	BeginTest(FT, "32x8 Mixer: Crosstalk test");
			
	if ( (theOutput - expectedOutput < .001) && (theOutput - expectedOutput > -.001) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected %f, but got %f with Inputs 1 (%f), 5 (%f), 16 (%f), 17 (%f), 22 (%f)", 
						expectedOutput, theOutput, .01*(1+1), .01*(5+1), .01*(16+1), .01*(17+1), .01*(22+1) );
	/**********************************************************************************************************************/
	/**********************************************************************************************************************/
	/*	Connect mixer inputs 2, 9, 10, 11, 12 to output 2	*/
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 2, 2 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 9, 2 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 10, 2 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 11, 2 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );
	
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 12, 2 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	expectedOutput	= .01*(2+1) + .01*(9+1) + .01*(10+1) + .01*(11+1) + .01*(12+1);
	
	StartInstrument( theInputVoltage, NULL );
	StartInstrument( theMixerInstrument, NULL );
	
	ReadProbePart( theTestProbe, 2, &theOutput );	
		
	StopInstrument( theInputVoltage, NULL );
	StopInstrument( theMixerInstrument, NULL );
			
	BeginTest(FT, "32x8 Mixer: Crosstalk test");
			
	if ( (theOutput - expectedOutput < .001) && (theOutput - expectedOutput > -.001) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected %f, but got %f with Inputs 2 (%f), 9 (%f), 10 (%f), 11 (%f), 12 (%f)", 
						expectedOutput, theOutput, .01*(2+1), .01*(9+1), .01*(10+1), .01*(11+1), .01*(12+1) );
	/**********************************************************************************************************************/
	/**********************************************************************************************************************/
	/*	Connect mixer inputs 21, 23, 8, 30, 4 to output 3	*/
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 4, 3 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 8, 3 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 21, 3 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 23, 3 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );
	
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 30, 3 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	expectedOutput	= .01*(4+1) + .01*(8+1) + .01*(21+1) + .01*(23+1) + .01*(30+1);
	
	StartInstrument( theInputVoltage, NULL );
	StartInstrument( theMixerInstrument, NULL );
	
	ReadProbePart( theTestProbe, 3, &theOutput );	
		
	StopInstrument( theInputVoltage, NULL );
	StopInstrument( theMixerInstrument, NULL );
			
	BeginTest(FT, "32x8 Mixer: Crosstalk test");
			
	if ( (theOutput - expectedOutput < .001) && (theOutput - expectedOutput > -.001) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected %f, but got %f with Inputs 4 (%f), 8 (%f), 21 (%f), 23 (%f), 30 (%f)", 
						expectedOutput, theOutput, .01*(4+1), .01*(8+1), .01*(21+1), .01*(23+1), .01*(30+1) );
	/**********************************************************************************************************************/
	/**********************************************************************************************************************/
	/*	Connect mixer inputs 3, 14, 15, 19, 29 to output 4	*/
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 3, 4 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 14, 4 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 15, 4 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 19, 4 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );
	
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 29, 4 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	expectedOutput	= .01*(3+1) + .01*(14+1) + .01*(15+1) + .01*(19+1) + .01*(29+1);
	
	StartInstrument( theInputVoltage, NULL );
	StartInstrument( theMixerInstrument, NULL );
	
	ReadProbePart( theTestProbe, 4, &theOutput );	
		
	StopInstrument( theInputVoltage, NULL );
	StopInstrument( theMixerInstrument, NULL );
			
	BeginTest(FT, "32x8 Mixer: Crosstalk test");
			
	if ( (theOutput - expectedOutput < .001) && (theOutput - expectedOutput > -.001) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected %f, but got %f with Inputs 3 (%f), 14 (%f), 15 (%f), 19 (%f), 29 (%f)", 
						expectedOutput, theOutput, .01*(3+1), .01*(14+1), .01*(15+1), .01*(19+1), .01*(29+1) );
	/**********************************************************************************************************************/
	/**********************************************************************************************************************/
	/*	Connect mixer inputs 5, 6, 18, 26, 28 to output 5	*/
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 5, 5 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 6, 5 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 18, 5 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 26, 5 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );
	
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 28, 5 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	expectedOutput	= .01*(5+1) + .01*(6+1) + .01*(18+1) + .01*(26+1) + .01*(28+1);
	
	StartInstrument( theInputVoltage, NULL );
	StartInstrument( theMixerInstrument, NULL );
	
	ReadProbePart( theTestProbe, 5, &theOutput );	
		
	StopInstrument( theInputVoltage, NULL );
	StopInstrument( theMixerInstrument, NULL );
			
	BeginTest(FT, "32x8 Mixer: Crosstalk test");
			
	if ( (theOutput - expectedOutput < .001) && (theOutput - expectedOutput > -.001) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected %f, but got %f with Inputs 5 (%f), 6 (%f), 18 (%f), 26 (%f), 28 (%f)", 
						expectedOutput, theOutput, .01*(5+1), .01*(6+1), .01*(18+1), .01*(26+1), .01*(28+1) );
	/**********************************************************************************************************************/
	/**********************************************************************************************************************/
	/*	Connect mixer inputs 20, 24, 25 to output 6	*/
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 20, 6 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 24, 6 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 25, 6 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	expectedOutput	= .01*(20+1) + .01*(24+1) + .01*(25+1);
	
	StartInstrument( theInputVoltage, NULL );
	StartInstrument( theMixerInstrument, NULL );
	
	ReadProbePart( theTestProbe, 6, &theOutput );	
		
	StopInstrument( theInputVoltage, NULL );
	StopInstrument( theMixerInstrument, NULL );
			
	BeginTest(FT, "32x8 Mixer: Crosstalk test");
			
	if ( (theOutput - expectedOutput < .001) && (theOutput - expectedOutput > -.001) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected %f, but got %f with Inputs 20 (%f), 24 (%f), 25 (%f)", 
						expectedOutput, theOutput, .01*(20+1), .01*(24+1), .01*(25+1) );
	/**********************************************************************************************************************/
	/**********************************************************************************************************************/
	/*	Connect mixer inputs 27, 28, 29, 30, 31 to output 7	*/
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 27, 7 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 28, 7 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 29, 7 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 30, 7 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );
	
	theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, 31, 7 );
	SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );

/*	expectedOutput	= .01*(27+1) + .01*(28+1) + .01*(29+1) + .01*(30+1) + .01*(31+1);
*/
	
	StartInstrument( theInputVoltage, NULL );
	StartInstrument( theMixerInstrument, NULL );
	
	ReadProbePart( theTestProbe, 7, &theOutput );	
		
	StopInstrument( theInputVoltage, NULL );
	StopInstrument( theMixerInstrument, NULL );
			
	BeginTest(FT, "32x8 Mixer: Crosstalk test");
			
	if ( (theOutput - 1.0 < .001) && (theOutput - 1.0 > -.001) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected 1.0, but got %f with Inputs 27 (%f), 28 (%f), 29 (%f), 30 (%f), 31 (%f)", 
						theOutput, .01*(27+1), .01*(28+1), .01*(29+1), .01*(30+1), .01*(31+1) );
	/**********************************************************************************************************************/
	/**********************************************************************************************************************/
	/*	Connect mixer inputs 27, 28, 29, 30, 31 to output 7	*/

	/*	Set the voltage for the last five parts of the knob to negative
	*/	
	for ( i = 27; i < 32; i++ )
		SetKnobPart( theVoltageKnob, i, -.01*(i+1) );


/*	expectedOutput	= -.01*(27+1) - .01*(28+1) - .01*(29+1) - .01*(30+1) - .01*(31+1);
*/
	
	
	StartInstrument( theInputVoltage, NULL );
	StartInstrument( theMixerInstrument, NULL );
	
	ReadProbePart( theTestProbe, 7, &theOutput );	
		
	StopInstrument( theInputVoltage, NULL );
	StopInstrument( theMixerInstrument, NULL );
			
	BeginTest(FT, "32x8 Mixer: Crosstalk test");
			
	if ( (theOutput + 1.0 < .001) && (theOutput + 1.0 > -.001) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected -1.0, but got %f with Inputs 27 (%f), 28 (%f), 29 (%f), 30 (%f), 31 (%f)", 
						theOutput, -.01*(27+1), -.01*(28+1), -.01*(29+1), -.01*(30+1), -.01*(31+1) );
	/**********************************************************************************************************************/
	/**********************************************************************************************************************/
	/*	Connect mixer inputs 27, 28, 29, 30, 31 to output 7	*/

	/*	Set the voltage for the last three parts of the knob to positive
	*/	
	for ( i = 29; i < 32; i++ )
		SetKnobPart( theVoltageKnob, i, .01*(i+1) );


	expectedOutput	= -.01*(27+1) - .01*(28+1) + .01*(29+1) + .01*(30+1) + .01*(31+1);
	
	StartInstrument( theInputVoltage, NULL );
	StartInstrument( theMixerInstrument, NULL );
	
	ReadProbePart( theTestProbe, 7, &theOutput );	
		
	StopInstrument( theInputVoltage, NULL );
	StopInstrument( theMixerInstrument, NULL );
			
	BeginTest(FT, "32x8 Mixer: Crosstalk test");
			
	if ( (theOutput - expectedOutput < .001) && (theOutput - expectedOutput > -.001) )
		LogResult(PASSED, "");
	else
		LogResult(FAILED, "Expected %f, but got %f with Inputs 27 (%f), 28 (%f), 29 (%f), 30 (%f), 31 (%f)", 
						expectedOutput, theOutput, -.01*(27+1), -.01*(28+1), .01*(29+1), .01*(30+1), .01*(31+1) );
	/**********************************************************************************************************************/


	/****************************************************************************************************/
	/*	Test #4:	Test MixerSpecToFlags																*/
	/****************************************************************************************************/
		mixerSpec1			= MakeMixerSpec(2, 2, 0);
		mixerSpec2			= MakeMixerSpec(2, 2, AF_F_MIXER_WITH_AMPLITUDE);
		mixerSpec3			= MakeMixerSpec(2, 2, AF_F_MIXER_WITH_LINE_OUT);
		mixerSpec4			= MakeMixerSpec(2, 2, AF_F_MIXER_WITH_AMPLITUDE | AF_F_MIXER_WITH_LINE_OUT);
		
	BeginTest(FT, "MixerSpecToFlags");
	if ( (theFlags = MixerSpecToFlags(mixerSpec1))  == 0 )
				LogResult(PASSED, "");
		else
				LogResult(FAILED, "Expected 0, but MixerSpecToFlags(0) yielded %d", theFlags );
	
	BeginTest(FT, "MixerSpecToFlags");
	if ( (theFlags = MixerSpecToFlags(mixerSpec2))  == AF_F_MIXER_WITH_AMPLITUDE )
				LogResult(PASSED, "");
		else
				LogResult(FAILED, "Expected AF_F_MIXER_WITH_AMPLITUDE, but MixerSpecToFlags(AF_F_MIXER_WITH_AMPLITUDE) yielded %d", theFlags );
	
	BeginTest(FT, "MixerSpecToFlags");
	if ( (theFlags = MixerSpecToFlags(mixerSpec3))  == AF_F_MIXER_WITH_LINE_OUT )
				LogResult(PASSED, "");
		else
				LogResult(FAILED, "Expected AF_F_MIXER_WITH_LINE_OUT, but MixerSpecToFlags(AF_F_MIXER_WITH_LINE_OUT) yielded %d", theFlags );
	
	BeginTest(FT, "MixerSpecToFlags");
	if ( (theFlags = MixerSpecToFlags(mixerSpec4))  == (AF_F_MIXER_WITH_LINE_OUT | AF_F_MIXER_WITH_AMPLITUDE) )
				LogResult(PASSED, "");
		else
				LogResult(FAILED, "Expected AF_F_MIXER_WITH_LINE_OUT | AF_F_MIXER_WITH_AMPLITUDE, but MixerSpecToFlags(AF_F_MIXER_WITH_LINE_OUT | AF_F_MIXER_WITH_AMPLITUDE) yielded %d", theFlags );
	
	DeleteProbe( theTestProbe );
	UnloadPatchTemplate(constantVoltagePatch);
	DeleteMixerTemplate(theMixerTemplate);							/*	Removes the template and the instrument	*/

	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT, theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}


/*	
 *	SetMixerGainKnobValue
 *
 *	This function quickly sets the gain knob values on an n x 2 mixer.  whichInput
 *	specifies which input should be used to calculate the output gains.  value
 *	specifies what the gain should be set.  numOfInputs indicates the size of the mixer.
 *
 *	If whichInput < 0, then the function initalizes the gain for all the inputs to 0.
 *	If whichInput >= numOfInputs, then the function sets the gain for all inputs to 1.0
 */
void SetMixerGainKnobValue( int whichInput, float32 value, int numOfInputs, Item theMixerKnob, MixerSpec theMixerSpec )
{

	int32	 i, theMixerKnobPartNum;
	

	/*	Set the volume to all knobs to Maximum		*/
	for ( i = 0; i < numOfInputs; i++ )
	{

		if ( whichInput < 0 )						/*	Set all gain knobs to 0	*/
		{
			theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, i, 0 );
			SetKnobPart( theMixerKnob, theMixerKnobPartNum, 0 );
		
			theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, i, 1 );
			SetKnobPart( theMixerKnob, theMixerKnobPartNum, 0 );
		}
		else if ( whichInput >= numOfInputs )		/*	Set all gain knobs to 1.0	*/
		{
			theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, i, 0 );
			SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );
		
			theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, i, 1 );
			SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );
		}
		else										/*	Set specified gain knob to specified value	*/
		{
			theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, whichInput, 0 );
			SetKnobPart( theMixerKnob, theMixerKnobPartNum, value );
		
			theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, whichInput, 1 );
			SetKnobPart( theMixerKnob, theMixerKnobPartNum, value );
			
			break;
		}
	}
			
}
