/* @(#) PlaySampleDSPs_FT.c 96/07/29 1.3 */
/*	
 *	File:				PlaySampleDSPs_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *		
 *	This file contains the functional tests for the sample dsps.  The test program gets the
 *	name of the sample instrument and the corresponding sound file it will play from PlaySampleDSPs_FT_data.h.
 *	The sample files should be sound files sampled at a rate of 44.1 KHz.  The tests require the
 *	user to listen to the output and use the control pad to enter whether the test passed (Shift-A)
 *	or failed (Shift-X).  For all tests except for the variable rate tests, the sample should sound normal.
 *
 *	Before the tests, each instrument is loaded (calcRateDivider argument to LoadInstrument is 1), 
 *	attached to the sample file, and connencted to the line_out instrument. There are four tests.
 *
 *	Test 1:	Play the sample at the default sample rate and amplitude.  The sound file should be clear and
 *			intelligible.
 *
 *	Test 2:	Play the sample at the default rate and at a random amplitiude. If the amplitude is too low, the
 *			sample may not be heard, but the test will still pass.  Look in the debugger window to see what
 *			the amplitude is.
 *
 *	Test 3:	Play the sample at the default rate and at a negative random amplitiude.
 *
 *	Test 4:	If the sample instrument is variable, play the sample at the default amplitude and a random rate. If
 *			the rate is above 44.1 KHz, the sound file will sound fast.  If it is below 44.1 KHz, it will sound slow.
 *
 *	Uses the following calls:
 *
 *	OpenAudioFolio	 		CreateCue			LoadInstrument		LoadSample
 *	GetCueSignal			CreateAttachment	CreateKnob,
 *	ConnectInstrumentParts	StartInstrument		StopInstrument		WaitSignal
 *	SetKnobPart				DeleteAttachment	UnloadInstrument	UnloadSample
 *
 *	Dsp Instruments used:
 *
 *	sampler_16_f1.dsp		sampler_16_v1.dsp		sampler_8_f1.dsp		sampler_8_v1.dsp
 *	sampler_raw_f1.dsp		sampler_adp4_v1.dsp		sampler_sqs2_f1.dsp		sampler_sqs2_v1.dsp
 *	sampler_cbd2_f1.dsp		sampler_cbd2_v1.dsp		sampler_16_f2.dsp		sampler_16_v2.dsp
 *	sampler_8_f2.dsp		sampler_8_v2.dsp		sampler_cbd2_f2.dsp		sampler_cbd2_v2.dsp
 *
 *	Audio files Used (These files must be present in the current directory)
 *
 *	44K_16B_M.AIFF				44K_16B_M.AIFF				44K_8B_M.AIFF				44K_8B_M.AIFF
 *	44K_16B_M.AIFF				44K_16B_M_ADP4_FAST.AIFF	44K_16B_M_SQS2_FAST.AIFF	44K_16B_M_SQS2_FAST.AIFF	
 *	44K_16B_M_CBD2_FAST.AIFF	44K_16B_M_CBD2_FAST.AIFF	44K_16B_S.AIFF				44K_16B_S.AIFF
 *	44K_8B_S.AIFF				44K_8B_S.AIFF				44K_16B_S_CBD2_FAST.AIFF	44K_16B_S_CBD2_FAST.AIFF
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <audio/parse_aiff.h>
#include <stdio.h>
#include <string.h>

#include "TestMacros.h"
#include "TestUtilities.h"
#include <tests/3do_scaf.h>

/**************************************Test Environment Globals****************/
/******************************************************************************/

int main (int argc, char ** argv)
{
	#include "PlaySampleDSPs_FT_data.h"
	
	Err		theError;
	float32	amplitude, rate;
	int32	p[10], theSignal, theNumber;
	Item 	theItem, theLineOut, theSample, theAttachment,
			theCue, theAmplitude, theRate;



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);

	#define expected 0

	/*	Since we're using ARGUMENT_02, we must initialize p[2] and increment it	*/
	for ( p[1]=0, p[2]=0; p[1]<NUM_PARAMS_01; p[1]++, p[2]++ )
	{
		/****************************************************************************************************/
		/*	Load the sample instrument and the line_out instrument (which provides the sound).				*/
		/*	Load the sample and attach it to the sample instrument.  Call MonitorAttachment which will		*/
		/*	let you know through theCue when the end of the sample has been reached when the instrument is	*/
		/*	played.  theSignal is theSignal for that particular attachment.									*/
		
		theItem			= LoadInstrument(ARGUMENT_01, 1, 100);
		theLineOut		= LoadInstrument("line_out.dsp", 1, 100);
		theSample		= LoadSample(ARGUMENT_02);
		theAttachment	= CreateAttachment( theItem, theSample, NULL );
		MonitorAttachment( theAttachment, theCue, CUE_AT_END );
		theSignal		= GetCueSignal( theCue );
		
		theAmplitude	= CreateKnob( theItem, "Amplitude", NULL );

		/****************************************************************************************************/
		/*	Monophonic instrument areare listed before the 10th element in the structure					*/
		/*	Stereo instruments are found after the 10th element												*/
		
		ConnectInstrumentParts(	theItem, "Output", 0, 
								theLineOut, "Input", 0 );		/*	Connect the left channel	*/

		if ( p[1] < 10 )
			ConnectInstrumentParts(	theItem, "Output", 0, 
									theLineOut, "Input", 1 );	/*	Connect the right channel - Mono Sample	*/
		else
			ConnectInstrumentParts(	theItem, "Output", 1, 
									theLineOut, "Input", 1 );	/*	Connect the right channel - Stereo Sample	*/

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sample with Default Amplitude and Default Sample Rate										*/
		/****************************************************************************************************/
		BeginTest(FT, "%s", ARGUMENT_01);
		LogComment("%d. Playing %s instrument at default amplitude and sample rate", p[1], ARGUMENT_01 );

		StartInstrument( theItem, NULL );					/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		WaitSignal( theSignal );							/*	Enter a Wait state until sample is finished	*/

		/****************************************************************************************************/
		StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );

		LogResult(InteractiveVerify("You should have heard the sample at its normal amplitude and rate. Did Test Pass?"), "");

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sample with a Random Amplitude and Default Sample Rate										*/
		/****************************************************************************************************/
		theNumber	= GetRandomNumber( 30, 100);
		
		amplitude	= theNumber / 100.0;
		
		BeginTest(FT, "%s", ARGUMENT_01);
		LogComment("%d. Playing %s instrument with %f amplitude", p[1], ARGUMENT_01, amplitude );
		
		SetKnobPart( theAmplitude, 0, amplitude );

		StartInstrument( theItem, NULL );					/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		WaitSignal( theSignal );							/*	Enter a Wait state until sample is finished	*/

		/****************************************************************************************************/
		StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );

		LogResult(InteractiveVerify("You should have heard the sample at the given amplitude and normal rate. Did Test Pass?"), "");

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sample with a Random Negative Amplitude and Default Sample Rate							*/
		/****************************************************************************************************/
		BeginTest(FT, "%s", ARGUMENT_01);
		LogComment("%d. Playing %s instrument with %f amplitude", p[1], ARGUMENT_01, -amplitude );
		
		SetKnobPart( theAmplitude, 0, -amplitude );

		StartInstrument( theItem, NULL );					/*	Play sound	*/
		StartInstrument( theLineOut, NULL );
		
		/****************************************************************************************************/
		
		WaitSignal( theSignal );							/*	Enter a Wait state until sample is finished	*/

		/****************************************************************************************************/
		StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
		StopInstrument( theLineOut, NULL );

		LogResult(InteractiveVerify("You should have heard the sample at the given amplitude and normal rate. Did Test Pass?"), "");

		/****************************************************************************************************/
		/****************************************************************************************************/
		/*	Play Sample with Default Amplitude and a Random Sample Rate										*/
		/****************************************************************************************************/
		if ( p[1] % 2 != 0 )
		{
			theRate		= CreateKnob( theItem, "SampleRate", NULL );
			SetKnobPart( theAmplitude, 0, 1 );					/*	Set the amplitude to the maximum	*/

			if (	(strcmp( ARGUMENT_01, "sampler_adp4_v1.dsp"	 ) == 0	)	||
					(strcmp( ARGUMENT_01, "sampler_cbd2_v1.dsp"	 ) == 0	)	||
					(strcmp( ARGUMENT_01, "sampler_cbd2_v2.dsp"	 ) == 0	)		)
				theNumber	= GetRandomNumber( 220500, 441000);			/*	These instruments have a sample rate range of 0 to 44100.  Going beyond this rate will risk
																			the sample being played at a tremendously slow speed	*/
			else
				theNumber	= GetRandomNumber( 220500, 882000);

			rate	= theNumber / 10.0;
		
			BeginTest(FT, "%s", ARGUMENT_01);
			LogComment("%d. Playing %s instrument with %f sample rate", p[1], ARGUMENT_01, rate );
		
			SetKnobPart( theRate, 0, rate );

			StartInstrument( theItem, NULL );					/*	Play sound	*/
			StartInstrument( theLineOut, NULL );
		
			/****************************************************************************************************/
		
			WaitSignal( theSignal );							/*	Enter a Wait state until sample is finished	*/

			/****************************************************************************************************/
			StopInstrument( theItem, NULL );					/*	Stop Playing the sound	*/
			StopInstrument( theLineOut, NULL );
		
			if (rate < 44100.0)
				LogResult(InteractiveVerify("The sample should have sounded slower and lower-pitched. Did Test Pass?"), "");
			else if (rate > 44100.0)
				LogResult(InteractiveVerify("The sample should have sounded faster and higher-pitched. Did Test Pass?"), "");
			else
				LogResult(InteractiveVerify("The sample should have sounded normal. Did Test Pass?"), "");

		}
		/****************************************************************************************************/

		DeleteAttachment( theAttachment );
		UnloadSample( theSample );
		UnloadInstrument(theItem);
		UnloadInstrument(theLineOut);
		/****************************************************************************************************/
	}	
	
	DeleteCue( theCue );

	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT,theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}

