/* @(#) Randomhold_FT.c 96/07/22 1.2 */
/*	
 *	File:				Randomhold_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests for the randomhold.dsp.  The instrument is loaded and the knobs 
 *	are created.  After setting the knobs to a particular amplitude and frequency, the tests checks the 
 *	output values of the instrument after a certain time interval.  If the value of the present output is
 *	different than the value of a previous output then the test has passed because a "random" output has been
 *	generated.  The output values are printed to the screen so that the user may see them.
 *
 *	Uses the following calls:
 *
 *	OpenAudioFolio	 		CreateCue			LoadInstrument	
 *	CreateProbe				CreateKnob			StopInstrument		
 *	ReadProbe				StartInstrument		SleepUntilTime
 *	SetKnob					UnloadInstrument
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <stdio.h>

#include "TestMacros.h"
#include "TestUtilities.h"
#include <tests/3do_scaf.h>

/**************************************Test Environment Globals****************/
/******************************************************************************/

int main (int argc, char ** argv)
{
	
	Err		theError;
	float32	theOutputValue, previousOutputValue = 0;
	int		i;
	Item 	theItem, theTestProbe, theCue, frequency, amplitude;



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);


	/****************************************************************************************************
	 *	Load the randomhold instrument ant create the two knobs for it.
	 ****************************************************************************************************/
		
	theItem			= LoadInstrument("randomhold.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,theItem,"LoadInstrument",cleanup);
		
	amplitude	= CreateKnob( theItem, "Amplitude", NULL );
	CHECKRESULT_SETUP(FT,amplitude,"CreateKnob",cleanup);
	
	frequency	= CreateKnob( theItem, "Frequency", NULL );
	CHECKRESULT_SETUP(FT,frequency,"CreateKnob",cleanup);

	/*	Create the test probe	*/
	theTestProbe	= CreateProbe( theItem, "Output", NULL );
	CHECKRESULT_SETUP(FT,theTestProbe,"CreateProbe",cleanup);
	
	SetKnob( amplitude, 1.0 );
	SetKnob( frequency, 22050.0 );
	StartInstrument( theItem, NULL );
	
	for ( i = 0; i < 6; i++ )
	{
		SleepUntilTime( theCue, GetAudioTime() + 44100/22050 );
		ReadProbe( theTestProbe, &theOutputValue );
		
		if ( i != 0 )
		{
			BeginTest(FT, "randomhold.dsp");
			if ( theOutputValue != previousOutputValue )
				LogResult(PASSED, "" );
			else
				LogResult(FAILED, "New value is the same as previous value." );
		}
		
		previousOutputValue	= theOutputValue;
	}
		
	SetKnob( amplitude, .5 );

	for ( i = 0; i < 6; i++ )
	{
		SleepUntilTime( theCue, GetAudioTime() + 44100/22050 );
		ReadProbe( theTestProbe, &theOutputValue );
		
		if ( i != 0 )
		{
			BeginTest(FT, "randomhold.dsp");
			if ( (theOutputValue != previousOutputValue) && (theOutputValue  < .5) && (theOutputValue  > -.5) )
				LogResult(PASSED, "" );
			else
				LogResult(FAILED, "New value is the same as previous value or greater than abs(.5)." );
		}
		
		previousOutputValue	= theOutputValue;
	}
		

	SetKnob( amplitude, 1.0 );
	SetKnob( frequency, -11025.0 );
	
	for ( i = 0; i < 6; i++ )
	{
		SleepUntilTime( theCue, GetAudioTime() + 44100/11025 );
		ReadProbe( theTestProbe, &theOutputValue );
		
		if ( i != 0 )
		{
			BeginTest(FT, "randomhold.dsp");
			if ( theOutputValue != previousOutputValue )
				LogResult(PASSED, "" );
			else
				LogResult(FAILED, "New value is the same as previous value." );
		}
		
		previousOutputValue	= theOutputValue;
	}
		

	SetKnob( amplitude, -.5 );

	for ( i = 0; i < 6; i++ )
	{
		SleepUntilTime( theCue, GetAudioTime() + 44100/22050 );
		ReadProbe( theTestProbe, &theOutputValue );
		
		if ( i != 0 )
		{
			BeginTest(FT, "randomhold.dsp");
			if ( (theOutputValue != previousOutputValue) && (theOutputValue < .5) && (theOutputValue > -.5) )
				LogResult(PASSED, "" );
			else
				LogResult(FAILED, "New value is the same as previous value or greater than abs(-.5)." );
		}
		
		previousOutputValue	= theOutputValue;
	}
		

	StopInstrument( theItem, NULL );

	UnloadInstrument( theItem );
	DeleteCue( theCue );
	DeleteProbe( theTestProbe );
	
	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT, theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}

