/* @(#) SvFilter_FT.c 96/07/29 1.3 */
/*	
 *	File:				SvFilter_FT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests for svfilter.dsp.  This program draws the frequency response of
 *	svfilter with a given resonance to a sinewave centered around middle C.  The frequency range tested
 *	goes from 130.5 to 523 Hz.  The lowpass (output) frequency response is drawn on the television screen.
 *	The vertical line represents the critical frequency. By pressing the down button on the control pad, the user can
 *	also see the lowpass, bandpass, and the highpass frequency responses.  Pressing the down button, will take
 *	the user back to a previous frequency response.  The user can exit by pressing the X button.
 *
 *	The user can set test parameters on the command line.  There are nine parameters the user can set.
 *
 *	-a		svfilter's amplitude. default is 1.0
 *	-f		svfilter's critical frequeny. default is 261.0 Hz
 *	-r		svfilter's resonance.. default is .125
 *	-l		the low end of the frequency range. default is 130.0
 *	-h		the high end of the frequency range. default is 523.0
 *	-d		the time (in ticks) to wait before the test program reads the response to a certain frequency 
 *			(the longer, the better - 60 ticks should be the maximum value needed. Any longer will make the
 *			test program way too slow). default is 60.0
 *	-o		which filter output the user wants displayed first. default is SCALED_LOWPASS_OUTPUT
 *	-p		print the graphed values into the debugger window. Default is 0 - "do not print"
 *	-v		vertical graphing scaling factor. default is .9
 *
 *	This test program works by creating a loop ranging from the lowest frequency to the highest frequency.
 *	Each time in the loop, the sinewave sample is played at that frequency into the svfilter's input.  The
 *	output of the svfilter is connected the an envfollower which follows the level of the svfilter's output.
 *	The test program waits a specified amout of time before reading the envfollower's output into an array.
 *	(We must allow the envfollower time to rise to the final output value.  The test program takes this array
 *	of values and plots them on the TV screen.
 *
 *	In order to build this program, you must -laudiodemo to the LIBS line of Integrator_FT.make, and you must
 *	place libaudiodemo.a in the 3DO:Interfaces:Libs:Default folder.
 *
 *	Audio Functions called:
 *
 *		CloseAudioFolio		CreateCue			CreateAttachment		LoadInstrument
 *		OpenAudioFolio		CreateInstrument	CreateKnob				CreateProbe
 *		SetKnobPart			StartInstrument		StopInstrument			LoadSample
 *		SleepUntilTime		DeleteAttachment	DeleteCue				UnloadInstrument
 *		DeleteProbe			UnloadSample		ConnectInstrumentParts
 *
 *	Dsp Instruments used:
 *
 *	sampler_16_f1.dsp		svfilter.dsp		envfollower.dsp
 *
 *	Audio files Used (These files must be present in the current directory)
 *
 *	sinewave.aiff				
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <audio/parse_aiff.h>
#include <graphics/graphics.h>
#include <kernel/operror.h>
#include <misc/event.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <audiodemo/portable_graphics.h>
#include <TestMacros.h>
#include <TestUtilities.h>
#include <tests/3do_scaf.h>

/**************************************Test Environment Globals****************/
/******************************************************************************/

#define	MAX_NUM_OF_SAMPLES			2000
#define	SCALED_LOWPASS_OUTPUT		0
#define	LOWPASS_OUTPUT				1
#define	BANDPASS_OUTPUT				2
#define	HIGHPASS_OUTPUT				3

void WaitButton( uint32 button );
void DoThisButton( void );


float32		 				outSampleValues[MAX_NUM_OF_SAMPLES], lowSampleValues[MAX_NUM_OF_SAMPLES],
							bandSampleValues[MAX_NUM_OF_SAMPLES], highSampleValues[MAX_NUM_OF_SAMPLES],
							vertScale	= .9,
							freqArg		= 261.0;
int32						diffFreq, lowFreq	= 130, highFreq	= 523, whichOutput	= SCALED_LOWPASS_OUTPUT;
PortableGraphicsContext		*pgcon;

int main (int argc, char ** argv)
{
	
	char		c;
	Err			theError;
	float32		theOutputValue, digitalFreq, scaledValue, flipValue,
				resArg	= .125, ampArg	= 1.0, freqCounter;
	int			doPrint	= 0, newArgc;
	int32		i, duration	= 60;
	Item 		svfilter, svFrequency, svResonance, svAmplitude, sineWave,
				out_envfollower, low_envfollower, band_envfollower, high_envfollower,
				outTestProbe, lowTestProbe, bandTestProbe, highTestProbe,
				theCue, sleepCue, theAttachment, theSample;
				


	newArgc	= InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	/******************************************************************************/

	while( --newArgc > 0  && (*++argv)[0] == '-' )
	{
		c	= *++argv[0];
		switch( c )
		{
			case	'a'	:									/*	Get amplitude from the command line	*/
							if ( --argc > 0 )
							{
								ampArg	= atof32( *++argv );
								if ( (ampArg < 0) || (ampArg > 1) ) 
								{
									printf( "\nSvFilter_FT: %f is out of range.\tAmplitude must be between 0.0 and 1.0\n", ampArg );
									exit(1);
								}
							}
							break;
			case	'f'	:									/*	Get frequency from the command line	*/
							if ( --argc > 0 )
							{
								freqArg	= atof32( *++argv );
								if ( (freqArg < 130.5) || (freqArg > 523) )  
								{
									printf( "\nSvFilter_FT: %f is out of range.\tCritical frequency must be between 130.5 and 523.0\n", freqArg );
									exit(1);
								}
							}
							break;
			case	'r'	:									/*	Get resonance from the command line	*/
							if ( --argc > 0 )
							{
								resArg	= atof32( *++argv );
								if ( (resArg < 0) || (resArg > 1.0) )  
								{
									printf( "\nSvFilter_FT: %f is out of range.\tResonance must be between 0.0 and 1.0\n", resArg );
									exit(1);
								}
							}
							break;
			case	'l'	:									/*	Get low frequency range  from the command line	*/
							if ( --argc > 0 )
							{
								lowFreq	= atol( *++argv );
								if ( (lowFreq < 130) || (lowFreq > 523) ) 
								{
									printf( "\nSvFilter_FT: %d is out of range.\tLow Frequency Range must be between 130 and 523 Hz\n", lowFreq );
									exit(1);
								}
							}
							break;
			case	'h'	:									/*	Get high frequency range  from the command line	*/
							if ( --argc > 0 )
							{
								highFreq	= atol( *++argv );
								if ( (highFreq < 130) || (highFreq > 523) ) 
								{
									printf( "\nSvFilter_FT: %d is out of range.\tHigh Frequency Range must be between 130 and 523 Hz\n", highFreq );
									exit(1);
								}
							}
							break;
			case	'd'	:									/*	Get duration (the time to wait before reading probes)  from the command line	*/
							if ( --argc > 0 )
							{
								duration	= atol( *++argv );
								if ( (duration < 0) || (duration > 60) )
								{
									printf( "\nSvFilter_FT: %d is out of range.\tWait time must be between 0 and 60\n", duration );
									exit(1);
								}
							}
							break;
			case	'o'	:									/*	Determine which output to show first from the command line	*/
							if ( --argc > 0 )
							{
								if ( strcmp( *++argv, "output" ) == 0 )
									whichOutput	= SCALED_LOWPASS_OUTPUT;
								else if ( strcmp( *argv, "lowpass" ) == 0 )
									whichOutput	= LOWPASS_OUTPUT;
								else if ( strcmp( *argv, "bandpass" ) == 0 )
									whichOutput	= BANDPASS_OUTPUT;
								else if ( strcmp( *argv, "highpass" ) == 0 )
									whichOutput	= HIGHPASS_OUTPUT;
								else
								{
									printf( "\nSvFilter_FT: %s is not an option\tOptions are <output>, <lowpass>, <bandpass>, or <highpass>\n", *argv );
									exit(1);
								}
							}
							break;
			case	'p'	:									/*	Get the printing option from the command line	*/
							if ( --argc > 0 )
							{
								doPrint	= atoi( *++argv );
								if ( doPrint < 0 ) 
								{
									printf( "\nSvFilter_FT: %d is out of range.\tPrint option must be either 1 to print or 0 disable printing\n", ampArg );
									exit(1);
								}
							}
							break;
			case	'v'	:									/*	Get vertical scaling factor from the command line	*/
							if ( --argc > 0 )
							{
								vertScale	= atof32( *++argv );
								if ( (vertScale < 0) || (vertScale > 1) ) 
								{
									printf( "\nSvFilter_FT: %f is out of range.\tVertical scale must be between 0.0 and 1.0\n", ampArg );
									exit(1);
								}
							}
							break;
			default		:	
							printf( "\nSvFilter_FT: Illegal option -%c\n", c );
							printf( "\nUsage:\tSvFilter_FT\t-a (amplitude) -f (critical frequency) -r (resonance)\n\t\t\t\t\t-l (low frequency) -h(high frequency) -d(wait duration)\n" );
							exit(1);
							break;
		}
	}
	
	if ( highFreq < lowFreq )
	{
		printf("Error: High part of the frequency range is lower than low part of frequency range.\nExiting program...\n");
		exit(1);
	}
		



	/* Initialize the EventBroker. */
	theError = InitEventUtility(1, 0, LC_FocusListener);
	CHECKRESULT_SETUP(FT,theError,"InitEventUtility",cleanup);

	/* Set up double buffered display. */
	theError = pgCreateDisplay( &pgcon, 2 );
	CHECKRESULT_SETUP(FT,theError,"pgCreateDisplay",cleanup);

	/* Draw Background so we can see pixel colors. */
	pgClearDisplay( pgcon, 0.0, 0.0, 0.0 );


	theError =	 OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);

	sleepCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,sleepCue,"CreateCue",cleanup);


	/****************************************************************************************************
	 *	Load the svfilter instrument and create the amplitude and offset knobs
	 ****************************************************************************************************/
		
	svfilter		= LoadInstrument("svfilter.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,svfilter,"LoadInstrument(integrator.dsp)",cleanup);
		
	svAmplitude	= CreateKnob( svfilter, "Amplitude", NULL );
	CHECKRESULT_SETUP(FT,svAmplitude,"CreateKnob",cleanup);
	
	svResonance		= CreateKnob( svfilter, "Resonance", NULL );
	CHECKRESULT_SETUP(FT,svResonance,"CreateKnob",cleanup);
	
	svFrequency		= CreateKnob( svfilter, "Frequency", NULL );
	CHECKRESULT_SETUP(FT,svFrequency,"CreateKnob",cleanup);
	

	/****************************************************************************************************
	 *	Load the sine wave instrument
	 ****************************************************************************************************/
		
	sineWave		= LoadInstrument("sampler_16_v1.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,sineWave,"LoadInstrument(sampler_16_v1.dsp)",cleanup);
		
	theSample		= LoadSample("sinewave.aiff");
	CHECKRESULT_SETUP(FT,theSample,"LoadSample(sinewave.aiff)",cleanup);
	
	theAttachment	= CreateAttachment( sineWave, theSample, NULL );
	MonitorAttachment( theAttachment, theCue, CUE_AT_END );


	/****************************************************************************************************
	 *	Create the envelope follower instruments
	 ****************************************************************************************************/

	out_envfollower		= LoadInstrument("envfollower.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,out_envfollower,"LoadInstrument(envfollower.dsp)",cleanup);
		
	low_envfollower		= LoadInstrument("envfollower.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,low_envfollower,"LoadInstrument(envfollower.dsp)",cleanup);
		
	band_envfollower	= LoadInstrument("envfollower.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,band_envfollower,"LoadInstrument(envfollower.dsp)",cleanup);
		
	high_envfollower	= LoadInstrument("envfollower.dsp", 1, 100);
	CHECKRESULT_SETUP(FT,high_envfollower,"LoadInstrument(envfollower.dsp)",cleanup);
		
		
	/****************************************************************************************************
	 *	Create the testprobes for the envelope follower instruments
	 ****************************************************************************************************/

	outTestProbe	= CreateProbe( out_envfollower, "Output", NULL );
	CHECKRESULT_SETUP(FT,outTestProbe,"CreateProbe",cleanup);
	
	lowTestProbe	= CreateProbe( low_envfollower, "Output", NULL );
	CHECKRESULT_SETUP(FT,lowTestProbe,"CreateProbe",cleanup);
	
	bandTestProbe	= CreateProbe( band_envfollower, "Output", NULL );
	CHECKRESULT_SETUP(FT,bandTestProbe,"CreateProbe",cleanup);
	
	highTestProbe	= CreateProbe( high_envfollower, "Output", NULL );
	CHECKRESULT_SETUP(FT,highTestProbe,"CreateProbe",cleanup);
	
	
	/****************************************************************************************************
	 *	Connect the output of the sampler_16_f1.dsp to the input of the svfilter.dsp.
	 *	Connect the Output of the svfilter.dsp to input of out_envfollwer..dsp
	 *	Connect the LowPass Output of the svfilter.dsp to input of low_envfollwer..dsp
	 *	Connect the BandPass Output of the svfilter.dsp to input of band_envfollwer..dsp
	 *	Connect the HighPass Output of the svfilter.dsp to input of high_envfollwer..dsp
	 ****************************************************************************************************/
		
	ConnectInstrumentParts(	sineWave, "Output", 0, svfilter, "Input", 0 		);
	ConnectInstrumentParts(	svfilter, "Output", 0, out_envfollower, "Input", 0 		);
	ConnectInstrumentParts(	svfilter, "LowPass", 0, low_envfollower, "Input", 0 		);
	ConnectInstrumentParts(	svfilter, "BandPass", 0, band_envfollower, "Input", 0 		);
	ConnectInstrumentParts(	svfilter, "HighPass", 0, high_envfollower, "Input", 0 		);

							
	/*	Calculate the number which controls the svfilter's frequency	*/		
	digitalFreq	= 2.0 * sinf(M_PI * freqArg / 44100.0);
		
	SetKnob( svFrequency, digitalFreq);
	SetKnob( svResonance, resArg);
	SetKnob( svAmplitude, ampArg);

	LogComment("\tsvfilter.dsp Setup: (Freq. = %f; Amp. = %f; Res. = %f)", freqArg, ampArg, resArg);		
	LogComment("\tFrequency range is from %d Hz to %d Hz.  Wait time is %d ticks.  Vertical scaling factor is %f", lowFreq, highFreq, duration, vertScale);		
		
	StartInstrument( svfilter, NULL );
	StartInstrument( out_envfollower, NULL );
	StartInstrument( low_envfollower, NULL );
	StartInstrument( band_envfollower, NULL );
	StartInstrument( high_envfollower, NULL );

	printf("Gathering data.  Please wait ...\n\n");		
	
	diffFreq	= highFreq - lowFreq;
	freqCounter	= lowFreq;
	
	for ( i = 0; i < diffFreq; i++ )
	{
		StartInstrumentVA( sineWave, AF_TAG_FREQUENCY_FP, ConvertFP_TagData(freqCounter), TAG_END);
		SleepUntilTime( sleepCue, GetAudioTime() + duration );

		ReadProbe( outTestProbe, &theOutputValue );
		outSampleValues[i]	= theOutputValue;
		
		ReadProbe( lowTestProbe, &theOutputValue );
		lowSampleValues[i]	= theOutputValue;
		
		ReadProbe( bandTestProbe, &theOutputValue );
		bandSampleValues[i]	= theOutputValue;
		
		ReadProbe( highTestProbe, &theOutputValue );
		highSampleValues[i]	= theOutputValue;
		
		freqCounter	+= 1.0;
		
	}	

	StopInstrument( sineWave, NULL );
	StopInstrument( svfilter, NULL );
	StopInstrument( out_envfollower, NULL );
	StopInstrument( low_envfollower, NULL );
	StopInstrument( band_envfollower, NULL );
	StopInstrument( high_envfollower, NULL );


	/* Draw white rectangle. */
	pgSetColor( pgcon, 1.0, 1.0, 1.0 );

	/*	Draw the lowpass scaled output to the screen. Scale the x values so that they will map to 0.0 - 1.0
		screen.  Invert the y values so that graph will look right side up.
	*/
	pgMoveTo( pgcon, 0, 1.0 );
	for ( i = 0; i < diffFreq; i++ )
	{
		scaledValue	= (float32)i/(float32)diffFreq;
		
		switch( whichOutput )
		{
			case	SCALED_LOWPASS_OUTPUT	:
												flipValue	= 1.0 - outSampleValues[i]*vertScale;
												break;
			case	LOWPASS_OUTPUT			:
												flipValue	= 1.0 - lowSampleValues[i]*vertScale;
												break;
			case	BANDPASS_OUTPUT			:
												flipValue	= 1.0 - bandSampleValues[i]*vertScale;
												break;
			case	HIGHPASS_OUTPUT			:
												flipValue	= 1.0 - highSampleValues[i]*vertScale;
												break;
			default							:
												flipValue	= 1.0;
												break;
		}
		pgDrawTo( pgcon, scaledValue,flipValue );
	}
	
	pgSetColor( pgcon, .4, .7, 1.0 );

	/*	Draw a line representing the critical frequency */
	pgMoveTo( pgcon, ((float32)(freqArg-lowFreq))/(float32)diffFreq, 1.0 );
	pgDrawTo( pgcon, ((float32)(freqArg-lowFreq))/(float32)diffFreq, 0.0 );

	/* Draw in pink. */
	pgSetColor( pgcon, .7, .5, .8 );
	pgMoveTo( pgcon, .4, .9 );
	
	switch( whichOutput )
	{
		case	SCALED_LOWPASS_OUTPUT	:
											pgDrawText( pgcon, "Output" );
											break;
		case	LOWPASS_OUTPUT			:
											pgDrawText( pgcon, "LowPass" );
											break;
		case	BANDPASS_OUTPUT			:
											pgDrawText( pgcon, "BandPass" );
											break;
		case	HIGHPASS_OUTPUT			:
											pgDrawText( pgcon, "HighPass" );
											break;
	}

	if ( doPrint != 0 )
	{
		for ( i = 0; i < diffFreq; i++ )
			printf("%3d Hz. out  = %f\t\tlow  = %f\t\tband = %f\t\thigh = %f\n",	i+lowFreq, outSampleValues[i], lowSampleValues[i],
																					bandSampleValues[i], highSampleValues[i] 			);
	}

	printf("\nPress the Up or down button on the control pad to scan through the different frequency response plots\nPress the X button to exit program\n\n");
	
	DoThisButton();
	
	/* Clear the screen before exiting */
	pgClearDisplay( pgcon, 0.0, 0.0, 0.0 );

	DeleteAttachment( theAttachment );
	UnloadSample( theSample );
	UnloadInstrument( svfilter );
	UnloadInstrument( sineWave );
	UnloadInstrument( out_envfollower );
	UnloadInstrument( low_envfollower );
	UnloadInstrument( band_envfollower );
	UnloadInstrument( high_envfollower );

	DeleteCue( theCue );
	DeleteCue( sleepCue );

	DeleteProbe( outTestProbe );
	DeleteProbe( lowTestProbe );
	DeleteProbe( bandTestProbe );
	DeleteProbe( highTestProbe );
	
	KillEventUtility();

	pgDeleteDisplay( &pgcon );
	
	BeginTest(FT, "svfilter.dsp");
	LogResult(InteractiveVerify("Did the frequency response plot look correct. Did Test Pass?"), "");

	cleanup:
	theError	= CloseAudioFolio();
	CHECKRESULT(FT,theError,"CloseAudioFolio");


	/******************************************************************************/
	TermScaffold();
	
	return 1;
}


void WaitButton( uint32 button )
{
	ControlPadEventData cped;

	do
	{
		GetControlPad (1, TRUE, &cped);   /* Wait for button press. */
	}
	while( (cped.cped_ButtonBits & button) == 0 );
}

void DoThisButton()
{
	ControlPadEventData cped;
	float32				scaledValue, flipValue;
	int					i;

	do
	{
		GetControlPad (1, TRUE, &cped);   /* Wait for button press. */
		
		if ( (cped.cped_ButtonBits & ControlDown) == ControlDown )
		{
			if ( whichOutput > SCALED_LOWPASS_OUTPUT )
				whichOutput--;
				
			pgClearDisplay( pgcon, 0.0, 0.0, 0.0 );
			pgSetColor( pgcon, 1.0, 1.0, 1.0 );

			pgMoveTo( pgcon, 0, 1.0 );
			for ( i = 0; i < diffFreq; i++ )
			{
				scaledValue	= (float32)i/(float32)diffFreq;
		
				switch( whichOutput )
				{
					case	SCALED_LOWPASS_OUTPUT	:
														flipValue	= 1.0 - outSampleValues[i]*vertScale;
														break;
					case	LOWPASS_OUTPUT			:
														flipValue	= 1.0 - lowSampleValues[i]*vertScale;
														break;
					case	BANDPASS_OUTPUT			:
														flipValue	= 1.0 - bandSampleValues[i]*vertScale;
														break;
					case	HIGHPASS_OUTPUT			:
														flipValue	= 1.0 - highSampleValues[i]*vertScale;
														break;
					default							:
														flipValue	= 1.0;
														break;
				}
		
				pgDrawTo( pgcon, scaledValue, flipValue );
			}
	
			pgSetColor( pgcon, .4, .7, 1.0 );

			/*	Draw a line representing the critical frequency */
			pgMoveTo( pgcon, ((float32)(freqArg-lowFreq))/(float32)diffFreq, 1.0 );
			pgDrawTo( pgcon, ((float32)(freqArg-lowFreq))/(float32)diffFreq, 0.0 );

			/* Draw in pink. */
			pgSetColor( pgcon, .7, .5, .8 );
			pgMoveTo( pgcon, .4, .9 );
	
			switch( whichOutput )
			{
				case	SCALED_LOWPASS_OUTPUT	:
													pgDrawText( pgcon, "Output" );
													break;
				case	LOWPASS_OUTPUT			:
													pgDrawText( pgcon, "LowPass" );
													break;
				case	BANDPASS_OUTPUT			:
													pgDrawText( pgcon, "BandPass" );
													break;
				case	HIGHPASS_OUTPUT			:
													pgDrawText( pgcon, "HighPass" );
													break;
			}
	
		}
		else if ( (cped.cped_ButtonBits & ControlUp) == ControlUp )
		{
			if ( whichOutput < HIGHPASS_OUTPUT )
				whichOutput++;
				
			pgClearDisplay( pgcon, 0.0, 0.0, 0.0 );
			pgSetColor( pgcon, 1.0, 1.0, 1.0 );

			pgMoveTo( pgcon, 0, 1.0 );
			for ( i = 0; i < diffFreq; i++ )
			{
				scaledValue	= (float32)i/(float32)diffFreq;
		
				switch( whichOutput )
				{
					case	SCALED_LOWPASS_OUTPUT	:
														flipValue	= 1.0 - outSampleValues[i]*vertScale;
														break;
					case	LOWPASS_OUTPUT			:
														flipValue	= 1.0 - lowSampleValues[i]*vertScale;
														break;
					case	BANDPASS_OUTPUT			:
														flipValue	= 1.0 - bandSampleValues[i]*vertScale;
														break;
					case	HIGHPASS_OUTPUT			:
														flipValue	= 1.0 - highSampleValues[i]*vertScale;
														break;
					default							:
														flipValue	= 1.0;
														break;
				}
		
				pgDrawTo( pgcon, scaledValue, flipValue );
			}
	
			pgSetColor( pgcon, .4, .7, 1.0 );

			/*	Draw a line representing the critical frequency */
			pgMoveTo( pgcon, ((float32)(freqArg-lowFreq))/(float32)diffFreq, 1.0 );
			pgDrawTo( pgcon, ((float32)(freqArg-lowFreq))/(float32)diffFreq, 0.0 );

			/* Draw in pink. */
			pgSetColor( pgcon, .7, .5, .8 );
			pgMoveTo( pgcon, .4, .9 );
	
			switch( whichOutput )
			{
				case	SCALED_LOWPASS_OUTPUT	:
													pgDrawText( pgcon, "Output" );
													break;
				case	LOWPASS_OUTPUT			:
													pgDrawText( pgcon, "LowPass" );
													break;
				case	BANDPASS_OUTPUT			:
													pgDrawText( pgcon, "BandPass" );
													break;
				case	HIGHPASS_OUTPUT			:
													pgDrawText( pgcon, "HighPass" );
													break;
			}
	

		}
		
			
	}
	while( (cped.cped_ButtonBits & ControlX) == 0 );
}

