/* @(#) JugglerFT.c 96/07/29 1.3 */
/*	
 *	File:				JugglerFT.c
 *	Author:				Eric L. Hayes
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional tests for the juggler.  First of all, the Event structure is created. It
 *	contains the relative time in which the instrument will be played, the Item number of the instrument to be
 *	played, any tag arguments which should go with starting the instrument, and Boolean variable which represents
 *	whether the instrument should be stopped or started.  The event list for each sequence
 *	is initialized during the corresponding juggler's UserStartFunc_SeqX function.  When the juggler runs, and
 *	receives an event object, it either starts or stops the instrument indicated by the objects's Item number.
 *	Whether the instrument is stopped or started depends on the objects Boolean on value.  Objects in the Event
 *	list are arranged so that an object starting an instrument is followed by an object stopping the instrument.
 *
 *	This test creates a collection made of three sequences.  The first sequence
 *	plays four dsp sounds at various frequencies.  The second sequence is a mono 8 bit music sample.  The third
 *	sequence is a 16 bit stereo sample of people talking and eerie music in the background.  All sequences start at
 *	the same time.  The second sequence is played twice within the collection.  The collection itself is played twice.
 *	This test program uses a minimum number of TestComment statements because the slowness of the I/O will throw
 *	off the timing of the juggler.
 *
 *	Uses the following calls:
 *
 *	OpenAudioFolio	 	InitJuggler			TermJuggler				LoadInstrument			
 *	LoadSample			CreateKnob			StopInstrument			CreateAttachment	
 *	MakeMixerSpec		StartInstrument		CreateMixerTemplate		ConnectInstrumentParts
 *	SetKnob				UnloadInstrument	UnloadSample			CreateInstrument
 *	DeleteAttachment	GetAudioTime		StartObject				StopObject
 *	CalcMixerGainPart	DestroyObject		DisconnectInstrumentParts
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */



#include <audio/audio.h>
#include <audio/music.h>
#include <audio/parse_aiff.h>
#include <audio/juggler.h>
#include <kernel/operror.h>
#include <stdio.h>

#include <TestMacros.h>
#include <tests/3do_scaf.h>


#define	PLAYTIME_1		240
#define	PLAYTIME_2		PLAYTIME_1 + 720
#define	PLAYTIME_3		PLAYTIME_2 + 240
#define	PLAYTIME_4		PLAYTIME_3 + 120
#define	PLAYTIME_5		PLAYTIME_4 + 240
#define	PLAYTIME_6		PLAYTIME_5 + 360
#define	PLAYTIME_7		PLAYTIME_6 + 240
#define	PLAYTIME_8		PLAYTIME_7 + 240
#define	PLAYTIME_9		PLAYTIME_8 + 240
#define	PLAYTIME_10		PLAYTIME_9 + 1440


/*	Define global variables	*/
Collection		*col	= NULL;
Item			gNoiseDspInstrumentItem,
				gSawtoothDspInstrumentItem,
				gSquareDspInstrumentItem,
				gTriangleDspInstrumentItem,
				gMixerInstrumentItem,
				gSampleFileItem,
				gSampleFileItem2,
				gSampleInstrumentItem,
				gSampleInstrumentItem2,
				gAttachmentItem,
				gAttachmentItem2;

Sequence		*seq1 = NULL, *seq2 = NULL, *seq3 = NULL;


		
/* Define what each event in the event list will look like	*/
typedef struct
{
	Time	playTime;
	Item	dspInstrumentItem;
	TagArg	Tags[6];
	Boolean	on;
} TestEvent;


TestEvent TestData[8], TestData_2[2], TestData_3[2];

/* Define callback functions for juggler. 						*/
/*	This function will be called when the sequence is started	*/ 
int32 UserStartFunc_Seq1 ( Jugglee *Self, Time StartTime )
{
	TOUCH(Self);
        TOUCH(StartTime);
	/*	Assign values to the Event List					*/
	TestData[0].playTime			= PLAYTIME_1;
	TestData[0].dspInstrumentItem	= gNoiseDspInstrumentItem;
	TestData[0].Tags[0].ta_Tag		= AF_TAG_AMPLITUDE_FP;
	TestData[0].Tags[0].ta_Arg		= ConvertFP_TagData(.2);
	TestData[0].Tags[1].ta_Tag		= TAG_END;
	TestData[0].on					= 1;

	TestData[1].playTime			= PLAYTIME_2;
	TestData[1].dspInstrumentItem	= gNoiseDspInstrumentItem;
	TestData[1].Tags[0].ta_Tag		= TAG_END;
	TestData[1].on					= 0;

	TestData[2].playTime			= PLAYTIME_3;
	TestData[2].dspInstrumentItem	= gSawtoothDspInstrumentItem;
	TestData[2].Tags[0].ta_Tag		= AF_TAG_FREQUENCY_FP;
	TestData[2].Tags[0].ta_Arg		= ConvertFP_TagData(2000.0);
	TestData[2].Tags[1].ta_Tag		= AF_TAG_AMPLITUDE_FP;
	TestData[2].Tags[1].ta_Arg		= ConvertFP_TagData(.1);
	TestData[2].Tags[2].ta_Tag		= TAG_END;
	TestData[2].on					= 1;

	TestData[3].playTime			= PLAYTIME_4;
	TestData[3].dspInstrumentItem	= gSawtoothDspInstrumentItem;
	TestData[3].Tags[0].ta_Tag		= TAG_END;
	TestData[3].on					= 0;

	TestData[4].playTime			= PLAYTIME_5;
	TestData[4].dspInstrumentItem	= gSquareDspInstrumentItem;
	TestData[4].Tags[0].ta_Tag		= AF_TAG_FREQUENCY_FP;
	TestData[4].Tags[0].ta_Arg		= ConvertFP_TagData(100.0);
	TestData[4].Tags[1].ta_Tag		= AF_TAG_AMPLITUDE_FP;
	TestData[4].Tags[1].ta_Arg		= ConvertFP_TagData(.3);
	TestData[4].Tags[2].ta_Tag		= TAG_END;
	TestData[4].on					= 1;

	TestData[5].playTime			= PLAYTIME_6;
	TestData[5].dspInstrumentItem	= gSquareDspInstrumentItem;
	TestData[5].Tags[0].ta_Tag		= TAG_END;
	TestData[5].on					= 0;

	TestData[6].playTime			= PLAYTIME_7;
	TestData[6].dspInstrumentItem	= gTriangleDspInstrumentItem;
	TestData[6].Tags[0].ta_Tag		= AF_TAG_FREQUENCY_FP;
	TestData[6].Tags[0].ta_Arg		= ConvertFP_TagData(1000.0);
	TestData[6].Tags[1].ta_Tag		= AF_TAG_AMPLITUDE_FP;
	TestData[6].Tags[1].ta_Arg		= ConvertFP_TagData(.3);
	TestData[6].Tags[2].ta_Tag		= TAG_END;
	TestData[6].on					= 1;

	TestData[7].playTime			= PLAYTIME_8;
	TestData[7].dspInstrumentItem	= gTriangleDspInstrumentItem;
	TestData[7].Tags[0].ta_Tag		= TAG_END;
	TestData[7].on					= 0;

	return 0;
}

int32 UserInterpFunc_Seq1 ( Jugglee *Self, TestEvent *te, void *UserContext)
{
	TOUCH(Self);
        TOUCH(UserContext);
	if ( te->on == 1 )
	{
		StartInstrument( te->dspInstrumentItem, te->Tags );

	}
	else
	{

		StopInstrument( te->dspInstrumentItem, NULL );

	}

	return 0;
}


/* Define callback functions for juggler. 						*/
/*	This function will be called when the sequence is stopped	*/ 
int32 UserStopFunc_Seq1 ( Jugglee *Self, Time StartTime )
{
	TOUCH(Self);
        TOUCH(StartTime);
	return 0;
}


int32 UserStartFunc_Seq2 ( Jugglee *Self, Time StartTime )
{
	TOUCH(Self);
        TOUCH(StartTime);
	/*	Assign values to the Event List					*/
	TestData_2[0].playTime			= PLAYTIME_1;
	TestData_2[0].dspInstrumentItem	= gSampleInstrumentItem;
	TestData_2[0].Tags[0].ta_Tag	= AF_TAG_AMPLITUDE_FP;
	TestData_2[0].Tags[0].ta_Arg	= ConvertFP_TagData(.3);
	TestData_2[0].Tags[1].ta_Tag	= TAG_END;
	TestData_2[0].on				= 1;

	TestData_2[1].playTime			= PLAYTIME_8;
	TestData_2[1].dspInstrumentItem	= gSampleInstrumentItem;
	TestData_2[1].Tags[0].ta_Tag	= TAG_END;
	TestData_2[1].on				= 0;

	return 0;
}


int32 UserInterpFunc_Seq2 ( Jugglee *Self, TestEvent *te, void *UserContext)
{
	TOUCH(Self);
        TOUCH(UserContext);	
	
	if ( te->on == 1 )
	{
		StartInstrument( te->dspInstrumentItem, te->Tags );

	}
	else
	{

		StopInstrument( te->dspInstrumentItem, NULL );

	}

	return 0;
}


/* Define callback functions for juggler. 						*/
/*	This function will be called when the sequence is stopped	*/ 
int32 UserStopFunc_Seq2 ( Jugglee *Self, Time StartTime )
{
	TOUCH(Self);
        TOUCH(StartTime);
	return 0;
}

int32 UserStartFunc_Seq3 ( Jugglee *Self, Time StartTime )
{
	TOUCH(Self);
        TOUCH(StartTime);
	/*	Assign values to the Event List					*/
	TestData_3[0].playTime			= PLAYTIME_1;
	TestData_3[0].dspInstrumentItem	= gSampleInstrumentItem2;
	TestData_3[0].Tags[0].ta_Tag	= AF_TAG_AMPLITUDE_FP;
	TestData_3[0].Tags[0].ta_Arg	= ConvertFP_TagData(.3);
	TestData_3[0].Tags[1].ta_Tag	= TAG_END;
	TestData_3[0].on				= 1;

	TestData_3[1].playTime			= PLAYTIME_4;
	TestData_3[1].dspInstrumentItem	= gSampleInstrumentItem;
	TestData_3[1].Tags[0].ta_Tag	= TAG_END;
	TestData_3[1].on				= 0;

	return 0;
}


int32 UserInterpFunc_Seq3 ( Jugglee *Self, TestEvent *te, void *UserContext)
{
	TOUCH(Self);
	TOUCH(UserContext);	
	
	if ( te->on == 1 )
	{
		StartInstrument( te->dspInstrumentItem, te->Tags );

	}
	else
	{

		StopInstrument( te->dspInstrumentItem, NULL );

	}

	return 0;
}


/* Define callback functions for juggler. 						*/
/*	This function will be called when the sequence is stopped	*/ 
int32 UserStopFunc_Seq3 ( Jugglee *Self, Time StartTime )
{
	TOUCH(Self);
        TOUCH(StartTime);
	return 0;
}

int32 UserStartFunc_Col ( Jugglee *Self, Time StartTime )
{
	TOUCH(Self);
        TOUCH(StartTime);
	return 0;
}


int32 UserStopFunc_Col ( Jugglee *Self, Time StartTime )
{
	TOUCH(Self);
        TOUCH(StartTime);
	return 0;
}


int32 UserRepeatFunc_Col ( Jugglee *Self, Time StartTime )
{
	TagArg	Tags[2];
	TOUCH(Self);
	TOUCH(StartTime);
	
	RemoveNthFromObject( col, 1 );	
        
	Tags[0].ta_Tag = JGLR_TAG_MANY;
    Tags[0].ta_Arg = (TagData) 4;
	Tags[1].ta_Tag = 0;

	SetObjectInfo( seq1, Tags );
	
	return 0;
}



int main(int argc, char ** argv)
{
	Err			theError;
	int32		Result, NextSignals, theMixerKnobPartNum, i;
	Item		theCue, theMixerTemplate, theMixerKnob;
	MixerSpec	theMixerSpec;
	Time 		CurTime, NextTime;



	InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	
	printf("\nThis file contains the tests the juggler.  It creates a collection made of three sequewnces.  The first sequence\n");
	printf("plays four dsp sounds at various frequencies.  The second sequence is a mono 8 bit music sample.  The third\n");
	printf("sequence is a 16 bit stereo sample of people talking and eerie music in the background.  All sequences start at\n");
	printf("the same time.  The third sequence is played twice within the collection.  The collection itself is played twice\n");
	printf("The collection's Repeat function removes the the second sequence from the collection and reduces the number of\n");
	printf("events(sounds) the first sequence plays by half.\n");
	printf("This test program uses a minimum number of TestComment statements because the slowness of the IO will throw\n");
	printf("off the timing of the juggler.\n\n");


	theError = OpenAudioFolio();
	CHECKRESULT_SETUP(FT,theError,"OpenAudioFolio",cleanup);

	theError = InitJuggler();
	CHECKRESULT_SETUP(FT,theError,"InitJuggler",cleanup);

	theCue	= CreateCue( NULL );

/* Create and use objects of the Class 'SequenceClass'. */
	seq1 = (Sequence *) CreateObject( &SequenceClass );
	seq2 = (Sequence *) CreateObject( &SequenceClass );
	seq3 = (Sequence *) CreateObject( &SequenceClass );

	col = (Collection *) CreateObject( &CollectionClass );

/* Internal data is accessed by using SetObjectInfo or GetObjectInfo
   This allows us to control access to internal data, enforce ranges,
   filter illegal values, etc.
*/
    {
        TagArg Tags[8];

        Tags[0].ta_Tag = JGLR_TAG_START_FUNCTION;
        Tags[0].ta_Arg = (TagData) UserStartFunc_Seq1;
        Tags[1].ta_Tag = JGLR_TAG_STOP_FUNCTION;
        Tags[1].ta_Arg = (TagData) UserStopFunc_Seq1;
        Tags[2].ta_Tag = JGLR_TAG_INTERPRETER_FUNCTION;
        Tags[2].ta_Arg = (TagData) UserInterpFunc_Seq1;
        Tags[3].ta_Tag = JGLR_TAG_MAX;
        Tags[3].ta_Arg = (TagData) 8;
        Tags[4].ta_Tag = JGLR_TAG_MANY;
        Tags[4].ta_Arg = (TagData) 8;
        Tags[5].ta_Tag = JGLR_TAG_EVENTS;
        Tags[5].ta_Arg = (TagData) &TestData[0];
        Tags[6].ta_Tag = JGLR_TAG_EVENT_SIZE;
        Tags[6].ta_Arg = (TagData) sizeof(TestEvent);
        Tags[7].ta_Tag =  0;

        SetObjectInfo(seq1, Tags);

        Tags[0].ta_Tag = JGLR_TAG_START_FUNCTION;
        Tags[0].ta_Arg = (TagData) UserStartFunc_Seq2;
        Tags[1].ta_Tag = JGLR_TAG_STOP_FUNCTION;
        Tags[1].ta_Arg = (TagData) UserStopFunc_Seq2;
        Tags[2].ta_Tag = JGLR_TAG_INTERPRETER_FUNCTION;
        Tags[2].ta_Arg = (TagData) UserInterpFunc_Seq2;
        Tags[3].ta_Tag = JGLR_TAG_MAX;
        Tags[3].ta_Arg = (TagData) 2;
        Tags[4].ta_Tag = JGLR_TAG_MANY;
        Tags[4].ta_Arg = (TagData) 2;
        Tags[5].ta_Tag = JGLR_TAG_EVENTS;
        Tags[5].ta_Arg = (TagData) &TestData_2[0];
        Tags[6].ta_Tag = JGLR_TAG_EVENT_SIZE;
        Tags[6].ta_Arg = (TagData) sizeof(TestEvent);
        Tags[7].ta_Tag =  0;

        SetObjectInfo(seq2, Tags);

        Tags[0].ta_Tag = JGLR_TAG_START_FUNCTION;
        Tags[0].ta_Arg = (TagData) UserStartFunc_Seq3;
        Tags[1].ta_Tag = JGLR_TAG_STOP_FUNCTION;
        Tags[1].ta_Arg = (TagData) UserStopFunc_Seq3;
        Tags[2].ta_Tag = JGLR_TAG_INTERPRETER_FUNCTION;
        Tags[2].ta_Arg = (TagData) UserInterpFunc_Seq3;
        Tags[3].ta_Tag = JGLR_TAG_MAX;
        Tags[3].ta_Arg = (TagData) 2;
        Tags[4].ta_Tag = JGLR_TAG_MANY;
        Tags[4].ta_Arg = (TagData) 2;
        Tags[5].ta_Tag = JGLR_TAG_EVENTS;
        Tags[5].ta_Arg = (TagData) &TestData_3[0];
        Tags[6].ta_Tag = JGLR_TAG_EVENT_SIZE;
        Tags[6].ta_Arg = (TagData) sizeof(TestEvent);
        Tags[7].ta_Tag =  0;

        SetObjectInfo(seq3, Tags);

        Tags[0].ta_Tag = JGLR_TAG_START_FUNCTION;
        Tags[0].ta_Arg = (TagData) UserStartFunc_Col;
        Tags[1].ta_Tag = JGLR_TAG_STOP_FUNCTION;
        Tags[1].ta_Arg = (TagData) UserStopFunc_Col;
        Tags[2].ta_Tag = JGLR_TAG_REPEAT_FUNCTION;
        Tags[2].ta_Arg = (TagData) UserRepeatFunc_Col;
		Tags[3].ta_Tag =  0;
		
		SetObjectInfo(col, Tags);
    }


	col->Class->Add(col, seq1, 1);
	col->Class->Add(col, seq2, 2);
	col->Class->Add(col, seq3, 1);

	PrintObject( seq1 );
	PrintObject( seq2 );
	PrintObject( seq3 );
	PrintObject( col );
	
    {
        TagArg Tags[3];
   
		Tags[0].ta_Tag = JGLR_TAG_MAX;
		Tags[1].ta_Tag = JGLR_TAG_MANY;
		Tags[2].ta_Tag = 0;

		GetObjectInfo( seq1, Tags );
		BeginTest(FT, "Juggler");
		LogComment("GetObjectInfo( Sequence 1 ): JGLR_TAG_MAX = %d, JGLR_TAG_MANY = %d", Tags[0].ta_Arg, Tags[1].ta_Arg );
		LogResult(InteractiveVerify("You should see information on the three sequences and one collection, and tag information for the first sequence. Did Test Pass?"), "");

	}


	printf("\nLoading all instruments and samples...\n");

	/*	Load all of the instruments which will be played in the sequence #1	*/
	gNoiseDspInstrumentItem		= LoadInstrument( "noise.dsp", 1, 100 );
	CHECKRESULT_SETUP(FT,gNoiseDspInstrumentItem,"LoadInstrument(noise.dsp)",cleanup);

	gSawtoothDspInstrumentItem	= LoadInstrument( "sawtooth.dsp", 1, 100 );
	CHECKRESULT_SETUP(FT,gSawtoothDspInstrumentItem,"LoadInstrument(sawtooth.dsp)",cleanup);

	gSquareDspInstrumentItem	= LoadInstrument( "square.dsp", 1, 100 );
	CHECKRESULT_SETUP(FT,gSquareDspInstrumentItem,"LoadInstrument(square.dsp)",cleanup);

	gTriangleDspInstrumentItem	= LoadInstrument( "triangle.dsp", 1, 100 );
	CHECKRESULT_SETUP(FT,gTriangleDspInstrumentItem,"LoadInstrument(triangle.dsp)",cleanup);



	/*	Load all of the instruments which will be played in the sequence #2	*/
	gSampleInstrumentItem		= LoadInstrument( "sampler_adp4_v1.dsp", 1, 100 );
	CHECKRESULT_SETUP(FT,gSampleInstrumentItem,"LoadInstrument(sampler_adp4_v1.dsp)",cleanup);

	gSampleFileItem				= LoadSample("22K_16B_M_ADP4_MUSIC.AIFF");
	CHECKRESULT_SETUP(FT,gSampleFileItem,"LoadSample(22K_16B_M_ADP4_MUSIC.AIFF)",cleanup);

	gAttachmentItem				= CreateAttachment( gSampleInstrumentItem, gSampleFileItem, NULL );
	CHECKRESULT_SETUP(FT,gAttachmentItem,"CreateAttachment",cleanup);
	

	/*	Load all of the instruments which will be played in the sequence #3	*/
	gSampleInstrumentItem2		= LoadInstrument( "sampler_16_f2.dsp", 1, 100 );
	CHECKRESULT_SETUP(FT,gSampleInstrumentItem2,"LoadInstrument(sampler_16_f2.dsp)",cleanup);

	gSampleFileItem2			= LoadSample("44K_16B_S.AIFF");
	CHECKRESULT_SETUP(FT,gSampleFileItem2,"LoadSample(44K_16B_S.AIFF)",cleanup);

	gAttachmentItem2			= CreateAttachment( gSampleInstrumentItem2, gSampleFileItem2, NULL );
	CHECKRESULT_SETUP(FT,gAttachmentItem2,"CreateAttachment",cleanup);
	
	

	/*	Create Mixer	*/
	theMixerSpec		= MakeMixerSpec(7, 2, AF_F_MIXER_WITH_LINE_OUT);
		
	theMixerTemplate		= CreateMixerTemplate( theMixerSpec, NULL );
	gMixerInstrumentItem	= CreateInstrument( theMixerTemplate, NULL );
	CHECKRESULT_SETUP(FT,gMixerInstrumentItem,"CreateInstrument(theMixer)",cleanup);

	theMixerKnob			= CreateKnob( gMixerInstrumentItem, "Gain", NULL );
	CHECKRESULT_SETUP(FT,theMixerKnob,"CreateKnob(Gain)",cleanup);

	/*	Set the volume to all knobs to Maximum		*/
	for ( i = 0; i <= 6; i++ )
	{

		/*	Set the gain between the input and the output			*/
		theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, i, 0 );
		SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );
		
		theMixerKnobPartNum	= CalcMixerGainPart( theMixerSpec, i, 1 );
		SetKnobPart( theMixerKnob, theMixerKnobPartNum, 1.0 );
	}
			
	/*	Connect all of the instruments to the mixer	*/
	
	theError	= ConnectInstrumentParts(	gNoiseDspInstrumentItem, "Output", 0, 
											gMixerInstrumentItem, "Input", 0 		);
	CHECKRESULT_SETUP(FT,theError,"ConnectInstrumentParts(Noise to Mixer)",cleanup);
							
	theError	= ConnectInstrumentParts(	gSawtoothDspInstrumentItem, "Output", 0, 
											gMixerInstrumentItem, "Input", 1 		);	
	CHECKRESULT_SETUP(FT,theError,"ConnectInstrumentParts(Sawtooth to Mixer)",cleanup);

	theError	= ConnectInstrumentParts(	gSquareDspInstrumentItem, "Output", 0, 
											gMixerInstrumentItem, "Input", 2 		);	
	CHECKRESULT_SETUP(FT,theError,"ConnectInstrumentParts(Square to Mixer)",cleanup);

	theError	= ConnectInstrumentParts(	gTriangleDspInstrumentItem, "Output", 0, 
											gMixerInstrumentItem, "Input", 3 		);	
	CHECKRESULT_SETUP(FT,theError,"ConnectInstrumentParts(Triangle to Mixer)",cleanup);


	theError	= ConnectInstrumentParts(	gSampleInstrumentItem, "Output", 0, 
											gMixerInstrumentItem, "Input", 4 		);	
	CHECKRESULT_SETUP(FT,theError,"ConnectInstrumentParts(Music Sample to Mixer)",cleanup);

	theError	= ConnectInstrumentParts(	gSampleInstrumentItem2, "Output", 0, 
											gMixerInstrumentItem, "Input", 5 		);	
	CHECKRESULT_SETUP(FT,theError,"ConnectInstrumentParts(Voice Sample(Left) to Mixer)",cleanup);

	theError	= ConnectInstrumentParts(	gSampleInstrumentItem2, "Output", 1, 
											gMixerInstrumentItem, "Input", 6 		);	
	CHECKRESULT_SETUP(FT,theError,"ConnectInstrumentParts(Voice Sample(Right) to Mixer)",cleanup);


	StartInstrument( gMixerInstrumentItem, NULL );

	printf("\nStarting Collection...\n\n");
	/*	Tell the Juggler to start the sequences at the given times.
		Nothing will happen until you call BumpJuggler() */

	CurTime = GetAudioTime(); 
	StartObject(col, CurTime, 2, NULL);

	do
	{
		CurTime = GetAudioTime(); 

		Result = BumpJuggler( CurTime, &NextTime, 0, &NextSignals );
		
		if (Result == 0)
			SleepUntilTime(theCue, NextTime);
			
	} while ( Result == 0);  /* Will be > 0 when done, < 0 if error */

	CurTime = GetAudioTime(); 
	StopObject(col, CurTime + 10);

	LogComment("Play Juggler Sequences Test is Over.\n");
	
	BeginTest(FT, "Juggler");
	LogResult(InteractiveVerify("You should have heard the collection of sequences being played. Did Test Pass?"), "");

	
	/*	Disconnect all instruments from the mixer	*/
	
	DisconnectInstrumentParts( gMixerInstrumentItem, "Input", 0 );	
	DisconnectInstrumentParts( gMixerInstrumentItem, "Input", 1 );	
	DisconnectInstrumentParts( gMixerInstrumentItem, "Input", 2 );	
	DisconnectInstrumentParts( gMixerInstrumentItem, "Input", 3 );	
	DisconnectInstrumentParts( gMixerInstrumentItem, "Input", 4 );	
	DisconnectInstrumentParts( gMixerInstrumentItem, "Input", 5 );	
	DisconnectInstrumentParts( gMixerInstrumentItem, "Input", 6 );	

	/*	Load all of the instruments which will be played in the sequence	*/
	UnloadInstrument( gNoiseDspInstrumentItem );
	UnloadInstrument( gSawtoothDspInstrumentItem );
	UnloadInstrument( gSquareDspInstrumentItem );
	UnloadInstrument( gTriangleDspInstrumentItem );
	UnloadInstrument( gMixerInstrumentItem );
	DeleteAttachment( gAttachmentItem );
	DeleteAttachment( gAttachmentItem2 );
	UnloadSample( gSampleFileItem );
	UnloadSample( gSampleFileItem2 );
	UnloadInstrument( gSampleInstrumentItem );
	UnloadInstrument( gSampleInstrumentItem2 );


	DeleteCue( theCue );

cleanup:
	DestroyObject( (COBObject *) seq1 );
	DestroyObject( (COBObject *) seq2 );
	DestroyObject( (COBObject *) seq3 );
	DestroyObject( (COBObject *) col );

	TermJuggler();
	
	theError	= CloseAudioFolio();
	CHECKRESULT(FT, theError,"CloseAudioFolio");

	/******************************************************************************/
	TermScaffold();

	return 0;
}
