/* @(#) SampleFT.c 96/07/25 1.3 */
/*	
 *	File:				SampleFT.c
 *	Author:				Phil Burk
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional test for creating a sample.  This test is an old Opera
 *	test which was converted for use with M2.  It creates a stereo sample of a sawtooth wave
 *	in C, prints debug information for the sample and then plays the sample for 30 seconds.
 *
 *	Audio Functions called:
 *
 *		CloseAudioFolio		CreateCue			LoadInstrument		LoadSample			
 *		OpenAudioFolio		StartInstrument		StopInstrument		SleepUntilTime	
 *		DeleteCue			ReleaseInstrument	UnloadInstrument	UnloadSample
 *		ConnectInstrumentParts
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 */


/* Include this when using the Audio Folio */
#include <audio/audio.h>
#include <audio/parse_aiff.h>
#include <audio/soundspooler.h>

#include <file/filefunctions.h>
#include <file/filesystem.h>
#include <kernel/debug.h>
#include <kernel/operror.h>
#include <kernel/mem.h>
#include <kernel/types.h>
#include <stdio.h>
#include <stdlib.h>

#include <tests/3do_scaf.h>
#include <TestMacros.h>
#include <TestUtilities.h>

#define NUMCHANNELS (2)
#define NUMFRAMES (32*1024)
#define SAMPSIZE (sizeof(int16)*NUMFRAMES*NUMCHANNELS)

#define	PRT(x)	{ LogComment x; }
#define	ERR(x)	PRT(x)
#define	DBUG(x)	PRT(x)


int main(int argc, char *argv[])
{
	Err		Result;
	Item 	SamplerIns;
	Item 	SampleItem;
	Item 	OutputIns;
	Item 	Attachment, theCue;
	float32 freq;
	int16 	*Data;
	TagArg 	Tags[6];

/*	PRT(( "Usage: %s <samplefile> <speed>\n", argv[0])); */


	argc	= InitScaffold(AUDIO, argc, argv);					/* initializes this program */
	

/* Initialize audio, return if error. */ 
	Result = OpenAudioFolio();
	CHECKRESULT_SETUP(FT, Result, "OpenAudioFolio", cleanup);
	
	theCue	= CreateCue( NULL );
	CHECKRESULT_SETUP(FT,theCue,"CreateCue",cleanup);

/* Use directout instead of mixer. */
	OutputIns = LoadInstrument("line_out.dsp",  0, 0);
	CHECKRESULT(FT, OutputIns,"LoadInstrument");
	StartInstrument( OutputIns, NULL );

/* Load Sampler instrument */
	SamplerIns = LoadInstrument("sampler_16_f2.dsp",  0, 100);
	CHECKRESULT(FT, SamplerIns,"LoadInstrument");

/* Make it stereo */

	Data = (int16 *) AllocMem( SAMPSIZE, MEMTYPE_NORMAL|MEMTYPE_TRACKSIZE );
	if( Data == NULL )
	{
		ERR(( "Not enough memory for sample .\n"));
		goto cleanup;
	}

	/* Fill sample with a sawtooth at a high and low frequency in different channels. */
	{
		int32 i;
		int16 *tData = Data;

		for (i=0; i<NUMFRAMES; i++)
		{
			*tData++ = (int16) ((((i << 8) & 0xFFFF) * i) / NUMFRAMES);
			*tData++ = (int16) ((((i << 6) & 0xFFFF) * (NUMFRAMES - i)) / NUMFRAMES);  /* Lower frequency. */
		}
	}

	/* Create sample Item after writing data to take advantage of free WriteBackDCache() */
	SampleItem = CreateSampleVA(	AF_TAG_ADDRESS, (int16 *)Data, AF_TAG_FRAMES, NUMFRAMES, AF_TAG_CHANNELS, 2,
									AF_TAG_SUSTAINBEGIN, 0, AF_TAG_SUSTAINEND, (NUMFRAMES-1), AF_TAG_AUTO_FREE_DATA,
									TRUE, TAG_END);
	CHECKRESULT(FT, SampleItem,"CreateSample");
	
	if (argc > 2)
	{
		freq = atof32( argv[2] );
	}
	else
	{
		freq =.5;
	}
	DBUG(( "Frequency = %f\n", freq));

/* Look at sample information. */
	DebugSample(SampleItem);
	
/* Connect Sampler to Mixer */
	Result = ConnectInstrumentParts (SamplerIns, "Output", 0, OutputIns, "Input", 0);
	CHECKRESULT(FT, Result,"ConnectInstruments");
	Result = ConnectInstrumentParts (SamplerIns, "Output", 1, OutputIns, "Input", 1);
	CHECKRESULT(FT, Result,"ConnectInstruments");
		
/* Attach the sample to the instrument. */
	Attachment = CreateAttachment(SamplerIns, SampleItem, NULL);
	CHECKRESULT(FT, Attachment,"CreateAttachment");
	

/* Play one note a long time. */
#define DURATION 30
DBUG(( "Play sample for %d seconds.", DURATION));
	Tags[0].ta_Tag = AF_TAG_AMPLITUDE_FP;
	Tags[0].ta_Arg = ConvertFP_TagData(1.0);
	Tags[1].ta_Tag = AF_TAG_RATE_FP;
	Tags[1].ta_Arg =  ConvertFP_TagData(freq);
	Tags[2].ta_Tag = TAG_END;
	Result = StartInstrument( SamplerIns, &Tags[0] );
	CHECKRESULT(FT, Result,"StartInstrument");
	SleepUntilTime( theCue, GetAudioTime() + 60*DURATION );
	
	Result	= ReleaseInstrument( SamplerIns, NULL);
	CHECKRESULT(FT, Result,"ReleaseInstrument");
	SleepUntilTime( theCue, GetAudioTime() + 10*DURATION );

	BeginTest(FT, "Sample");
	LogResult(InteractiveVerify("You should have heard the sawtooth sample. Did Test Pass?"), "");

	DeleteCue( theCue );
	DeleteSample( SampleItem );
	UnloadInstrument( SamplerIns );
	UnloadInstrument( OutputIns );
	DeleteAttachment( Attachment );

cleanup:
/* The Audio Folio is immune to passing NULL values as Items. */

	Result	= CloseAudioFolio();
	CHECKRESULT(FT,Result,"CloseAudioFolio");
	
	TermScaffold();
	return((int) Result);
}
