/* @(#) Score_FT.c 96/08/28 1.5 */
/* 	
 *	File:				Score_FT.c
 *	Author:				Phil Burk
 *						Software Attic
 *	Creation Date:		
 *
 *	Last Modified by:	Eric L. Hayes
 *	Modification Date:	3/15/96
 *
 *	Version:			M2 2.0
 *
 *	This file contains the functional test for the Score MIDI functions.
 *	This test is an old Opera test which was converted for use with M2.  
 *
 *	Load three MIDI files and play them using Juggler and Score toolbox. 
 *	The three files can be played back in varying order using the A, B, 
 *	and C buttons on the Joypad.  
 *
 *	Holding down the left shift key causes the note durations to be lengthened.
 *	The right shift key causes them to be shortened.
 *
 *  A Button = play sequence1 when current sequence is finished
 *  B Button = play sequence2 when current sequence is finished
 *  C Button = play sequence3 when current sequence is finished
 * 
 *	Sequence1 is always first, and will loop until the stop button is pressed, or
 *  B or C button is used to change the order of playback.
 *
 *
 *	Audio Functions called:
 *
 *		CloseAudioFolio			OpenAudioFolio			InitJuggler					TermJuggler			
 *		CreateScoreContext		MakeMixerSpec			CreateScoreMixer			LoadPIMap	
 *		GetNthFromObject		CreateObject			SetObjectInfo				GetObjectInfo
 *		MFLoadCollection		GetAudioClockRate		MFUnloadCollection			DestroyObject
 *		UnloadPIMap				DeleteScoreMixer		DeleteScoreContext			StartObject
 *		CreateCue				GetCueSignal			GetAudioTime				StopObject
 *		SignalAtTime			BumpJuggler				AbortTimerCue				DeleteCue
 *		WaitSignal
 *
 *
 *	MIDI files Used (These files must be present in the current directory)
 *
 *	AGroove.pimap				AGroove1.mf				AGroove2.mf				AGroove3.mf
 *
 *
 *	Audio files Used (These files must be present in the current directory)
 *
 *	44K_16B_M.AIFF				Trumpet_44K_16B_M.aiff				CongaHC.aiff				CongaHO.aiff
 *	CongaLC.aiff				CongaLO.aiff
 *
 *
 *  Copyright (C) 1996, an unpublished work by The 3DO Company.
 *  All rights reserved. This material contains confidential
 *  information that is the property of The 3DO Company. Any
 *  unauthorized duplication, disclosure or use is prohibited.
 *  
 *
 */

#include <audio/audio.h>
#include <audio/music.h>
#include <audio/parse_aiff.h>
#include <audio/juggler.h>
#include <kernel/operror.h>
#include <kernel/task.h>
#include <misc/event.h>
#include <stdio.h>

#include <TestMacros.h>
#include <tests/3do_scaf.h>


#define  VERSION "V1.1"

#define	PRT(x)	{ printf x; }
#define	ERR(x)	PRT(x)
#define	DBUG(x)	/* PRT(x) */


/* Save memory by lowering MAXPROGRAMNUM to highest program number. */
#define MAXPROGRAMNUM 		(10)  
#define MAX_SCORE_VOICES  	(8)

/* Raise this value till the sound clips then drop it till you feel safe.
 * (MAXDSPAMPLITUDE/MAX_SCORE_VOICES) is guaranteed safe if all you are
 * doing is playing scores
 */
#define MIXER_AMPLITUDE 	(0x6000)

/* These are used to alter the playback speed while the scores are playing. 
 * The up scalar is frac16-speak for 1.25 and the down .85.  These values
 * are multiplied by the duration value for every note in a collection.
 */
#define NOTE_DUR_UP_SCALAR		((1 << 16) | 0x4000)
#define NOTE_DUR_DOWN_SCALAR	(0xB000)

/*****************************************************************
 	Function Prototypes.
 *****************************************************************/
static	Err 	StartItUp( void );
static	void 	ShutItDown( void );

static	ScoreContext *SetupScoreContext( char *mapfile );
static	Err 		 CleanupScoreContext( ScoreContext *scon );

static	Collection	*LoadScore( ScoreContext *scon, char *scorefile );
static	Err 	LoopScores( Collection *Coll1, Collection *Coll2, Collection *Coll3 );
static	Err 	UnloadScore( ScoreContext *scon, Collection *CollectionPtr );

static	Err 	ScaleSequenceTimes( Sequence *Self, frac16 Scalar );
static 	Err 	ScaleCollectionTimes( Collection *Coll, frac16 Scalar );

static 	void 	MyStopFunction( Collection *obj, uint32 TimeStopped );

/*****************************************************************
 * Static global which contains a pointer to the next collection 
 * to play...  there ought to be a better way to do this, but I 
 * can't think of one right now.
 *****************************************************************/
static	Collection	*NextCollPtr;


/**************************************Test Environment Globals****************/
/******************************************************************************/

/*****************************************************************
 	Open folios and initalize stuff.
 *****************************************************************/
static	Err StartItUp( void )
{
	int32	result;
	
	/* Initialize the EventBroker. */
	result = InitEventUtility(1, 0, LC_FocusListener);
	CHECKRESULT_SETUP(FT,result,"InitEventUtility",cleanup1);

	/* Initialize audio, return if error. */ 
	result = OpenAudioFolio();
	CHECKRESULT_SETUP(FT, result,"OpenAudioFolio",cleanup1);
	
	/* Required before playing scores. */
	InitJuggler();

	/* Everything was cool */
	return 0;
	
	cleanup1:
	
	return -1;

}

/*****************************************************************
 	Close folios and cleanup stuff.
 *****************************************************************/
static	void ShutItDown( void )
{
	Err		Result;
	
	
	TermJuggler();
	
	BeginTest(FT, "Score");
	LogResult(InteractiveVerify("You should have heard sounds as you press the A, B, and C buttons. Did Test Pass?"), "");

	Result	= CloseAudioFolio();
	CHECKRESULT(FT, Result,"CloseAudioFolio");
}

/******************************************************************
 * Create a ScoreContext, and load a PIMap from a text file.
 ******************************************************************/
static	ScoreContext *SetupScoreContext( char *mapfile )
{
	int32 			result;
	MixerSpec		theMixerSpec;
	ScoreContext 	*scon;
	
	/* Create a context for interpreting a MIDI score and tracking notes. */	
	scon = CreateScoreContext( MAXPROGRAMNUM );
	if ( scon == NULL )
		return NULL;
	
	theMixerSpec			= MakeMixerSpec(8, 2, AF_F_MIXER_WITH_AMPLITUDE | AF_F_MIXER_WITH_LINE_OUT);
	
	/* Specify a mixer to use for the score voices. */
	result = CreateScoreMixer( scon, theMixerSpec, MAX_SCORE_VOICES, MIXER_AMPLITUDE );
	CHECKRESULT_SETUP(FT,  result,"CreateScoreMixer",cleanup);

	/* Load Instrument Templates from disk and fill Program Instrument Map.
	 * As an alternative, you could use SetPIMapEntry() 
	 */
	result = LoadPIMap( scon, mapfile );
	CHECKRESULT_SETUP(FT, result, "LoadPIMap",cleanup);
	
	return scon;
	
cleanup:
	CleanupScoreContext( scon );
	return NULL;
}

/******************************************************************
 * Scale the timestamps in a Juggler Sequence.  This allows you
 * to play the score without having to muck with the audio timer.
 ******************************************************************/
static	Err ScaleSequenceTimes( Sequence *Self, frac16 Scalar )
{
	int32 	indx, eventSize, many;
	Time 	*eventPtr;
	TagArg 	Tags[4];
	
	/* Get info from Sequence using Tags.  */
	Tags[0].ta_Tag = JGLR_TAG_MANY;
	Tags[1].ta_Tag = JGLR_TAG_EVENTS;
	Tags[2].ta_Tag = JGLR_TAG_EVENT_SIZE;
	Tags[3].ta_Tag = TAG_END;
	GetObjectInfo( Self, Tags );
	
	many = (int32) Tags[0].ta_Arg;
	eventPtr = (Time *) Tags[1].ta_Arg;
	eventSize = (int32) Tags[2].ta_Arg;
	
	DBUG(("ScaleSequenceTimes: many = %d, eventPtr = %d, eventSize = %d\n",
		many, eventPtr, eventSize ));
		
	/* Scan through array and scale each timestamp. */
	indx = 0;
	while ( indx < many )
		{
		/* Scale time stamp. */
		*eventPtr = MulSF16( (*eventPtr), Scalar );
		
		/* Index to next Event */
		eventPtr = (Time *) (((char *)eventPtr) + eventSize);

		indx++;
		}

	return 0;
}

/******************************************************************
 * Scale the timestamps in all the sequences in a Juggler Collection.
 ******************************************************************/
static 	Err ScaleCollectionTimes( Collection *Coll, frac16 Scalar )
{
	int32		result = 0;
	int32		i, many;
	TagArg		Tags[2];
	Sequence	*seq;
	
	/* Query how many sequences it contains. */
	Tags[0].ta_Tag = JGLR_TAG_MANY;
	Tags[1].ta_Tag = TAG_END;
	GetObjectInfo( Coll, Tags );

	many = (int32) Tags[0].ta_Arg;

	for( i = 0; i < many; i++ )
		{
		result = GetNthFromObject( Coll, i, &seq );
		CHECKRESULT_SETUP(FT,  result, "GetNthFromCollection" ,cleanup);

		/* WARNING - assumes contents of Collection are sequences. */
		result = ScaleSequenceTimes( seq, Scalar );
		CHECKRESULT_SETUP(FT,  result, "ScaleSequenceTimes" ,cleanup);
		}
		
cleanup:
	return result;
}

/******************************************************************
 * Create a collection and load a score into it from a MIDI File.
 ******************************************************************/
static	Collection *LoadScore( ScoreContext *scon, char *scorefile )
{
	int32 		result;
	int32 		many;
	float32 	originalRate;
	TagArg 		Tags[2];
	frac16 		scalar;

	Collection *collectionPtr;

	MIDIFileParser MFParser;

	/* Create a collection and load MIDI File into it. */
	collectionPtr = (Collection *) CreateObject( &CollectionClass );
	if ( collectionPtr == NULL )
		{
		ERR(("LoadScore: Failure to create Collection\n"));
		goto cleanup;
		}
	
	/* Set context used to play collection. */
	Tags[0].ta_Tag = JGLR_TAG_CONTEXT;
	Tags[0].ta_Arg = (void *) scon;
	Tags[1].ta_Tag = TAG_END;
	SetObjectInfo( collectionPtr, Tags );

	/* Load it from a MIDIFile */
	result = MFLoadCollection ( &MFParser, scorefile, collectionPtr );
	if (result)
		{
		ERR(("Error loading MIDI File = $%x\n", result));
		goto cleanup;
		}
	
	/* Query how many sequences it contains. */
	Tags[0].ta_Tag = JGLR_TAG_MANY;
	GetObjectInfo(collectionPtr, Tags);

	many = (int32) Tags[0].ta_Arg;
	PRT(("Collection contains %d sequences.\n", many));
	
	/* Don't change the audio clock because that might interfere with other tasks. */
	originalRate = GetAudioClockRate( AF_GLOBAL_CLOCK, &originalRate);
	scalar = DivSF16( originalRate, MFParser.mfp_Rate );
	PRT(("Scale sequence times. mfp_Rate = 0x%x, scalar = 0x%x\n",
		MFParser.mfp_Rate, scalar));
		
	/* Traverse all of the times stamps in every sequence in the collection and 
	 * scale them so that we dont have to muck with the Audio clock.
	 */
	ScaleCollectionTimes( collectionPtr, scalar );
	
	return collectionPtr;
	
cleanup:
	UnloadScore( scon, collectionPtr );
	return NULL;
}

/******************************************************************
 * Unload the collection which frees the sequences, then
 * destroy collection.
 ******************************************************************/
static	Err UnloadScore( ScoreContext *scon, Collection *collectionPtr )
{
	TOUCH(scon);
	if (collectionPtr != NULL)
		{
		MFUnloadCollection( collectionPtr );
		DestroyObject( (COBObject *) collectionPtr );
		}

	return 0;
}

/******************************************************************
 * Unload the PIMap which frees the instruments and samples, then
 * delete the ScoreContext.
 ******************************************************************/
static	Err CleanupScoreContext( ScoreContext *scon )
{
	if( scon != NULL )
		{
		UnloadPIMap( scon );
		DeleteScoreMixer( scon );
		DeleteScoreContext( scon );
		}
	
	return 0;
}

/******************************************************************
 * Callback function to pass to the juggler.  This is used to
 * chain the collections.  When a collection stops, it's StopFunction
 * is called, which in this case starts the next collection in the 
 * chain using NextCollPtr.  This value is set in LoopScores() based on 
 * Joypad input.
 ******************************************************************/
static 	void MyStopFunction( Collection *obj, uint32 TimeStopped )
{
	TOUCH(obj);
	DBUG(("Time Stopped = %ld, Now = %ld\n", (long)TimeStopped, GetAudioTime()) );

/* 1.3 passed a bogus value for TimeStopped */
#ifdef WORKAROUND_1P3
	StartObject( NextCollPtr, GetAudioTime(), 1, NULL );
#else
	StartObject( NextCollPtr, TimeStopped, 1, NULL ); /* 940804 Use TimeStopped */
#endif
}

/*****************************************************************
 * Main playback loop to service juggler and deal with user input.
 *****************************************************************/
static	Err LoopScores( Collection *Coll1, Collection *Coll2, Collection *Coll3 )
{
	int32 	result = 0;								/* success or failure */
	int32	numReps;							/* times to repeat collection */
	int32 	nextSignals, cueSignal, signalsGot;	/* various signals */
	int32	timerPending;						/* are we waiting for cue signal */
	Item	myCue;								/* cue for Juggler timing */
	AudioTime curTime, nextTime;				/* for use with BumpJuggler() */

	ControlPadEventData 	jpState;			/* JoyPad context */

	/* Initalize things. */
	timerPending	= FALSE;
	numReps			= 1;
	nextSignals		= 0;
	

	/* Create a Cue for timing the score. */
	myCue = CreateCue( NULL );
	CHECKRESULT_SETUP(FT, myCue, "CreateCue",cleanup);

	cueSignal = GetCueSignal( myCue );

	/* Drive Juggler using Audio timer.
	 * Delay start by adding ticks to avoid stutter on startup. Optional. 
	 */
	nextTime = GetAudioTime() + 40;
	curTime = nextTime;
	
	DBUG(("PlayScore: Start at time %d\n", nextTime ));

	/* This tells the Juggler to process collection 1 when BumpJuggler() is
	 * later called.  
	 */
	StartObject( Coll1, nextTime, numReps, NULL );
	
	/* Set up Coll1 to repeat until something else happens */
	NextCollPtr = Coll1;
	
	do
		{
		/* Read current state of Control Pad. */
		GetControlPad (1, TRUE, &jpState);

		/* Forced Quit by hitting ControlX */
		if( (jpState.cped_ButtonBits & ControlX) == ControlX ) 
			{
			/* Need to figure out how to ask Juggler which object(s) are 
			 * currently executing
			 */
			StopObject( Coll1, curTime );
			StopObject( Coll2, curTime );
			StopObject( Coll3, curTime );
			break;
			}

		if ( (jpState.cped_ButtonBits & ControlA) == ControlA ) 
			{
			/* Point to collection 1 so when our user stop function is called
			 * it will setup this sequence to play when the currently executing
			 * sequence is done.
			 */
			NextCollPtr = Coll1;
			
			/* For fun, if the left shift button was down, scale the note
			 * durations up a bit; if right shift then down a bit.
			 */
			if ( (jpState.cped_ButtonBits & ControlLeftShift) == ControlLeftShift ) 
				ScaleCollectionTimes( Coll1, (frac16)NOTE_DUR_UP_SCALAR );
			else if ( (jpState.cped_ButtonBits & ControlRightShift) == ControlRightShift ) 
				ScaleCollectionTimes( Coll1, (frac16)NOTE_DUR_DOWN_SCALAR );
			}

		if ( (jpState.cped_ButtonBits & ControlB) == ControlB )  
			{
			/* Point to collection 2 so when our user stop function is called
			 * it will setup this sequence to play when the currently executing
			 * sequence is done.
			 */
			NextCollPtr = Coll2;
			
			/* For fun, if the left shift button was down, scale the note
			 * durations up a bit; if right shift then down a bit.
			 */
			if ( (jpState.cped_ButtonBits & ControlLeftShift) == ControlLeftShift ) 
				ScaleCollectionTimes( Coll2, (frac16)NOTE_DUR_UP_SCALAR );
			else if ( (jpState.cped_ButtonBits & ControlRightShift) == ControlRightShift ) 
				ScaleCollectionTimes( Coll2, (frac16)NOTE_DUR_DOWN_SCALAR );
			}

		if ( (jpState.cped_ButtonBits & ControlC) == ControlC ) 
			{
			/* Point to collection 3 so when our user stop function is called
			 * it will setup this sequence to play when the currently executing
			 * sequence is done.
			 */
			NextCollPtr = Coll3;
			
			/* For fun, if the left shift button was down, scale the note
			 * durations up a bit; if right shift then down a bit.
			 */
			if ( (jpState.cped_ButtonBits & ControlLeftShift) == ControlLeftShift )  
				ScaleCollectionTimes( Coll3, (frac16)NOTE_DUR_UP_SCALAR );
			else if ( (jpState.cped_ButtonBits & ControlRightShift) == ControlRightShift ) 
				ScaleCollectionTimes( Coll3, (frac16)NOTE_DUR_DOWN_SCALAR );
			}

		/* Request a timer wake up at the next event time. */
		if( !timerPending )
			{
			result = SignalAtTime( myCue, nextTime);
			if (result < 0) return result;
			timerPending = TRUE;
			}
	
		/* Wait for timer signal or signal(s) from BumpJuggler */
		signalsGot = WaitSignal( cueSignal | nextSignals );
		
		/* Did we wake up because of a timer signal? */
		if( signalsGot & cueSignal )
			{
			timerPending = FALSE;
			curTime = nextTime;
			}
		else
			{
			/* Get current time to inform Juggler. */
			curTime = GetAudioTime();
			}
		
		/* Tell Juggler to process any pending events, eg. Notes. result > 0 if done. */
		result = BumpJuggler( curTime, &nextTime, signalsGot, &nextSignals );
		
		} while ( result == 0 );

	AbortTimerCue( myCue );
	DeleteCue( myCue );
	
cleanup:
	return result;

}

/**************************************Test Environment Globals****************/
int32 gFlag;
Boolean gContinue;
char * gProgName = "MFLoopTest";	/* contains the name of the program */
char * gTestName = "MFLoopTest";		/* contains the name of the test    */
int32 gTestLevel = BREADTH;			/* contains what level of testing occurs in this test */
/******************************************************************************/
int main (int argc, char *argv[])
{
	TagArg	Tags[2];

	Collection *Coll1 = NULL;
	Collection *Coll2 = NULL;
	Collection *Coll3 = NULL;

	ScoreContext *scon = NULL;
	
/* Change these to your own file names. */
	char *mapfile = "/remote/AGroove.pimap";

	char *midifile1 = "/remote/AGroove1.mf";
	char *midifile2 = "/remote/AGroove2.mf";
	char *midifile3 = "/remote/AGroove3.mf";

/******************************************************************************/

	InitScaffold(AUDIO, argc, argv);					/* initializes this program */

	PRT(("Loop MIDI Files, %s\n", VERSION));

	/* Initalize folios and stuff */
	if ( StartItUp() )
		goto cleanup;

	/* Find and load the PIMap, this will cause all the samples and instruments
	 * we need to be set up in the score context 
	 */
	scon = SetupScoreContext( mapfile );
	if ( scon == NULL ) goto cleanup;
	
	/* Load and parse the MIDIfiles into juggler objects */
	Coll1 = LoadScore( scon, midifile1 );
	if ( Coll1 == NULL ) goto cleanup;

	Coll2 = LoadScore( scon, midifile2 );
	if ( Coll2 == NULL ) goto cleanup;

	Coll3 = LoadScore( scon, midifile3 );
	if ( Coll3 == NULL ) goto cleanup;

	/* Setup the User Stop Function. It will determine which sequence plays next. */
	Tags[0].ta_Tag = JGLR_TAG_STOP_FUNCTION;
	Tags[0].ta_Arg = (void *) &MyStopFunction;
	Tags[1].ta_Tag = TAG_END;
	
	SetObjectInfo( Coll1, Tags );
	SetObjectInfo( Coll2, Tags );
	SetObjectInfo( Coll3, Tags );

	/* Play the score collection. */
	LoopScores( Coll1, Coll2, Coll3 );
	
cleanup:
	UnloadScore( scon, Coll1 );
	UnloadScore( scon, Coll2 );
	UnloadScore( scon, Coll3 );

	CleanupScoreContext( scon );

	ShutItDown();

/******************************************************************************/
TermScaffold();

return 1;

}


