/********************************************************************************/
/* Written by:	Paige Wuestefeld						*/
/* SimpleLight.c								*/
/*										*/
/* Simple test of Mercury Lighting functions.					*/
/*										*/
/* 	Displays a torus with specular reflection and 3 different movable light	*/
/*	sources: directional, spotlight, and point.				*/
/*										*/
/* 	Some code taken from helloworld by Nick Triantos.			*/
/********************************************************************************/

#ifdef MACINTOSH
#include <kernel:types.h>
#include <kernel:mem.h>
#include <graphics:graphics.h>
#include <graphics:view.h>
#include <graphics:clt:gstate.h>
#include <misc:event.h>
#include <stdio.h>
#include <stdlib.h>
#else
#include <kernel/types.h>
#include <kernel/mem.h>
#include <graphics/graphics.h>
#include <graphics/view.h>
#include <graphics/clt/gstate.h>
#include <misc/event.h>
#include <stdio.h>
#include <stdlib.h>
#endif

#include "graphicsenv.h"
#include "mercury.h"
#include "matrix.h"
#include "datatorus.h"

  typedef struct AppData {
    CloseData *close;
    Matrix *skewMatrix;
    Matrix *camMatrix;
    uint32 curRenderScreen;
    float rx, ry, rz;		/* rotation of object */
    float tx, ty, tz;		/* location of object */
  }
AppData;


#define kMaxVertsPerPod     400
#define kMaxCmdListWords    400
#define kStartCamPosnX      0.0
#define kStartCamPosnY      0.0
#define kStartCamPosnZ      300.0
#define Pi                  3.141592653589793
#define kRadian             (Pi/180.0)
#define kIncrement          5.0
#define kMoveIncrement      4.0
#define kMoveCameraIncrement    (7.5f)

/* Prototypes for routines only used in this file */
void Program_Init (GraphicsEnv * genv, AppData ** app);
void Program_Begin (GraphicsEnv * genv, AppData * app);

/* --------------------------------------------------------------------- */


int
main (int argc, char *argv[])
{
  Err err;
  GraphicsEnv *genv;
  AppData *appData;
  int i;
  Vector3D *lightDir;

  printf ("Usage:\n");
  printf ("SimpleLightIT -d x y z\n");
  printf ("where x y z are pre-normalized directional light parameters\n");
  printf ("The default is 0 0 -1\n");

  lightDir = Vector3D_Construct ();
  Vector3D_Set (lightDir, 0.0, 0.0, -1.0);

  for (i = 1; i < argc; i++) {

    if (strcmp (argv[i], "-d") == 0) {
      if (i + 3 <= argc)
	Vector3D_Set (lightDir, (float) (atoi (argv[i + 1])),
		(float) (atoi (argv[i + 2])), (float) (atoi (argv[i + 3])));
    }
  }

  /* need to be normalized or lighting calculations wrong */
  Vector3D_Normalize (lightDir);

  printf ("Directional light direction (post-normalization): \n");
  Vector3D_Print (lightDir);

  /* Use this location as default directional light position */
  LightDirDataSpec.nx = lightDir->x;
  LightDirDataSpec.ny = lightDir->y;
  LightDirDataSpec.nz = lightDir->z;


  /* Setup Display, GState */
  genv = GraphicsEnv_Create ();
  if (genv == NULL) {
    printf ("Couldn't allocate graphics environment.  Exiting.\n");
    exit (1);
  }
  err = GraphicsEnv_Init (genv, 640, 480, 32);
  if (err < 0) {
    PrintfSysErr (err);
    exit (err);
  }

  /* Setup the EventBroker */
  err = InitEventUtility (1, 0, 0);
  if (err < 0) {
    PrintfSysErr (err);
    exit (err);
  }

  /* Load & initialize data, then begin */
  Program_Init (genv, &appData);
  Program_Begin (genv, appData);

  /* Cleanup */
  printf ("Program exiting.\n");
  GraphicsEnv_Delete (genv);

  return 0;
}


/* --------------------------------------------------------------------- */


void
Program_Init (GraphicsEnv * genv, AppData ** app)
{
  ViewPyramid vp;
  AppData *appData;

  *app = AllocMem (sizeof (AppData), MEMTYPE_NORMAL);
  if (*app == NULL) {
    printf ("Couldn't create AppData structure.  Exiting\n");
    exit (1);
  }
  appData = *app;

  /* Initialize Mercury */
  appData->close = M_Init (kMaxVertsPerPod, kMaxCmdListWords, genv->gs);
  if (appData->close == NULL) {
    printf ("Couldn't init Mercury.  Exiting\n");
    exit (1);
  }
  appData->close->fwclose = 1.01;
  appData->close->fwfar = 100000.0;
  appData->close->fscreenwidth = genv->d->width;
  appData->close->fscreenheight = genv->d->height;
  appData->close->depth = genv->d->depth;

  /* Setup camera */
  appData->skewMatrix = AllocMem (sizeof (Matrix), MEMTYPE_NORMAL);
  if (appData->skewMatrix == NULL) {
    printf ("Couldn't allocate skew matrix.  Exiting\n");
    exit (1);
  }

  vp.left = -1.0;
  vp.right = 1.0;
  vp.top = 0.75;		/* Basic 4/3 aspect ratio, with positive Y UP */
  vp.bottom = -0.75;
  vp.hither = 1.2;
  Matrix_Perspective (appData->skewMatrix, &vp, 0.0, (float) (genv->d->width),
		      0.0, (float) (genv->d->height), 1.0 / gHitherPlane);

  appData->camMatrix = Matrix_Construct ();
  if (appData->camMatrix == NULL) {
    printf ("Couldn't allocate camera matrix.  Exiting\n");
    exit (1);
  }
  {
    Vector3D center;
    Vector3D position;

    Vector3D_Set (&center, 0.0f, 0.0f, 0.0f);
    Matrix_Identity (appData->camMatrix);
    Matrix_Translate (appData->camMatrix,
		      kStartCamPosnX, kStartCamPosnY, kStartCamPosnZ);
    Matrix_GetTranslation (appData->camMatrix, &position);
    Matrix_LookAt (appData->camMatrix, (Point3D *) & center,
		   (Point3D *) & position, 0.0f);
  }

  M_SetCamera (appData->close, appData->camMatrix, appData->skewMatrix);

}


/* --------------------------------------------------------------------- */

static void
Controller_Update (AppData * appData, ControlPadEventData * cped,
		   Matrix * matrix)
{
  Vector3D center;
  Vector3D moveCamera;

  Vector3D_Zero (&moveCamera);

/* freely rotate about XY */

  if (cped->cped_ButtonBits & ControlA) {

    appData->rx += kIncrement * kRadian;
    appData->ry += kIncrement * kRadian;
    Matrix_Identity (matrix);
    Matrix_Move (matrix, appData->tx, appData->ty, appData->tz);
    Matrix_TurnXLocal (matrix, appData->rx);
    Matrix_TurnYLocal (matrix, appData->ry);
  }

/* Move object */
  if (cped->cped_ButtonBits & ControlB) {

    if (cped->cped_ButtonBits & ControlUp) {
      appData->tz -= kMoveIncrement;
    }
    if (cped->cped_ButtonBits & ControlDown) {
      appData->tz += kMoveIncrement;
    }
    if (cped->cped_ButtonBits & ControlRight) {
      appData->tx += kMoveIncrement;
    }
    if (cped->cped_ButtonBits & ControlLeft) {
      appData->tx -= kMoveIncrement;
    }
    if (cped->cped_ButtonBits & ControlLeftShift) {
      appData->ty -= kMoveIncrement;
    }
    if (cped->cped_ButtonBits & ControlRightShift) {
      appData->ty += kMoveIncrement;
    }
    Matrix_Identity (matrix);
    Matrix_Move (matrix, appData->tx, appData->ty, appData->tz);

  }

/* Reposition Point light source */
  if (cped->cped_ButtonBits & ControlC) {

    if (cped->cped_ButtonBits & ControlRightShift)
      LightPointData.x -= 3.0;
    if (cped->cped_ButtonBits & ControlLeftShift)
      LightPointData.x += 3.0;

  }

/* Reposition Spot light source */
  if (cped->cped_ButtonBits & ControlD) {

    if (cped->cped_ButtonBits & ControlRightShift)
      LightSpotData.z -= 3.0;
    if (cped->cped_ButtonBits & ControlLeftShift)
      LightSpotData.z += 3.0;

  }

/* Reset */
  if (cped->cped_ButtonBits & ControlStart) {

    Vector3D_Set (&center, 0.0f, 0.0f, 0.0f);
    Matrix_Identity (appData->camMatrix);
    Matrix_Identity (matrix);
    Matrix_Translate (appData->camMatrix,
		      kStartCamPosnX, kStartCamPosnY, kStartCamPosnZ);
  }

/* Move camera - simple Up, Down, Right, Left */
  if (!(cped->cped_ButtonBits & ControlB)) {

    if (cped->cped_ButtonBits & ControlUp)
      moveCamera.y = kMoveCameraIncrement;
    if (cped->cped_ButtonBits & ControlDown)
      moveCamera.y = -kMoveCameraIncrement;

    if (cped->cped_ButtonBits & ControlLeft)
      moveCamera.x = kMoveCameraIncrement;
    if (cped->cped_ButtonBits & ControlRight)
      moveCamera.x = -kMoveCameraIncrement;

    Matrix_MoveByVector (appData->camMatrix, &moveCamera);

  }
}


/* --------------------------------------------------------------------- */


void
Program_Begin (GraphicsEnv * genv, AppData * app)
{
  ControlPadEventData cped;
  Pod *firstPod;
  bool done = FALSE;
  bool LookCenter = FALSE;
  Vector3D *camDist, *ModelCenter;

  printf ("\n\nDisplaying torus centered at 0, 0, 0 with white material and three lights.\n");

  printf ("White directional with direction as specified by user (default: 0 0 -1)\n");
  printf ("Soft pink specular reflections.\n");

  printf ("Blue soft spot positioned at 0, 0, 100.\n");
  printf ("The spot can be moved in Z with Controller D and Right and Left Shift\n");

  printf ("Green point light originally positioned at 100, 0, 100.\n");
  printf ("The point light can be moved in X with Controller C and Right and Left Shift\n");

  printf ("Directions:\n");
  printf ("Arrows move camera.\n");
  printf ("Controller A rotates torus in XY.\n");
  printf ("Controller B moves torus (with Arrows and Shifts).\n");
  printf ("Controller C and Right and Left Shifts repositions point light in X.\n");
  printf ("Controller L and Right and Left Shifts repositions spot light in Z.\n");
  printf ("Controller P camera always looks at center of object.\n");
  printf ("Controller R print out camera and object matrix.\n");
  printf ("Controller X resets view.\n");

  firstPod = &gTorusPod;

  app->curRenderScreen = 0;
  app->rx = app->ry = app->rz = 0.0;
  app->tx = app->ty = 0.0;
  app->tz = 0.0;

  camDist = Vector3D_Construct ();
  ModelCenter = Vector3D_Construct ();
  Vector3D_Set (ModelCenter, 0.0, 0.0, 0.0);

  do {
    /* Get Control Pad info */
    GetControlPad (1, FALSE, &cped);

    if (cped.cped_ButtonBits & ControlX)
      done = TRUE;

    if (cped.cped_ButtonBits & ControlE) {
      if (!LookCenter)
	printf ("Camera is now always looking at the center of the object\n");
      LookCenter = TRUE;
    }

    Controller_Update (app, &cped, firstPod->pmatrix);

    if (LookCenter) {
      Matrix_GetTranslation (app->camMatrix, camDist);
      Matrix_Identity (app->camMatrix);
      Matrix_SetTranslationByVector (app->camMatrix, camDist);
      Matrix_LookAt (app->camMatrix, ModelCenter, Matrix_GetpTranslation (app->camMatrix), 0.0);
    }

    M_SetCamera (app->close, app->camMatrix, app->skewMatrix);

    CLT_ClearFrameBuffer (genv->gs, 0.0, 0.0, 0.0, 1.0, TRUE, TRUE);

    M_DBInit (GS_Ptr (genv->gs), 0, 0, genv->d->width, genv->d->height);

    M_Draw (firstPod, app->close);

    if (cped.cped_ButtonBits & ControlF) {
      printf ("Camera Matrix\n");
      Matrix_Print (app->camMatrix);
      printf ("Object Matrix\n");
      Matrix_Print (firstPod->pmatrix);
    }

    genv->gs->gs_SendList (genv->gs);

    GS_WaitIO (genv->gs);
    ModifyGraphicsItemVA (genv->d->view,
			  VIEWTAG_BITMAP, GS_GetDestBuffer (genv->gs),
			  TAG_END);
    app->curRenderScreen = (app->curRenderScreen + 1) % 2;
    GS_SetDestBuffer (genv->gs, genv->bitmaps[app->curRenderScreen]);
    GS_BeginFrame (genv->gs);

  } while (!done);

  M_End (app->close);
}
