/********************************************************************************/
/* Written by:	Paige Wuestefeld						*/
/* SimpleMat.c									*/
/*										*/
/* Simple test of Mercury cases with no texture					*/
/*										*/
/* Displays three objects with a multicolored partially transparent background	*/
/*										*/
/* Data located in : 	SimpleMatData.c						*/
/* Cases Tested:								*/
/* 			DynLitTrans, DynLit, PreLitTrans, PreLit		*/
/*			DynLitFogTrans, DynLitFog, PreLitFogTrans, PreLitFog	*/
/*										*/
/* 	Some code taken from helloworld by Nick Triantos.			*/
/* Modified 11/19 to use MP utility						*/
/********************************************************************************/

#ifdef MACINTOSH
#include <kernel:types.h>
#include <kernel:mem.h>
#include <graphics:graphics.h>
#include <graphics:view.h>
#include <graphics:clt:gstate.h>
#include <misc:event.h>
#else
#include <kernel/types.h>
#include <kernel/mem.h>
#include <graphics/graphics.h>
#include <graphics/view.h>
#include <graphics/clt/gstate.h>
#include <misc/event.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include "graphicsenv.h"
#include "mercury.h"
#include "matrix.h"
#include "SimpleMatData.h"
#include "scalemodel.h"

/* overarching structure */

  typedef struct AppData {
    CloseData *close;
    Matrix *skewMatrix;
    Matrix *camMatrix;
    uint32 curRenderScreen;
    float rx, ry, rz;		/* rotation of object */
    float tx, ty, tz;		/* location of object */
  }
AppData;

#define RUNMP 1

/* Mercury limits we have arbitrarily set -- lower if tight on mem */
#define kMaxVertsPerPod     400
#define kMaxCmdListWords    400

/* Initial camera position in world coords */
#define kStartCamPosnX      0.0
#define kStartCamPosnY      0.0
#define kStartCamPosnZ      300.0
#define MAXPODS		       20
#define kRadian             	(PI/180.0)
#define kMoveCameraIncrement    (7.5f)
#define kMoveIncrement			(2.0f)

/* Prototypes for routines only used in this file */
void Program_Init (GraphicsEnv * genv, AppData ** app);
void Program_Begin (GraphicsEnv * genv, AppData * app);
void AddFog (void);

/* --------------------------------------------------------------------- */


int
main (int argc, char *argv[])
{
  Err err;
  GraphicsEnv *genv;
  AppData *appData;

  (void) &argv;
  (void) &argc;

  /* Setup Display, GState */

  genv = GraphicsEnv_Create ();
  if (genv == NULL) {
    printf ("Couldn't allocate graphics environment.  Exiting.\n");
    exit (1);
  }
  err = GraphicsEnv_Init (genv, 320, 240, 32);
  if (err < 0) {
    PrintfSysErr (err);
    exit (err);
  }

  /* Setup the EventBroker */

  err = InitEventUtility (1, 0, 0);
  if (err < 0) {
    PrintfSysErr (err);
    exit (err);
  }

  /* Load & initialize data, then begin */

  Program_Init (genv, &appData);
  Program_Begin (genv, appData);

  /* Cleanup */

  printf ("Program exiting.\n");
  GraphicsEnv_Delete (genv);

  return 0;
}


/* --------------------------------------------------------------------- */


void
Program_Init (GraphicsEnv * genv, AppData ** app)
{
  ViewPyramid vp;
  AppData *appData;

  *app = AllocMem (sizeof (AppData), MEMTYPE_NORMAL);
  if (*app == NULL) {
    printf ("Couldn't create AppData structure.  Exiting\n");
    exit (1);
  }
  appData = *app;

  /* Initialize Mercury */
  appData->close = M_Init (kMaxVertsPerPod, kMaxCmdListWords, genv->gs);
#if RUNMP
  M_InitMP (appData->close, MAXPODS, 100000, 0);
#endif

  if (appData->close == NULL) {
    printf ("Couldn't init Mercury.  Exiting\n");
    exit (1);
  }

  /* required close data arguments */

  appData->close->fwclose = 1.01;
  appData->close->fwfar = 100000.0;
  appData->close->fscreenwidth = genv->d->width;
  appData->close->fscreenheight = genv->d->height;
  appData->close->depth = genv->d->depth;

  /* Setup camera */
  appData->skewMatrix = AllocMem (sizeof (Matrix), MEMTYPE_NORMAL);
  if (appData->skewMatrix == NULL) {
    printf ("Couldn't allocate skew matrix.  Exiting\n");
    exit (1);
  }

  vp.left = -1.0;
  vp.right = 1.0;
  vp.top = 0.75;		/* Basic 4/3 aspect ratio, with positive Y UP */
  vp.bottom = -0.75;
  vp.hither = 1.2;

  Matrix_Perspective (appData->skewMatrix, &vp, 0.0, (float) (genv->d->width),
		      0.0, (float) (genv->d->height), 1.0 / gHitherPlane);

  appData->camMatrix = Matrix_Construct ();
  if (appData->camMatrix == NULL) {
    printf ("Couldn't allocate camera matrix.  Exiting\n");
    exit (1);
  }
  {
    Vector3D center;
    Vector3D position;

    Vector3D_Set (&center, 0.0f, 0.0f, 0.0f);
    Matrix_Identity (appData->camMatrix);
    Matrix_Translate (appData->camMatrix,
		      kStartCamPosnX, kStartCamPosnY, kStartCamPosnZ);
    Matrix_GetTranslation (appData->camMatrix, &position);
    Matrix_LookAt (appData->camMatrix, (Point3D *) & center,
		   (Point3D *) & position, 0.0f);
  }

  M_SetCamera (appData->close, appData->camMatrix, appData->skewMatrix);

}


/* --------------------------------------------------------------------- */

static void
Controller_Update (AppData * appData, ControlPadEventData * cped,
		   Matrix * matrixObj, Matrix * matrixBack)
{
  Vector3D moveObj;
  Vector3D moveBack;
  Vector3D moveCamera;

  Vector3D_Zero (&moveObj);
  Vector3D_Zero (&moveBack);
  Vector3D_Zero (&moveCamera);

  if (cped->cped_ButtonBits & ControlA) {
    if (cped->cped_ButtonBits & ControlUp)
      moveObj.y = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlDown)
      moveObj.y = -kMoveIncrement;
    if (cped->cped_ButtonBits & ControlRight)
      moveObj.x = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlLeft)
      moveObj.x = -kMoveIncrement;
    if (cped->cped_ButtonBits & ControlLeftShift)
      moveObj.z = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlRightShift)
      moveObj.z = -kMoveIncrement;
  }

  if (cped->cped_ButtonBits & ControlB) {
    if (cped->cped_ButtonBits & ControlUp)
      moveBack.y = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlDown)
      moveBack.y = -kMoveIncrement;
    if (cped->cped_ButtonBits & ControlRight)
      moveBack.x = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlLeft)
      moveBack.x = -kMoveIncrement;
    if (cped->cped_ButtonBits & ControlLeftShift)
      moveBack.z = kMoveIncrement;
    if (cped->cped_ButtonBits & ControlRightShift)
      moveBack.z = -kMoveIncrement;
  }

  if ((cped->cped_ButtonBits & ControlD)) {
    Matrix_TurnX (matrixObj, kRadian * 2.0);
    Matrix_TurnY (matrixObj, kRadian * 2.0);
  }

  if (cped->cped_ButtonBits & ControlStart) {
    Matrix_Identity (matrixObj);
    Matrix_Identity (matrixBack);
  }

  if ((!(cped->cped_ButtonBits & ControlA) && !(cped->cped_ButtonBits & ControlB))) {
    if (cped->cped_ButtonBits & ControlUp)
      moveCamera.y = kMoveCameraIncrement;
    if (cped->cped_ButtonBits & ControlDown)
      moveCamera.y = -kMoveCameraIncrement;

    if (cped->cped_ButtonBits & ControlLeft)
      moveCamera.x = kMoveCameraIncrement;
    if (cped->cped_ButtonBits & ControlRight)
      moveCamera.x = -kMoveCameraIncrement;

    if (cped->cped_ButtonBits & ControlLeftShift)
      moveCamera.z = kMoveCameraIncrement;
    if (cped->cped_ButtonBits & ControlRightShift)
      moveCamera.z = -kMoveCameraIncrement;
  }


  Matrix_TranslateByVector (matrixObj, &moveObj);
  Matrix_TranslateByVector (matrixBack, &moveBack);
  Matrix_TranslateByVector (appData->camMatrix, &moveCamera);

}


/* --------------------------------------------------------------------- */

void
Program_Begin (GraphicsEnv * genv, AppData * app)
{
  ControlPadEventData cped;
  Pod *firstPod;
  Pod *secondPod;
  bool done = FALSE;
  bool moveBack = FALSE;
  Bitmap *myDestBM;
  int numPods = 3;
  int frame = 0;
  int i;
  float inout;

  printf ("Usage:\n");
  printf ("SimpleMatIT\n");
  printf ("Displaying dynamically lit box and torus and prelit cube with paritially transparent background\n");
  printf ("Back plane is a BillBoardY object\n\n");
  printf ("Arrows move camera\n");
  printf ("Control A and arrows+shifts moves objects\n");
  printf ("Control B and arrows+shifts moves background\n");
  printf ("Control L rotates in XY\n");
  printf ("Control P translates background in and out\n");
  printf ("Control R fogs display\n");
  printf ("Control X exits fogged display\n");

  /* scale each vertex in the torus */
  Model_Scale (&gTorusDynPod, .5);

  firstPod = &gCubePrePod;
  secondPod = &gPlanePrePod;

  app->curRenderScreen = 0;

  for (i = 0; i < 2; i++) {

    firstPod = M_Sort (numPods, firstPod, NULL);

    do {

      /* Get Control Pad info */
      GetControlPad (1, FALSE, &cped);

      if ((!i) && (cped.cped_ButtonBits & ControlF))
	done = TRUE;

      if ((i) && (cped.cped_ButtonBits & ControlX))
	done = TRUE;

      Controller_Update (app, &cped, firstPod->pmatrix, secondPod->pmatrix);

      /* On Control P translate transparent background in an out to check tranparency */
      if (((cped.cped_ButtonBits & ControlE) || moveBack)) {
	frame++;
	moveBack = TRUE;
	if (((frame / 800) % 2) == 0)
	  inout = gHitherPlane / 40.0;
	else
	  inout = -gHitherPlane / 40.0;

	Matrix_Move (secondPod->pmatrix, 0.0, 0.0, inout);
      }

      Matrix_BillboardY (secondPod->pmatrix, Matrix_GetpTranslation (app->camMatrix));

      M_SetCamera (app->close, app->camMatrix, app->skewMatrix);

      /* Clear screen and Z-buffer */
      CLT_ClearFrameBuffer (genv->gs, 0.0, 0.0, 0.0, 1.0, TRUE, TRUE);

      /* init destination buffer to known state (do each frame) */
      M_DBInit (GS_Ptr (genv->gs), 0, 0, genv->d->width, genv->d->height);

      /* set src address for transparency */
      myDestBM = (Bitmap *) LookupItem (GS_GetDestBuffer (genv->gs));
      app->close->srcaddr = (uint32) myDestBM->bm_Buffer;

      app->close->fogcolor = M_PackColor (0.0, 0.0, 0.0);

      /* draw pods */
      M_Draw (firstPod, app->close);
      M_Draw (secondPod, app->close);
#if RUNMP
	  M_DrawEnd (app->close);
#endif

      /* send cmd list */
      genv->gs->gs_SendList (genv->gs);

      GS_WaitIO (genv->gs);
      ModifyGraphicsItemVA (genv->d->view,
			    VIEWTAG_BITMAP, GS_GetDestBuffer (genv->gs),
			    TAG_END);
      app->curRenderScreen = (app->curRenderScreen + 1) % 2;
      GS_SetDestBuffer (genv->gs, genv->bitmaps[app->curRenderScreen]);
      GS_BeginFrame (genv->gs);

    } while (!done);

    printf ("Adding fog\n");
    /* now test cases with fog */
    AddFog ( );
    done = FALSE;
  }

  M_End (app->close);
#if RUNMP
  M_EndMP (app->close);
#endif
}

void
AddFog (void)
{

  /* change pcase, alpha, and lights */

  gCubePrePod.pcase = M_SetupPreLitFog;
  gCubePrePod.pmaterial->base.a = 0.7;
  gCubePrePod.plights = gFogOnlyLight;

  gPlanePrePod.pcase = M_SetupPreLitFogTrans;
  gPlanePrePod.plights = gFogOnlyTransLight;

  gBoxDynPod.pcase = M_SetupDynLitFogTrans;
  gBoxDynPod.plights = gDirPointFogLightList;
  gBoxDynPod.pmaterial->base.a = 0.7;

  gTorusDynPod.pcase = M_SetupDynLitFog;
  gTorusDynPod.plights = gDirPointFogLightList;
  gTorusDynPod.pmaterial->base.a = 0.7;

}



