/* @(#) scaf_MemDebug.c 96/07/11 1.1 */
#include <kernel/debug.h>
#include <kernel/mem.h>
#include <kernel/types.h>
#include <strings.h>

#define	ErrReport		kprintf
#define	ErrNumReport	PrintfSysErr
#define	Report			kprintf


static	Boolean	gScafMemDebugInited = FALSE;		/* set to true if MemDebug initialized */
static	Boolean gDisplayForThisTaskOnly = FALSE;	/* set to true if we display for this task only */
static	Boolean gDisplaySuper = FALSE;				/* set to true if we display supervisor mode allocations */

/* ------------------------------------------------------------------------------- */
/*	startup memory debugging based on the options in the option string */
/*	see the switch statement in this routine for available options */

void	StartMemDebug( char *optionStr )
{
	Err		err;
	uint32	controlFlags = 0x00000000;
	uint32	n;
	uint32	len;
	
	
	Report( "### Start Memory Debugging -- option string = \"%s\"\n", optionStr );

	if ( (err = CreateMemDebug( NULL )) < 0 )				/* initialize MemDebug */
	{
		ErrNumReport( err );
		ErrReport( "Error $%x calling CreateMemDebug\n");
		gScafMemDebugInited = FALSE;
		return;
	}
	
	gScafMemDebugInited = TRUE;
	
	len = strlen( optionStr );
	
	for( n = 0; n < len; n++ )
		switch ( optionStr[n] )	{							/* set up the option flags based on the option string */
		
			case 'A':		controlFlags |= MEMDEBUGF_ALLOC_PATTERNS;
							break;
							
			case 'F':		controlFlags |= MEMDEBUGF_FREE_PATTERNS;
							break;
							
			case 'P':		controlFlags |= MEMDEBUGF_PAD_COOKIES;
							break;
							
			case 'B':		controlFlags |= MEMDEBUGF_BREAKPOINT_ON_ERRORS;
							break;
							
			case 'K':		controlFlags |= MEMDEBUGF_CHECK_ALLOC_FAILURES;
							break;
							
			case 'T':		gDisplayForThisTaskOnly = TRUE;		/* display for this task only */
							break;
							
			case 'S':		gDisplaySuper = TRUE;				/* display supervisor mode allocations also */
							break;
							
							
		}
		
	if ( (err = ControlMemDebug( controlFlags )) < 0 )				/* setup MemDebug options */
	{
		ErrNumReport( err );
		ErrReport( "Error $%x calling ControlMemDebug(\"%s\")\n", err, optionStr );
		return;
	}
}


	
	
/* ------------------------------------------------------------------------------- */
/*	terminates memory debugging and dump the statistics */

void	TermMemDebug( void )
{
	Err		err;
	TagArg	tag[5];
	uint32	indx = 0;
	
	if ( !gScafMemDebugInited )	return;			/* don't do if we didn't initialize the memdebug stuff */
	gScafMemDebugInited = FALSE;
	Report( "### Terminate Memory Debugging\n" );
	
	if ( !gDisplayForThisTaskOnly )
	{
		tag[indx].ta_Tag =		DUMPMEMDEBUG_TAG_TASK;	/* default is for all tasks but */
		tag[indx++].ta_Arg = 	0;						/* this tag with zero val prints for all tasks */
	}
	if ( gDisplaySuper )
	{
		tag[indx].ta_Tag =		DUMPMEMDEBUG_TAG_SUPER;	/* display supervisor mode allocations */
		tag[indx++].ta_Arg = 	(void *) TRUE;
	}
	
	tag[indx].ta_Tag =	TAG_END;					/* don't forget to terminate the tag array */
	tag[indx].ta_Arg = 	0;
	
	
	if (!gScafMemDebugInited)	return;				/* do nothing if Memdebug was not initialized */
	
	if ( (err = DumpMemDebug( tag )) < 0 )			/* Dump MemDebug */
	{
		ErrNumReport( err );
		ErrReport( "Error $%x calling DumpMemDebug\n");
	}
	
	if ( ( err = DeleteMemDebug( ) ) < 0 )			/* Delete MemDebug */
	{
		ErrNumReport( err );
		ErrReport( "Error $%x calling DeleteMemDebug\n");
	}
	
}

/* ------------------------------------------------------------------------------- */
/*
This function checks all current memory allocations to see if any of the memory
cookies have been corrupted. This is useful when trying to track down at which
point in a program's execution memory cookies are being trashed. 
 */

void	CheckMemDebug( void )
{
	#define BANNER	"Scaffold Memory Check:"
	Err	err;

	if ( !gScafMemDebugInited )	return;			/* don't do if we didn't initialize the memdebug stuff */
	Report( "### Check Memory Debugging\n" );

	if ( (err = SanityCheckMemDebug( BANNER, NULL)) < 0 )			/* Do a MemDebug Sanity Check*/
	{
		ErrNumReport( err );
		ErrReport( "Error $%x calling SanityCheckMemDebug\n");
	}
}

/* ------------------------------------------------------------------------------- */
/*
This function Dumps the current memdebug statistics. 
 */

void	PrintMemDebug( void )
{
	Err	err;
	uint32	indx = 0;
	TagArg	tag[5];

	Report( "### Print Memory Debugging Statistics\n" );

	if ( !gScafMemDebugInited )	return;			/* don't do if we didn't initialize the memdebug stuff */
	
	if ( !gDisplayForThisTaskOnly )
	{
		tag[indx].ta_Tag =		DUMPMEMDEBUG_TAG_TASK;	/* default is for all tasks but */
		tag[indx++].ta_Arg = 	0;						/* this tag with zero val prints for all tasks */
	}
	if ( gDisplaySuper )
	{
		tag[indx].ta_Tag =		DUMPMEMDEBUG_TAG_SUPER;	/* display supervisor mode allocations */
		tag[indx++].ta_Arg = 	(void *) TRUE;
	}
	
	tag[indx].ta_Tag =	TAG_END;					/* don't forget to terminate the tag array */
	tag[indx].ta_Arg = 	0;
	
	
	if (!gScafMemDebugInited)	return;				/* do nothing if Memdebug was not initialized */
	
	if ( (err = DumpMemDebug( tag )) < 0 )			/* Dump MemDebug */
	{
		ErrNumReport( err );
		ErrReport( "Error $%x calling DumpMemDebug\n");
	}
}
